#! /usr/bin/perl

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'config.pl,v control version 1.5 2002/01/05 02:39:07'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "User interface functions man page config items";


#
#  copyright (c) 2000, 2001, Gordon B. Aaseng
#                      Personal IR&D
#                      Glendale, AZ
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  

#
#  FILE NAME
#
#  config.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This file contains functions for displaying and editing man page
#  configuration elements.  These elements include header, directory
#  and support files.  Other elements may be added as needed.  The
#  man page header data is displayed in the configuration displays.
#  Other elements are displayed in the main user interface.  A set of
#  control buttons are provided to determine which elements are
#  included in an initialization file and which are maintained in
#  memory only.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  config.pl,v
#  Revision 1.5  2002/01/05 02:39:07  gordon
#  Changed license to GPL.
#
#  Revision 1.4  2001/12/28 05:48:11  gordon
#  Added controls for HTML formatting options, extraction options, and
#  storage class selections.
#
#  Revision 1.3  2001/11/24 13:59:37  gordon
#  Added controls for HTML directory and index, and page format controls
#  (man or html) to the items saved in the initialization file.
#
#  Revision 1.2  2001/01/21 14:26:10  gordon
#  Added a check of the FileSelect return to make sure that the name
#  is in a valid directory; a workaround for a FileSelect bug.
#
#  Revision 1.1  2000/12/14 05:33:38  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use subs qw( ConfigUsage ConfigCmInfo );
use subs qw( DisplayConfig UpdateConfig SelectRcFile WriteConfig );

# ***************************************************************
#
#  Subroutine USAGE
#
#  Print script help message
#
# ***************************************************************

sub ConfigUsage {
   printf( STDOUT "Script synopsis:\n" );
   printf( STDOUT "Usage:   require 'config.pl'\n" );
   return;
}

# ***************************************************************
#
#  Subroutine CMINFO
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub ConfigCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}


# ***************************************************************
#
#  Subroutine SetConfig
#
#  Description:  This function creates a widget for editing the man page
#                header and author information.  The widget is a separate
#                top level widget that is popped up when selected from the
#                main user interface. It places the default configuration
#                data into text and entry widgets, and updates data
#                structures when the 'OK' button is selected.
#
# ***************************************************************

sub SetConfig {
   my $top        = $_[0];

   my $reinitbtn;

   my $cfgtop = $top->Toplevel( -title => 'Set Man Cofiguration' );

#  This is the configuration information frame, which has author, section,
#  source and anything else in the srcdocrc initialization file.

   my $infoframe = $cfgtop->Frame();
   $infoframe->pack( -side =>'top', -ipadx => '0.2i' );

   my $infolabel = $infoframe->Label( -text => "Man Page Configuration",
               -width              => 40,
               -height             => 2,
               -wraplength         => '4i',
               -anchor             => 'n'
             );
   $infolabel->pack( -side => 'top' );

   my $authorframe = $infoframe->Frame();
   $authorframe->pack( -side => 'top' );

   my $authorlabel = $authorframe->Label(
               -text => "Man Page Author",
               -justify => 'left',
               -width   => 40,
               -anchor  => 'w'
             );
   $authorlabel->pack( -side => 'top' );

   my $AuthorText = $authorframe->Text(
               -state  => 'normal',
               -height => 10,
               -width  => 40,
               -wrap   => 'word'
             );
   $AuthorText->pack( -side => 'bottom' );

#  Create labels and entry widgets for the man page section, a number or small
#  number of letters

   my $infoht = 24;
   my $sectionframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $sectionframe->pack( -side => 'top' );

   my $label_size = 16;
   my $SectionLabel = $sectionframe->Label(
                 -text => "Man Page Section",
                 -width => $label_size,
                 -anchor  => 'w'
               );

   $SectionLabel->place( -in => $sectionframe, -relx => 0.0 );

   my $SectionEntry = $sectionframe->Entry( -width => 2 );
   $SectionEntry->place( -in => $SectionLabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );

#  Create labels and entry widgets for the man page source, which is a name
#  such as Unix, GNU, BSD, etc.

   my $srcframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $srcframe->pack( -side => 'top' );
   my $SrcLabel = $srcframe->Label(
                 -text => "Man Page Source",
                 -width => $label_size,
                 -anchor  => 'w'
               );
   $SrcLabel->place( -in => $srcframe, -relx => 0.0 );

   my $SrcEntry = $srcframe->Entry( -width => 12 );
   $SrcEntry->place( -in => $SrcLabel,
                     -relx => 1.0,
                     -rely => 0.5,
                     -anchor => 'w' );

   my $ManualFrame = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $ManualFrame->pack( -side => 'top' );

   my $ManualLabel = $ManualFrame->Label(
                 -text => "Manual Name ",
                 -width => $label_size,
                 -justify => 'left',
                 -anchor  => 'w'
               );

   $ManualLabel->place( -in => $ManualFrame, -relx => 0.0 );

   my $ManualEntry = $ManualFrame->Entry( -width => 20 );
   $ManualEntry->place( -in => $ManualLabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w' );

#  Create a separator line, then a set of checkbuttons for specifying which
#  things to save to file.

   my $sepframe = $infoframe->Frame(
                          -height => 6,
                          -width => 300,
                          -borderwidth => 2,
                          -relief => 'raised' )->pack( -side => 'top');

   my $itemlabel = $infoframe->Label(
                     -text => "Cfg Items\nSelect the items to be saved to file",
                     -relief => 'raised'
                   );
   $itemlabel->pack( -side => 'top', -ipady => '0.2i', -ipadx => '0.2i' );
   my $itemframe = $infoframe->Frame( -height => 30 )->pack(
                                            -side => 'top', -fill =>'x' );

   my $srcitem = $itemframe->Checkbutton(
                      -text     => "Source",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{source}
                   );
   $srcitem->place( -in     => $itemframe,
                    -relx   => 0.1,
                    -y      => 0,
                    -anchor => 'nw' );

   my $authoritem = $itemframe->Checkbutton(
                      -text => "Author",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{author}
                   );
   $authoritem->place( -in => $srcitem,
                    -relx => 1.0,
                    -rely => 0.5,
                    -anchor => 'w' );

   my $manualitem = $itemframe->Checkbutton(
                      -text => "Manual",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{manual}
                   );
   $manualitem->place( -in => $authoritem,
                    -relx => 1.0,
                    -rely => 0.5,
                    -anchor => 'w' );

   my $sectionitem = $itemframe->Checkbutton(
                      -text => "Section",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{section}
                   );
   $sectionitem->place( -in => $manualitem,
                    -relx => 1.0,
                    -rely => 0.5,
                    -anchor => 'w' );

#  Set up a second row of checkbuttons for the cfg itmes to be written.

   my $item2frame = $infoframe->Frame( -height => 30 )->pack(
                                            -side => 'top', -fill => 'x' );
   my $srcdiritem = $item2frame->Checkbutton(
                      -text => "Source Dir",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{srcdir}
                   );
   $srcdiritem->place( -in     => $item2frame,
                       -relx   => 0.02,
                       -y      => 0,
                       -anchor => 'nw' );

   my $tagitem = $item2frame->Checkbutton(
                      -text     => "Tag File",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{tagfile}
                   );
   $tagitem->place( -in     => $srcdiritem,
                    -relx   => 1.0,
                    -rely   => 0.5,
                    -anchor => 'w' );

   my $mandiritem = $item2frame->Checkbutton(
                      -text     => "Man Dir",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{mandir}
                   );
   $mandiritem->place( -in => $tagitem,
                       -relx => 1.0,
                       -rely => 0.5,
                       -anchor => 'w' );

   my $htmlidxitem = $item2frame->Checkbutton(
                      -text     => "HTML Index",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{indexfile}
                   );
   $htmlidxitem->place( -in => $mandiritem,
                       -relx => 1.0,
                       -rely => 0.5,
                       -anchor => 'w' );

#  Add a third row of checkbuttons for HTML formatting and function type
#  settings.

   my $item3frame = $infoframe->Frame( -height => 30 )->pack(
                                            -side => 'top', -fill => 'x' );

   my $htmldiritem = $item3frame->Checkbutton(
                      -text => "HTML Dir",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{htmldir}
                   );
   $htmldiritem->place( -in     => $item3frame,
                       -relx   => 0.05,
                       -y      => 0,
                       -anchor => 'nw' );

   my $fmtitem = $item3frame->Checkbutton(
                      -text     => "Page Formats",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{fmtsel}
                   );
   $fmtitem->place( -in     => $htmldiritem,
                    -relx   => 1.0,
                    -rely   => 0.5,
                    -anchor => 'w' );

   my $stgitem = $item3frame->Checkbutton(
                      -text     => "Function Types",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{stgmode}
                   );
   $stgitem->place( -in     => $fmtitem,
                    -relx   => 1.0,
                    -rely   => 0.5,
                    -anchor => 'w' );

#  Add another row for the comment extraction settings

   my $item4frame = $infoframe->Frame( -height => 30 )->pack(
                                            -side => 'top', -fill => 'x' );

   my $cmtitem = $item4frame->Checkbutton(
                      -text => "Comment Types",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{cmtflag}
                   );
   $cmtitem->place( -in     => $item4frame,
                    -relx   => 0.15,
                    -y      => 0,
                    -anchor => 'nw' );

   my $extitem = $item4frame->Checkbutton(
                      -text     => "Update Mode",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$RcSettings{updatemode}
                   );
   $extitem->place( -in     => $cmtitem,
                    -relx   => 1.0,
                    -rely   => 0.5,
                    -anchor => 'w' );

#  Create a panel for displaying and editing the initialization file name.
#  If the file name is changed, 'Save' operations use the specified file.
#  'Save As...' will ignore the name and bring up a file selector.  The
#  'Re-Init' button will read the file and update the display with data
#  from the maned file.

   my $manrcframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $manrcframe->pack( -side => 'top' );

   my $manrclabel = $manrcframe->Label(
                 -text => "Man Init File",
                 -width => $label_size,
                 -anchor  => 'w'
               );

   $manrclabel->place( -in => $manrcframe, -relx => 0.0 );
   my $manrcentry = $manrcframe->Entry(
                       -width        => 20,
                       -textvariable => \$DispInitFile
                      );
   $manrcentry->place( -in       => $manrclabel,
                       -relx   => 1.0,
                       -rely   => 0.5,
                       -anchor => 'w'
                     );
   if( -e $InitFile ) { $DispInitFile = $InitFile; }
   my $bcmd = sub { if( $DispInitFile ne $InitFile ) {
                      if( -e $DispInitFile ) {
                        $InitFile = $DispInitFile;
                        $reinitbtn->configure( -state => 'normal' );
                      }
                      else { PrintStat( "Invalid Man Directory $DispInitFile",
                                        'warning' );
                             $DispInitFile = $InitFile;
                           }
                    }
                  };
   $manrcentry->bind( '<Return>'   => $bcmd );
   $manrcentry->bind( '<FocusOut>' => $bcmd );
   $manrcentry->bind( '<Leave>'    => $bcmd );

   my $cfgbtns = $cfgtop->Frame();
   $cfgbtns->pack( -side => 'top', -ipady => '0.2i' );

   my $okbtn = $cfgbtns->Button(
                -text    => 'OK',
                -command => sub {
                   $manrc_desc{header}{section} = $SectionEntry->get();
                   UpdateConfig( $AuthorText );
                   $manrc_desc{header}{source} = $SrcEntry->get();
                   $manrc_desc{header}{manual} = $ManualEntry->get();
                   $cfgtop->withdraw
                }
               );
   $okbtn->pack( -side => 'left' );
   my $savebtn = $cfgbtns->Button(
                -text    => 'Save',
                -command => sub {
                   $manrc_desc{header}{section} = $SectionEntry->get();
                   UpdateConfig( $AuthorText );
                   $manrc_desc{header}{source} = $SrcEntry->get();
                   $manrc_desc{header}{manual} = $ManualEntry->get();
                   WriteConfig;
                 }
               );
   $savebtn->pack( -side => 'left' );

   my $save_asbtn = $cfgbtns->Button(
                -text    => 'Save As...',
                -command => sub {
                   $manrc_desc{header}{section} = $SectionEntry->get();
                   UpdateConfig( $AuthorText );
                   $manrc_desc{header}{source} = $SrcEntry->get();
                   $manrc_desc{header}{manual} = $ManualEntry->get();
                   SelectRcFile( $top );
                   $manrcentry->delete( 0, 'end' );
                   $manrcentry->insert( 'end', $InitFile );
                 }
               );
   $save_asbtn->pack( -side => 'left' );

   $reinitbtn = $cfgbtns->Button(
                -text    => 'Re-Init',
                -state   => 'disabled',
                -command => sub { ReadRcFile();
                                  DisplayConfig( $AuthorText, $SectionEntry,
                                                 $SrcEntry, $ManualEntry );
                                  $reinitbtn->configure( -state => 'disabled' );
                                }
               );
   $reinitbtn->pack( -side => 'left' );

   my $cnxbtn = $cfgbtns->Button(
                -text    => 'Cancel',
                -command => [ $cfgtop, 'withdraw' ]
               );
   $cnxbtn->pack( -side => 'left' );

   DisplayConfig( $AuthorText, $SectionEntry, $SrcEntry, $ManualEntry );

   return;
}

# ***************************************************************
#
#  Subroutine DisplayConfig
#
#  Description:  This function displays the man page header information in
#                the Man Configuration panels created in SetConfig.
#
# ***************************************************************

sub DisplayConfig {
   my $author_panel  = shift;
   my $section_panel = shift;
   my $src_panel     = shift;
   my $manual_panel  = shift;

   my $authortext = "";
   my $para_idx;
   my @author_paras = @{$manrc_desc{author} };

   $author_panel->delete( '1.0', 'end' );
   $section_panel->delete( 0, 'end' );
   $src_panel->delete( 0, 'end' );
   $manual_panel->delete( 0, 'end' );

   for $para_idx (0 .. $#author_paras) {
      $authortext = $authortext . $author_paras[$para_idx]->{text};
      if( $para_idx < $#author_paras ) { $authortext .= "\n\n"; }
   }
   $author_panel->insert( 'end', $authortext );

   $section_panel->insert( 0, $manrc_desc{header}{section} );
   $src_panel->insert( 0, $manrc_desc{header}{source} );
   $manual_panel->insert( 0, $manrc_desc{header}{manual} );

   return;
}

# ***************************************************************
#
#  Subroutine UpdateConfig
#
#  Description:  Reads the user interface Man Page Configuration panel
#                and updates the run-time data structures that contain
#                the preference or standard data.  This function is intended
#                to execute when the user changes the configuration
#                information.
#
# ***************************************************************

sub UpdateConfig {
   my $AuthorText = $_[0];

   my $newtext;
   my $textblk  = "";
   my $line;
   my @para;
   my $idx = 0;
   my @author_desc = ();


   $newtext = $AuthorText->get( '1.0', 'end' );
   $manrc_desc{author} = ();

   UpdateAuthor( $newtext, \@author_desc );

   $manrc_desc{author} = [ @author_desc ];
      
   return;
}

# ***************************************************************
#
#  Subroutine SelectRcFile
#
#  Description:  Invokes a file selector panel for selecting an
#                existing configuration file or specifying a new
#                file.
#
# ***************************************************************

sub SelectRcFile {
   my $top = $_[0];

   my $FS;
   my $new_initfile;
   my $dirname;

   $FS = $top->FileSelect( -directory => $RcStartDir );
   $new_initfile = $FS->Show;

#  Verify that the directory portion of the selected file is valid.  This is
#  done because sometimes FileSelect returns an invalid name, such as when
#  no file is selected.

   ($dirname) = ( $new_initfile =~
                  m|^ ( (?: .* / (?: \.\.?\Z(?!\n) )? )? ) [^/]* |xs );

   if( defined( $new_initfile ) and $new_initfile ne "" and -e $dirname ) {
      $InitFile = $new_initfile;
      WriteConfig;
   }

   return;
}

# ***************************************************************
#
#  Subroutine WriteConfig
#
#  Description:  Writes a configuration file that contains standard
#                man page ocnfiguration data, including section number,
#                title, and author.  Additional preference information
#                could be added to this section.
#
# ***************************************************************

sub WriteConfig {

   my  @author_paras = @{$manrc_desc{author}};
   my  @author_words;
   my  $para_ref;
   my  %para;
   my  $word;
   my  $linelen = 0;
   my  $para_no = 0;

   open( RCFILE, "> $InitFile" ) or die "Couldn't open init file $InitFile";
   if( $RcSettings{author} eq $True ) {
      printf( RCFILE "<Author>\n" );
   
      foreach $para_ref (@author_paras) {
         if( $para_no++ > 0 ) { printf( RCFILE "\n\n" ); }
         %para = %$para_ref;
         @author_words = split( ' ', $para{text} );
         foreach $word (@author_words) {
            if( $linelen == 0 ) {
               printf( RCFILE "\t%s", $word );
               $linelen = length( $word );
            } else {
               $linelen += length( $word );
               if( $linelen > 70 ) {
                  $linelen = length( $word );
                  printf( RCFILE "\n\t%s", $word );
               } else {
                  printf( RCFILE " %s", $word );
                  $linelen++;
               }
            }
         }
         $linelen = 0;
      }
      printf( RCFILE "\n</Author>\n" );
   }
   if( $RcSettings{manual} eq $True or $RcSettings{source} eq $True or
                                       $RcSettings{section} eq $True ) {
      printf( RCFILE "<ManHeader>\n" );
      if( $RcSettings{manual} eq $True ) {
         printf( RCFILE "\tMANUAL = \"%s\"\n", $manrc_desc{header}{manual} );
      }
      if( $RcSettings{source} eq $True ) {
         printf( RCFILE "\tSOURCE = \"%s\"\n", $manrc_desc{header}{source} );
      }
      if( $RcSettings{section} eq $True ) {
         printf( RCFILE "\tSECTION = \"%d\"\n", $manrc_desc{header}{section} );
      }
      printf( RCFILE "</ManHeader>\n" );
   }
   if( $RcSettings{srcdir} eq $True ) {
      printf( RCFILE "<SrcDir Name = \"$SrcDir\">\n" );
   }
   if( $RcSettings{mandir} eq $True ) {
      printf( RCFILE "<ManDir Name = \"$ManDir\">\n" );
   }
   if( $RcSettings{htmldir} eq $True ) {
      printf( RCFILE "<HtmlDir Name = \"$HtmlDir\">\n" );
   }
   if( $RcSettings{tagfile} eq $True ) {
      printf( RCFILE "<TagFile Name = \"$TagFile\">\n" );
   }
   if( $RcSettings{indexfile} eq $True ) {
      printf( RCFILE "<IndexFile Name = \"$IndexFile\">\n" );
   }
   if( $RcSettings{fmtsel} eq $True ) {
      my $spacer = "";
      printf( RCFILE "<PageFormat Type = \"" );
      if( $HtmlFormat eq $True ) {
         printf( RCFILE "HTML" );
         $spacer = ", ";
      }
      if( $ManFormat eq $True ) {
         printf( RCFILE "%sMAN", $spacer );
      }
      printf( RCFILE "\">\n" );
   }
   if( $RcSettings{stgmode} eq $True ) {
      my $spacer = "";
      my $sval;
      printf( RCFILE "<StorageMode Type = \"" );
      if( $StorageMode{public} eq $True ) { 
         printf( RCFILE "PUBLIC" );
         $spacer = ", ";
      }
      if( $StorageMode{static} eq $True ) {
         printf( RCFILE "%sSTATIC", $spacer );
      }
      printf( RCFILE "\">\n" );
   }
   if( $RcSettings{cmtflag} eq $True ) {
      printf( RCFILE "<Comment Type = \"$CmtFlag\">\n" );
   }
   if( $RcSettings{updatemode} eq $True ) {
      printf( RCFILE "<UpdateMode Type = \"$ReComment\">\n" );
   }

   close( RCFILE );

   return;
}

1;
