#! /usr/bin/perl

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'manproc.pl,v control version 1.6 2002/01/05 02:45:02'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "User interface functions for man page menu items";


#
#  copyright (c) 2000, 2001, Gordon B. Aaseng
#                      Personal IR&D
#                      Glendale, AZ
#

#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#  This software is protected as an unpublished work under the U.S. Copyright
#  Act of 1976.  All rights reserved by Gordon B. Aaseng.
#  

#
#  FILE NAME
#
#  manproc.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This file contains the functions for selecting source code modules for
#  generating man pages, and menu items for viewing and editing man pages.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  manproc.pl,v
#  Revision 1.6  2002/01/05 02:45:02  gordon
#  Changed license to GPL.
#
#  Revision 1.5  2001/12/28 05:51:31  gordon
#  Improved handling of selection of storage class.  Refreshes list of
#  functions when the selection of static or public functions changes.
#
#  Revision 1.4  2001/11/24 14:03:45  gordon
#  Corrected an anomaly in the $#{$mode} test that only showed up in debug
#  mode.
#
#  Revision 1.3  2001/01/21 13:44:23  gordon
#  Corrected SrcSelect function's ability to get the directory name from
#  a full path name.  Lifted the RE from File::Spec functions.
#
#  Revision 1.2  2000/12/16 17:34:27  gordon
#  Corrected name of CM display function.
#
#  Revision 1.1  2000/12/14 05:35:47  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use subs qw( mpUsage mpCmInfo );
use subs qw( ViewMan );

# ***************************************************************
#
#  Subroutine USAGE
#
#  Print script help message
#
# ***************************************************************

sub mpUsage {
   printf( STDOUT "Script synopsis:\n" );
   printf( STDOUT "Usage:   require 'manproc.pl'\n" );
   return;
}

# ***************************************************************
#
#  Subroutine CMINFO
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub mpCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}

# ***************************************************************
#
#  Main script body
#
# ***************************************************************


# ***************************************************************
#
#  Subroutine UpdateFuncList
#
#  Description:  This function calls a FileSelect function, runs the
#                ExtractSrc function, and then puts the list of function
#                names into the main window Listbox widget.
#
# ***************************************************************

sub UpdateFuncList {
   my $top    = $_[0];   # The top-level widget
   my $lb     = $_[1];   # The ListBox widget
   my $mode   = $_[2];   # The array of storage classes selected
   my $fnames = $_[3];   # The array of function names
   my $opmode = $_[4];   # Specifies REFRESH or SELECT from a file list

   my @newfunclist = ();
   my $func;

   if( $opmode eq "SELECT" ) {
      my $rval = SrcSelect( $top, "c" );
      if( $rval < 0 ) { return; }
      @GblFuncnames = ();
      %FuncList     = ();
      ExtractSrc;
   }
   $lb->delete( 0, 'end' );
 
   foreach $func (@GblFuncnames) {
      if( $mode->{public} eq $True && $FuncList{$func}->{stg} ne "static" ) {
         push( @newfunclist, $func );
         $lb->insert( 'end', $func );
      }
      if( $mode->{static} eq $True && $FuncList{$func}->{stg} eq "static" ) {
         push( @newfunclist, $func );
         $lb->insert( 'end', $func );
      }
   }
   @$fnames = @newfunclist;
   return;
}

# ***************************************************************
#
#  Subroutine GetManDir
#
#  Description:  Selects the man page directory using a FileSelector
#                widget.
#
# ***************************************************************

sub GetManDir {
   my $top      = shift;

   my $fs;
   my $newdir;

   $fs = $top->FileSelect( -directory => $ManDir );
   $fs->configure( -verify => ['-d' ] );
   $newdir = $fs->Show;
   if( defined( $newdir ) and $newdir ne '' ) {
      $ManDir = $newdir;
      $DispManDir = $ManDir;
      PrintStat( "Man Directory set to $ManDir", 'advise' );
   } elsif( !defined( $newdir ) ) {
      return;
   } else { PrintStat( "Invalid Directory $newdir", 'warning' ); }
  
   return;
}

# ***************************************************************
#
#  Subroutine ViewMan
#
#  Description:  Displays the selected man page in an 'xterm' window.
#                Uses the selected function name, the ManDir directory
#                and the man section specified in the 'srcdocrc' file or
#                the user interface.
#
# ***************************************************************

sub ViewMan {
   my $manfile = $_[0];
   my $op      = $_[1];
   my $manpage;
   my $sect;
   my $dir;

   if( %manrc_desc and $manrc_desc{header}{section} > 0 ) {
      $sect = $manrc_desc{header}{section};
   } else { $sect = 3; }

   $manpage = $manfile;
   $manpage =~ s/.*\///;
   $manpage =~ s/\..*$//;
   if( $ManDir =~ /man/ ) {
      ($dir) = ( $ManDir =~ /(.*\/)man.*$/ );
   } else { $dir = $ManDir; }

   if( $op eq $View ) {
      my $mancmd = "xterm -e man -S $sect -M $dir $manpage &";
      system( $mancmd );
   } elsif( $op eq $Editman ) {
      my $mansrc = $ManDir . "/" . $manfile . "." . $sect;
      my $editcmd = "gvim +\"set tag=$TagFile\" $mansrc &";
      system( $editcmd );
   }
   return;
}

# ***************************************************************
#
#  Subroutine SrcSelect
#
#  Description:  Invokes a file selector panel for selecting the
#                ctag file for use by 'vi', 'gvim' or 'emacs'.
#
# ***************************************************************

sub SrcSelect {
   my $top  = $_[0];
   my $type = $_[1];

   my $FS;
   my $f    = "*." . $type;
   my @flist;
   my $dirname;
   my $newtag;
   my $rval;

# NOTE:  The FileSelect widget appears to have a minor bug when nothing is
#        selected.  For example, when filtering for C files, then go to a
#        directory with no C files, so the directory and file lists are both
#        empty, then 'Accept', the return is a doubling of the path name,
#        e.g., /home/me//home/me.  A check for existence of the selected
#        file is used to avoid around the problem.

   $FS = $top->FileSelect( -directory => $SrcDir,
                           -filter    => $f
                          );

   if( $type eq "tag" ) {
      $newtag = $FS->Show;
      if( defined( $newtag ) and $newtag ne "" and -f $newtag ) {
         $TagFile = $newtag;
         $DispTagFile = $TagFile;
         $rval = 1;
      }
      else { $rval = -1; }
   } elsif( $type eq "c" ) {
      $FS->configure( -selectmode => 'extended' );
      (@flist) = $FS->Show;
      return( -1 ) if( !defined( $flist[0] ) );
      if( $#flist >= 0 and $flist[0] ne "" ) {
         @SrcFiles = @flist;

         ($dirname) = ( $flist[0] =~
                  m|^ ( (?: .* / (?: \.\.?\Z(?!\n) )? )? ) [^/]* |xs );

         if( defined( $dirname ) and -e $dirname ) {
            if( $dirname =~ /^\// ) { $SrcDir = $dirname; }
            elsif( $dirname =~ /^\.$/ ) {
               $SrcDir = qx/pwd/;
               chomp $SrcDir;
            } else { $SrcDir .= '/' . $dirname; }
            $DispSrcDir = $SrcDir;
            $rval = $#flist;
         } else { $rval = -1; }
      }
   }

   return( $rval );
}

# ***************************************************************
#
#  Subroutine GetSelection
#
#  Description:  Gets the selected items from the selection list
#                and calls the EditFile function to edit the selected
#                files.
#
# ***************************************************************

sub GetSelection {
   my $fl          = $_[0];
   my @func_names  = @{$_[1]};
   my $op          = $_[2];

   my  $idx;

   $psel_cnt++;

   my @active = $fl->curselection;
   for $idx ( @active ) {
      if( $Verbose eq $True ) { printf( "Active %s\n", $func_names[$idx] ); }
      if( $op eq $Edit or $op eq $Generate ) {
         EditFile( $func_names[$idx], $op,
                   $FuncList{$func_names[$idx]}->{srcfile} );
      } elsif( $op eq $View or $op eq $Editman ) {
         ViewMan( $func_names[$idx], $op );
      }
      
   }
   return;
}

1;
