#! /usr/bin/perl

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'printhtml.pl,v control version 1.4 2002/01/05 02:47:30'; }
$cpyrght = "Copyright (c) 2001, Gordon B. Aaseng";

$usage = "Functions for writing HTML man pages.";


#
#  copyright (c) 2001, Gordon B. Aaseng
#                      Personal IR&D
#                      Glendale, AZ
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  
#
#  This software is protected as an unpublished work under the U.S. Copyright
#  Act of 1976.  All rights reserved by Personal IR&D
#

#
#  FILE NAME
#
#  printhtml.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This file contains the routines for printing man pages in
#  HTML formats.  Formatting information is converted from troff
#  to HTML in some cases, and definitions in a css file are
#  assumed.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  printhtml.pl,v
#  Revision 1.4  2002/01/05 02:47:30  gordon
#  Changed license to GPL.
#
#  Revision 1.3  2001/12/28 05:54:07  gordon
#  Corrected error in printing functions with no parameters.
#
#  Revision 1.2  2001/11/30 12:59:10  gordon
#  Improved the HTML index generation so that it reads the existing index
#  and adds generated pages to it, if needed.
#
#  Revision 1.1  2001/11/22 13:33:16  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************


# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use vars qw( %FuncList %keylist %namelist @sys_includes @lcl_includes );
use vars qw( %FuncList %IndexEntries $PageWidth $ParamTableWidth );
use vars qw( $ParamLineLen );
use subs qw( PrintProto PrintHtmlBody PrintHtmlDesc PrintIndex IsPunctuation );
use subs qw( HtmlUsage HtmlCmInfo );

# ***************************************************************
#
#  Subroutine HtmlUsage
#
#  Print script help message
#
# ***************************************************************

sub HtmlUsage {
   printf( STDOUT "Usage:  require 'printhtml.pl'\n" );
   return;
}

# ***************************************************************
#
#  Subroutine HtmlCmInfo
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub HtmlCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}

# ***************************************************************
#
#  Global Data Intialization
#
# ***************************************************************

$IndexEntries{lbAA} = "";
$IndexEntries{lbAB} = "NAME";
$IndexEntries{lbAC} = "SYNOPSIS";
$IndexEntries{lbAD} = "DESCRIPTION";
$IndexEntries{lbAE} = "EXAMPLE";
$IndexEntries{lbAF} = "RETURN VALUES";
$IndexEntries{lbAG} = "BUGS";
$IndexEntries{lbAH} = "CREDITS";
$IndexEntries{lbAI} = "FUNCTIONS";
$IndexEntries{lbAJ} = "SEE ALSO";
$IndexEntries{lbAK} = "AUTHOR";

$PageWidth          = "8in";
$ParamTableWidth    = "7in";
$ParamLineLen       = 60;

# ***************************************************************
#
#  Subroutine PrintHtmlMan
#
#  Description:  Print a man page for a function from the information
#                obtained from the source code and the developer's
#                inputs.  It does a bunch of HTML formatting, using
#                bold and italics to highlight function names, data types
#                and section headings.  In some cases it converts the
#                previously completed troff formatting to HTML, such as the
#                bold and italic encoding.
#
# ***************************************************************

sub PrintHtmlMan {
   my $fcts = $_[0];

   my $hdrname;
   my $func;
   my $fname;
   my $inclkey;
   my $incl;
   my @lcl_incl_list;
   my @sys_incl_list;
   my $timebuf;
   my $manpage;
   my $shrt_hdr;
   my $sec;
   my $min;
   my $hour;
   my $mday;
   my $mon;
   my $year;
   my $wday;
   my $yday;
   my $isdst;
   my $hdrlen;
   my $fileext = "html";
   my $tsummary;
   my @index_list = ();

   $printman_cnt++;

   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
   $year += 1900;
   $mon += 1;
   $sec = 0; $min = 0; $hour = 0; $wday = 0; $yday = 0; $isdst = $False;
   $timebuf = $year . "/" . $mon . "/" . $mday;

   if( %manrc_desc ) {
      if( $manrc_desc{header}{manual} ne "" ) {
         $ManHeading = $manrc_desc{header}{manual};
      }
      if( $manrc_desc{header}{section} > 0 ) {
         $ManSection = $manrc_desc{header}{section};
      }
      if( $manrc_desc{header}{source} ne "" ) {
         $ManSource = $manrc_desc{header}{source};
      }
   }

#  Open a man page file, print headers

   $manpage = $fcts . "." . $fileext;
   $manpage = $HtmlDir . "/" . $manpage;
   $func = $FuncList{$fcts};
   $hdrname = $fcts;
   $hdrlen = int( ( 54 - length( $ManHeading ) ) / 2 );
   if( length( $hdrname ) > $hdrlen ) {
      ($shrt_hdr) = ($hdrname =~ /(\w{1,$hdrlen}).+/ );
      $hdrname = $shrt_hdr . "..";
   }

   open( MANP, ">$manpage" ) or die "Couldn't open man page file $manpage";

   printf( MANP "<!doctype html public \"-//w3c//dtd html 4.0 //en\">\n" );
   printf( MANP "<META http-equiv = \"Content-type\"" );
   printf( MANP "content = \"text/html\">\n\n" );
   printf( MANP "<HTML>\n\n" );

   printf( MANP "<HEAD>\n" );
   printf( MANP "   <TITLE>Manpage of %s</TITLE>\n", $fcts );
   printf( MANP "   <link REL=stylesheet TYPE=\"text/css\" " );
   printf( MANP "HREF=\"manpage.css\">\n\n" );

   printf( MANP "</HEAD>\n" );
   printf( MANP "<BODY>\n" );

   printf( MANP "<H1>%s</H1>\n", $fcts );
   printf( MANP "Section: %s (%d)\n",
                $ManHeading, $ManSection );
   printf( MANP "<BR>\n" );
   printf( MANP "Updated: %s\n", $timebuf );
   printf( MANP "<BR>\n" );

   printf( MANP "<A HREF=\"#index\">Index</A>\n" );

   printf( MANP "<BR><BR>\n" );
   printf( MANP "<HR>\n" );

   printf( MANP "<A NAME=\"lbAB\">&nbsp;</A>\n" );
   printf( MANP "<H2>NAME</H2>\n\n" );
   push( @index_list, "lbAB" );

#  Have to remove any HTML special chars like '<' and '>'; not dealing any
#  others yet.  Maybe want an HTML transform function that runs all text
#  through such a filter.

   $tsummary = $func->{summary};

   $tsummary =~ s/</&lt;/;
   $tsummary =~ s/>/&gt;/;

#  Summary line; it's put into a table to format those cases where it
#  doesn't all fit on one line.

   printf( MANP "<TABLE class = funcblock>\n" );
   printf( MANP "   <TR>\n" );
   printf( MANP "      <TH class = tagb>%s&nbsp;-</TH>\n", $fcts );
   printf( MANP "      <TD>%s</TD>\n", $tsummary );
   printf( MANP "   </TR>\n" );
   printf( MANP "</TABLE>\n" );
   printf( MANP "<A NAME=\"lbAC\">&nbsp;</A>\n" );
   push( @index_list, "lbAC" );

#  Print the SYNOPSIS section header.  List the include files needed and the
#  function prototype(s).

   @sys_incl_list = @{$func->{sysincl}};
   @lcl_incl_list = @{$func->{lclincl}};
   if( $func->{blktype} eq "SUBR"  or
          $#sys_incl_list >= 0   or
          $#lcl_incl_list >= 0 )    {
      printf( MANP "<H2>SYNOPSIS</H2>\n" );
   }
   for $incl (@sys_incl_list) {
       $incl =~ s/</&lt;/;
       $incl =~ s/>/&gt;/;
       printf( MANP "<B>#include %s</B>\n", $incl );
       printf( MANP "<BR>\n" );
   }
   for $incl (@lcl_incl_list) {
       ($inclkey, $fname) = ( $incl =~ /(.*)\s*\"(.*)\"/ );
       printf( MANP "<B>#include %s &#8220;%s&#8221;</B>\n",
                    $inclkey, $fname );
       printf( MANP "<BR>\n" );
   }

   if( $func->{blktype} eq "SUBR" ) {
      PrintProto( $func, $fcts );
      PrintHtmlBody( $func, $fcts, \@index_list );
   }
   PrintIndex( \@index_list );


   printf( MANP "</BODY>\n" );
   close( MANP );

   $PageIndexList{$fcts} = 1;

   return;
}

# ***************************************************************
#
#  Subroutine PrintProto
#
#  Description:  Print the function prototype.  It needs bold for the
#                name and types, and italic for the parameter names.  If
#                it's a long line, it should be broken up and indented
#                nicely, but if it all fits on one line, then it's best
#                left on one line.
#                The formatting has to handle three cases:
#                   1)  Everything (storage class, return type, function name
#                       and all parameters fit on one line, e.g.,
#                          void foo ( int n, char s );
#                   2)  Everything doesn't fit on one line, but the longest
#                       combination of function definition and parameters
#                       fits across the page, so the parameters line up to
#                       the right of the function definition, e.g.,
#                          void moderateName (
#                                              parmeterType *parmname,
#                                              int n );
#                   3)  The function definition plus longest parameter
#                       is longer than the page width so that the parameters
#                       have to be listed below the function def, e.g.,
#                          void someReallyLongFunctionName (
#                                 longParameterTypeName *bigWordyParameterName,
#                                 int                    n );
#
# ***************************************************************

sub PrintProto {
   my $func = shift;
   my $fcts = shift;

   my $stgclass;
   my $typename;
   my $totlen;
   my $funclen;
   my @prmlist;
   my $longparm = 0;
   my $parmlen;
   my $multiln;
   my $indent;
   my $idx;
   my $preindent = $False;
   my $fspace    = "";

   printf( MANP "<BR>\n" );

   $stgclass = $func->{stg};
   $typename = $func->{typ};
   $totlen = length( $stgclass ) + 1;
   $totlen += length( $typename ) + 1;
   $totlen += length( $fcts ) + 1;
   $funclen = $totlen;
   @prmlist = @{ $func->{params} };

   for $idx (0 .. $#prmlist ) {
      $parmlen  = length( $prmlist[$idx]->{ptype} ) + 1;
      $parmlen += length( $prmlist[$idx]->{pname} ) + 1;
      $totlen  += $parmlen;
      if( $parmlen > $longparm ) { $longparm = $parmlen; }
   }
   if( $totlen > 55 ) {
      $multiln = $True;

      if( ($funclen + $longparm) > 55 ) {
         $preindent = $True;
         if( $longparm <= 58 ) { $indent = (56 - $longparm); }
         else { $indent = 3; }
      }
   } else { $multiln = $False; }

#  For the multi-line version, put it into table headers and rows.
#  For the single-line case, just use bold/italic, or use the SPAN classes.

   if( $multiln eq $True ) {
      if( ($typename =~ /.*\ \*$/) != 1 ) { $fspace = "&nbsp;"; }
      else { $fspace = ""; }
      if( $preindent eq $True ) {
         printf( MANP "<DIV>\n" );
         if( $stgclass ne "" ) {
            printf( MANP "   <SPAN class = funcname>%s&nbsp;</SPAN>\n",
                         $stgclass );
         }
         printf( MANP "   <SPAN class = return>%s%s</SPAN>\n",
                      $typename, $fspace );
         printf( MANP "   <SPAN class = funcname>%s(</SPAN>\n", $fcts );
         printf( MANP "</DIV>\n" );
         $preindent = $False;
         printf( MANP "<TABLE class = funcblock STYLE = width:%s>\n",
                      $ParamTableWidth );
         printf( MANP "   <TR class = funcline>\n" );
         printf( MANP "      <TH class = funcdef STYLE = width:%d%%>",
                      ($indent/$ParamLineLen)*100 );
         printf( MANP "&nbsp;</TH>\n" );
      } else {
         printf( MANP "<TABLE class = funcblock>\n" );
         printf( MANP "   <TR class = funcline>\n" );
         printf( MANP "      <TH class = funcdef>\n" );
   
         if( $stgclass ne "" ) {
            printf( MANP "         <SPAN class = funcname>%s</SPAN>\n",
                         $stgclass );
         }
         printf( MANP "         <SPAN class = return>%s%s</SPAN>\n",
                      $typename, $fspace );
         printf( MANP "         <SPAN class = funcname>%s(&nbsp;</SPAN>\n",
                 $fcts  );
         printf( MANP "      </TH>\n" );
      }

#  Print the first parameter in a <TD> on the header row.

      $idx = 0;

      if( ($prmlist[$idx]->{ptype} =~ /.*\ \*$/) != 1) { $fspace = "&nbsp;"; }
      else { $fspace = ""; }

      printf( MANP "      <TD class = parms>\n" );
      printf( MANP "         <SPAN class = funcname>%s%s",
                   $prmlist[$idx]->{ptype}, $fspace );
      printf( MANP "</SPAN>\n" );
      if( $prmlist[$idx]->{pname} ne "" ) {
         printf( MANP "         <SPAN class = parmname>%s</SPAN>\n",
                      $prmlist[$idx]->{pname} );
      }
      if( $#prmlist > 0 ) {
         printf( MANP "         <SPAN class = funcname>,</SPAN>\n" );
      } else {
         printf( MANP "         <SPAN class = funcname>&nbsp;);</SPAN>\n" );
      }
      printf( MANP "      </TD>\n" );
      printf( MANP "   </TR>\n" );

#  Print the rest of the function parameters, one per row
      for $idx (1 .. $#prmlist) {
         printf( MANP "   <TR class = funcline><TH></TH>\n" );
         printf( MANP "      <TD class = parms>\n" );

         if( ($prmlist[$idx]->{ptype} =~ /.*\ \*$/) != 1) {
            $fspace = "&nbsp;";
         } else { $fspace = ""; }

         printf( MANP "         <SPAN class = funcname>%s%s",
                      $prmlist[$idx]->{ptype}, $fspace );
         printf( MANP "</SPAN>\n" );

         if( $prmlist[$idx]->{pname} ne "" ) {
            printf( MANP "         <SPAN class = parmname>%s</SPAN>\n",
                         $prmlist[$idx]->{pname} );
         }

         if( $idx eq 0 and $preindent eq $False ) {
            printf( MANP "         <SPAN class = funcname>,</SPAN>\n" );
            $preindent = $False;
         }
         elsif( $idx < $#prmlist ) {
            printf( MANP "         <SPAN class = funcname>,</SPAN>\n" );
         } else {
            printf( MANP "         <SPAN class = funcname>&nbsp;);</SPAN>\n" );
         }
         printf( MANP "      </TD>\n" );
         printf( MANP "   </TR>\n" );
      }
      printf( MANP "</TABLE>\n" );
   } else {
      printf( MANP "<DIV>\n" );
      if( $stgclass ne "" ) {
         printf( MANP "   <SPAN class = funcname>%s&nbsp;</SPAN>\n",
                      $stgclass );
      }
      if( ($typename =~ /.*\ \*$/) != 1 ) { $fspace = "&nbsp;"; }
      else { $fspace = ""; }
      printf( MANP "   <SPAN class = return>%s%s</SPAN>\n",
                   $typename, $fspace );

      if( $#prmlist < 0 ) {
         printf( MANP "   <SPAN class = funcname>%s(&nbsp;);</SPAN>\n", $fcts );
      } else {
         printf( MANP "   <SPAN class = funcname>%s(&nbsp;</SPAN>\n", $fcts );
      }
      for $idx (0 .. $#prmlist) {
         if( ($prmlist[$idx]->{ptype} =~ /.*\ \*$/) != 1 ) {
            $fspace = "&nbsp;";
         } else { $fspace = ""; }
         printf( MANP "   <SPAN class = funcname>%s</SPAN>\n",
                      $prmlist[$idx]->{ptype} );
         if( $prmlist[$idx]->{pname} ne "" ) {
            printf( MANP "   <SPAN class = parmname>%s%s</SPAN>\n",
                         $fspace, $prmlist[$idx]->{pname} );
         }
         if( $idx < $#prmlist ) {
            printf( MANP "   <SPAN class = funcname>,&nbsp;</SPAN>\n" );
         } else {
            printf( MANP "   <SPAN class = funcname>&nbsp;);</SPAN>\n" );
            printf( MANP "</DIV>\n" );
         }
      }
   }
}

# ***************************************************************
#
#  Subroutine PrintHtmlBody
#
#  Description:  Prints the description, return values and bugs, if any
#                are specified.  The PrintHtmlDesc subroutine is called
#                to look at each word to see if it is an 'interesting'
#                word that needs highlighting. 
#
# ***************************************************************

sub PrintHtmlBody {
   my $func = shift;
   my $fcts = shift;
   my $index_list = shift;

   if( @{ $func->{description} } ) {
      printf( MANP "<A NAME=\"lbAD\">&nbsp;</A>\n" );
      printf( MANP "<H2>DESCRIPTION</H2>\n\n" );
      PrintHtmlDesc( $func->{description}, $fcts );
      push( @{$index_list}, "lbAD" );
   }
   if( @{ $func->{example} } ) {
      printf( MANP "<A NAME=\"lbAE\">&nbsp;</A>\n" );
      printf( MANP "<H2>EXAMPLE</H2>\n\n" );
      PrintHtmlDesc( $func->{example}, $fcts );
      push( @{$index_list}, "lbAE" );
   }
   if( @{ $func->{ret_desc} } ) {
      printf( MANP "<A NAME=\"lbAF\">&nbsp;</A>\n" );
      printf( MANP "<H2>RETURN VALUES</H2>\n\n" );
      PrintHtmlDesc( $func->{ret_desc}, $fcts );
      push( @{$index_list}, "lbAF" );
   }
   if( @{ $func->{bug_desc} } ) {
      printf( MANP "<A NAME=\"lbAG\">&nbsp;</A>\n" );
      printf( MANP "<H2>BUGS</H2>\n\n" );
      PrintHtmlDesc( $func->{bug_desc}, $fcts );
      push( @{$index_list}, "lbAG" );
   }
   if( @{ $func->{credits} } ) {
      printf( MANP "<A NAME=\"lbAH\">&nbsp;</A>\n" );
      printf( MANP "<H2>CREDITS</H2>\n\n" );
      PrintHtmlDesc( $func->{credits}, $fcts );
      push( @{$index_list}, "lbAH" );
   }
   if( @{ $func->{funclist} } ) {
      printf( MANP "<A NAME=\"lbAI\">&nbsp;</A>\n" );
      printf( MANP "<H2>FUNCTIONS</H2>\n\n" );
      PrintHtmlDesc( $func->{funclist}, $fcts );
      push( @{$index_list}, "lbAI" );
   }
   if( @{ $func->{refs} } ) {
      printf( MANP "<A NAME=\"lbAJ\">&nbsp;</A>\n" );
      printf( MANP "<H2>SEE ALSO</H2>\n\n" );
      FormatRefs( $func );
      PrintHtmlDesc( $func->{refs}, $fcts );
      push( @{$index_list}, "lbAJ" );
   }
   if( @{ $func->{author} } ) {
      printf( MANP "<A NAME=\"lbAK\">&nbsp;</A>\n" );
      printf( MANP "<H2>AUTHOR</H2>\n\n" );
      PrintHtmlDesc( $func->{author}, $fcts );
   } elsif( %manrc_desc ) {
      printf( MANP "<A NAME=\"lbAK\">&nbsp;</A>\n" );
      printf( MANP "<H2>AUTHOR</H2>\n\n" );
      PrintHtmlDesc( $manrc_desc{author}, $fcts );
   } else {
      my $passwdent;
      my $username = $ENV{USER};
      my @pwstuff = getpwnam($username);
      my $author_name = $pwstuff[6];
 
      printf( MANP "<A NAME=\"lbAK\">&nbsp;</A>\n" );
      printf( MANP "<H2>AUTHOR</H2>\n\n" );
      printf( MANP "<P>%s</P>\n", $author_name );
   }
   push( @{$index_list}, "lbAK" );

}

# ***************************************************************
#
#  Subroutine PrintHtmlDesc
#
#  Description:  Processes a block of text for a 'man' page section.
#                Uses previously identified paragraph attributes to set
#                indents and line spacing.  For each word, sets bold and
#                italic highlights for function, data type and parameter
#                names.
#
# ***************************************************************

sub PrintHtmlDesc {
   my @fmtdesc    = @{$_[0]};
   my $fctname    = $_[1];

   my $paragraph;       # Hash reference to the paragraph.
   my $previndent = 0;
   my @words;
   my $word;
   my $filename;
   my $punc;
   my $fmtline;
   my $linelen    = 0;
   my @decomp;
   my $fmt;
   my $linesep    = "";
   my $pstyle     = "";
   my $istable    = $False;
   my $htmlfmt    = "";
   my $linespace  = 0;
   my $paracount  = 0;
   my $count;
   my $para_space = 0;
   my $bk_after_fmt = $False;

   $printdesc_cnt++;

#  Now process each paragraph, defined as a hash including text, indent,
#  line spacing, length and paragraph type.

   Reindent( \@fmtdesc );

   $previndent = $fmtdesc[0]->{indent};
   foreach $paragraph (@fmtdesc) {
      if( $paragraph->{indent} < $previndent ) {
         $previndent = $paragraph->{indent};
      }
   }
   if( $previndent < 0 ) {  # Adjust all indents so the lowest is 0
      my $indent_adjust = abs( $previndent );
      foreach $paragraph (@fmtdesc) {
         $paragraph->{indent} += $indent_adjust;
      }
   }

   $previndent = $fmtdesc[0]->{indent};

   FMTLINE: foreach $paragraph (@fmtdesc) {

      $pstyle = " STYLE = margin-left:" .
                   ($paragraph->{indent}/2) . "em";

      if( $istable eq $True and $paragraph->{fmttype} ne $Label ) {
         printf( MANP "</TABLE>\n" );
         $istable = $False;
      }
      for $linespace (1..$para_space) {
         if( $istable eq $True ) {
            printf( MANP "   <TR><TD STYLE = height:0.8em></TD></TR>\n" );
         } else { printf( MANP "<BR>\n" ); }
      }

      if( $paragraph->{fmttype} eq $Normal ) {
         $htmlfmt = $linesep . "<DIV" . $pstyle . ">";
         $linelen = length( $htmlfmt );
         printf( MANP $htmlfmt );
         $linesep = "";
         $pstyle = "";
      } elsif( $paragraph->{fmttype} eq $Label ) {
  
#    Labeled paragraphs are put in a table.  Need to see if an element is
#    the first label and write table header if it is.  Then write each table
#    row with label, separator and text.  After the last table element,
#    write a table footer.

         if( $istable eq $False ) {
            printf( MANP "<TABLE class = labellist%s>\n", $pstyle );
            $istable = $True;
         }
         printf( MANP "   <TR>\n" );
         my $tagword = $paragraph->{label};
         if( defined( $keylist{$tagword} ) ) { $fmt = $keylist{$tagword}; }
         elsif( defined( $namelist{$fctname}{$tagword} ) ) {
            $fmt = $namelist{$fctname}{$tagword};
         } else { $fmt = $paragraph->{tagfmt}; }
         printf( MANP
                 "      <TH class=tag%s><SPAN class=tagword>%s</SPAN></TH>\n",
                 lc($fmt), $paragraph->{label} );
         $htmlfmt = "      <TD class = labelbody>";
         $linelen = length( $htmlfmt );
         printf( MANP $htmlfmt );
         $linesep = "";
         $linespace = $paragraph->{spaces};
      } elsif( $paragraph->{fmttype} eq $Code ) {
         my $code_indent = $paragraph->{indent}/2;
         if( $code_indent < 0 ) { $code_indent = 0; }
         printf( MANP "<PRE class = ccode STYLE = margin-left:%dem>\n",
                      $code_indent );
         printf( MANP "%s", $paragraph->{text} );
         printf( MANP "</PRE>\n" );
         $linesep = "";
         next FMTLINE;
      }

      @words = split( ' ', $paragraph->{text} );
   #  Here it gets messy.  First, any trailing punctuation has to come off,
   #  because we don't want to highlight that.  Then the remainder needs to
   #  be separated into the alphanumeric parts so each word can be checked to
   #  see if it should be highlighted.  Associated C syntax, like '()', '*'
   #  needs to be highlighted with the word, but other punctuation shouldn't.
   #  Slashes need to separate words so we can deal with things like 'and/or',
   #  but path names need to be treated as an entity, and they should be
   #  highlighted also.

      WORD: for $word ( @words ) {
   #  Substitute '<', '>' characters with the HTML codes &lt, &gt
         $word =~ s/</&lt;/g;
         $word =~ s/>/&gt;/g;

         if( $word =~ /^\.{0,2}\/\w+/ ) {
            ($filename, $punc) = ($word =~ /^(\.{0,2}\/[\/\w\.]+)(\W*)$/ );
            if( defined( $punc ) and $punc ne "" ) {
               $fmtline = ".IR $filename $punc";
            } else { $fmtline = ".I $filename"; }
         } else {
            @decomp = split( '\b', $word );

            if( $#decomp > 0 ) {
               $fmtline = FormatWord( \@decomp, $fctname, "HTML" );
            }
            else { $fmtline = ""; }
         }

   #  Convert the man-page style formatting definition to HTML formatting
   #  using SPAN definitions in the css

   #  1)  Print formatted text, if it's designated as bold or italic
         if( $fmtline ne "" ) {
            my $fdidx    = 0;
            my @fmtlist  = split( ' ', $fmtline );
            my $fmtdesig = shift( @fmtlist );
            my @fdlist   = split( '', $fmtdesig );
            my $punctest = $False;
            my $numitems;
            shift( @fdlist );

            printf( MANP "%s", $linesep );
            if( $istable eq $True and $linesep eq "\n" ) {
               printf( MANP "      " );
            }
            $numitems = $#fmtlist;
            for $idx (0..$numitems) {
               printf( MANP "<SPAN class = medium%s>",
                            lc( $fdlist[$fdidx] ) );
               $punctest = IsPunctuation( $fmtlist[0] );
               if( $punctest eq $True ) { $linesep = "\n"; }
               else { $linesep = ""; }
               printf( MANP "%s</SPAN>%s", shift( @fmtlist ), $linesep );
               $fdidx++;
               if( $fdidx > $#fdlist ) { $fdidx = 0; }
            }

            if( $linesep eq "" ) { $bk_after_fmt = $True; }
            $linesep = "";
            $linelen = 0;
            next WORD;
         }
   #  2)  If it's a defined name or keyword, determine it's formatting and
   #      print it
         if( $bk_after_fmt eq $True ) {
            printf( MANP "\n" );
            $linesep = "";
            $bk_after_fmt = $False;
         }
         if( defined( $keylist{$word} ) ) { $fmt = $keylist{$word}; }
         elsif( defined( $namelist{$fctname}{$word} ) ) {
            $fmt = $namelist{$fctname}{$word}; }
         else { $fmt = ""; }
         if( $fmt ne "" ) { 
            if( $istable eq $True and $linesep eq "\n" ) {
               $linesep = $linesep . "      ";
            }
            printf( MANP "%s<SPAN class = medium%s>%s</SPAN>\n",
               $linesep, lc($fmt), $word );
            $linelen = 0;
            $linesep = "";
            next WORD;
         } else {
   #  3)  Print the word in plain text.  If the line is filled up, print
   #      a newline and start fresh.  Otherwise, print a space before the
   #      word unless it's the start of a line.
            my $wsep;
            if( $linelen == 0 ) {
               if( $istable eq $True ) { $wsep = "      "; }
               else { $wsep = ""; }
            } else { $wsep = " "; }
            $linelen += length( $word ) + length( $wsep );
            if( $linelen > 70 ) {
               printf( MANP "\n" );
               if( $istable eq $True ) { $wsep = "      "; }
               else { $wsep = ""; }
               $linelen = length( $word ) + length( $wsep );
            }
            printf( MANP "%s%s", $wsep, $word );
            $linesep = "\n";
         }
      }

      if( $istable eq $True ) {
         if( $linesep eq "\n" ) { $wsep = ""; }
         else { $wsep = "      "; }
         printf( MANP "%s</TD>\n   </TR>\n", $wsep );
      } elsif( $paragraph->{fmttype} eq $Normal ) {
         printf( MANP "</DIV>\n" );
      }
      $para_space = $paragraph->{spaces};
      $linesep = "";
      $linelen = 0;
      $paracount++;
   }
   if( $istable eq $True ) {
      printf( MANP "</TABLE>\n" );
      $linesep = "";
   }
   printf( MANP "$linesep" );
   return;
}

# ***************************************************************
#
#  Subroutine PrintIndex
#
#  Description:  Print the index section for each page segment used.
#                The list was collected while printing the man page
#                and passed as an array.  Uses the pre-initialized hash
#                table of 'lbAX' codes and name strings.
#
# ***************************************************************

sub PrintIndex {
   my $index_list = shift;

   my $idxitem;

   printf( MANP "<HR>\n" );
   printf( MANP "<A NAME=\"index\">&nbsp;</A><H2>Index</H2>\n" );
   printf( MANP "<DL>\n" );

   for $idxitem (@{$index_list}) {
      printf( MANP "<DT><A HREF=\"#%s\">%s</A></DT>\n",
                   $idxitem, $IndexEntries{$idxitem} );
   }
   printf( MANP "</DL>\n" );

   return;
}

# ***************************************************************
#
#  Subroutine PrintHtmlIndex
#
#  Description:  Print an HTML index page that lists all the functions and
#                a link to the man page.
#
#                For now, this lists every function, whether or not the man
#                page has been generated, whether or not the file exists.
#                It should check to see if the function is in the index
#                already, add it in alphabetic order if not found, and
#                provide options for nameing the file and location.
#
# ***************************************************************

sub PrintHtmlIndex {
   my $func;

   ReadHtmlIndex();

   if( defined( $IndexFile ) and $IndexFile ne "") {
      open( INDEX, ">$IndexFile" )
            or die "Couldn't open HTML Index file $IndexFile";
   } else {
      PrintStat( "HTML Index name '$IndexFile' not valid", 'warning' );
      return;
   }

   printf( INDEX "<HTML>\n" );
   printf( INDEX "<HEAD>\n" );
   printf( INDEX "   <TITLE>Man Page Index</TITLE>\n" );
   printf( INDEX "   <LINK REL = stylesheet TYPE = \"text.css\" " );
   printf( INDEX "HREF = \"manpage.css\"\n" );
   printf( INDEX "</HEAD>\n" );
   printf( INDEX "<BODY>\n" );
   printf( INDEX "<H1>Index of Man Pages</H1>\n" );
   printf( INDEX "Section: Srcdoc Programmer's Manual\n" );
   printf( INDEX "<BR>\n" );
   printf( INDEX "Updated: 2001/11/16\n" );
   printf( INDEX "<HR>\n" );

   foreach $func (sort(keys %PageIndexList) ) {
      printf( INDEX "<A HREF = %s.html>%s</A>\n", $func, $func );
      printf( INDEX "<BR>\n" );
   }
   close( INDEX );

   %PageIndexList = ();

   return;
}

# ***************************************************************
#
#  Subroutine IsPunctuation
#
#  Description:  Utility function to check if a word contains only punctuation
#                characters.  Returns TRUE if it does, FALSE otherwise.
#                Punctuation includes , . ; : & * in the current
#                implementation.
#
# ***************************************************************

sub IsPunctuation {
   my $str = shift;
   my $result = $False;

   if( $str =~ /^[\,\.\;\:\&\*]+$/ ) {
      $result = $True;
   }

   return $result;
}

# ***************************************************************
#
#  Subroutine ReadHtmlIndex
#
#  Description:  Reads the HTML index file and stores the page names in
#                a hash table.  HTML entries are in the form of
#                   <A HREF = filename>funcname</A>
#                This function reads the funcname part.
#
# ***************************************************************

sub ReadHtmlIndex {

   my $iline;
   my $filename;

   if( defined( $IndexFile ) and $IndexFile ne "" and -e $IndexFile ) {
      open( INDEXFILE, "<$IndexFile" )
            or die "Couldn't open HTML Index file $IndexFile";
   } else { return; }

   while( <INDEXFILE> ) {
      $iline = $_;
      chomp $iline;
      
      if( $iline =~ /<A.*>/ ) {
         ($filename) = ($_ =~ /<A.*>(.*)<\/A>/ );
         if( -e "$HtmlDir/$filename.html" ) {
            $PageIndexList{$filename} = 1;
         }
      }
   }
   close( INDEXFILE );
   return;
}

1;
