#if( defined( SOLARIS ) )
static char   ident[]   = "%Z%%M%  control version %I%  %G% %U%";
#elif( defined( LINUX ) )
static char   ident[]   = "srcdoc.c,v control version 1.7 2002/01/05 02:57:53";
#endif
static char   cpyrght[] = "Copyright (c) 2000, Gordon B. Aaseng";
static char   usage[]   = "Main routines for generating man pages from source code";

/*
 * srcextract - program for extracting documentation data from C source files
 *
 * copyright (c) 2000, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */


/*
 *
 * FILE NAME
 *
 * srcdoc.c
 *
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 *
 * FILE DESCRIPTION
 *
 * This module contains the main routine for extracting documentation
 * information from C source code.  It parses the source code with a
 * lex/yacc parser, then organizes the data by function and writes an
 * intermediate file which a perl program uses to generate the man page.
 * This module is generally called by the perl program or other program,
 * such as 'make'.
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * mkdoc.c,v
 * Revision 1.1  2000/06/24 03:57:18  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */


#define   _POSIX_SOURCE     1       /* posix.1             */
#ifndef   _POSIX_C_SOURCE
#define   _POSIX_C_SOURCE   199309  /* posix.1 and posix.4 */
#endif

/* ************************************************************************* */
/*                                                                           */
/* Included Files                                                            */
/*                                                                           */
/* ************************************************************************* */

/*
 * os specific header files.
 */
#include <unistd.h>

/*
 * standard c (ANSI) header files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>


/*
 * posix header files
 */
/*  (None.)  */

/*
 * project specific header files.
 */

#include "srcdocLex.h"
#include "srcdocActions.h"

/* ************************************************************************* */
/*                                                                           */
/* External References                                                       */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Scalar Data Types                                                         */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Data Structures                                                           */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Public Global Variables                                                   */
/*                                                                           */
/* ************************************************************************* */

int          funcStart = 0;
int          topOfFile = 0;
char        *parseFileName;
int          gblVbs;
funcDefType *gblFuncList;

/* ************************************************************************* */
/*                                                                           */
/* Private (static) Global Variables                                         */
/*                                                                           */
/* ************************************************************************* */

static int stgMode = 0;

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (Locally Used Functions)                                 */
/*                                                                           */
/* ************************************************************************* */


/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

static void  dumpFunc( char * );
static void  dumpParm( FILE *, parmListType *, int, parmClassType );
static void  showFuncList( FILE * );
static void  doHelp( void );

/* ---------------------------------------------------------------------- */

/* ************************************************************************* *
 *  Private Function doHelp
 *
 *  DESCRIPTION
 *     Prints help information for all valid command line switches.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void doHelp( void )
{
   printf( "srcextract [-h -v -z -m] [-f in_file] [-o out_file]\n" );
   printf( "Extract man page information from a C source file\n" );
   printf( "including function calls, include files and comments,\n" );
   printf( "and format a man page outline.\n" );
   printf( "This should generally be used to pipe the pre-compiler\n" );
   printf( "output into 'srcextract', specifying the output file name\n" );
   printf( "on the command line, all running from a Makefile.\n" );
   printf( "   where\n" );
   printf( "      -h : print this help message\n" );
   printf( "      -v : verbose, to print a bunch of run messages\n" );
   printf( "      -z : print module version data\n" );
   printf( "      -m : write storage class of functions to man page\n" );
   printf( "      -f <in_file> : file to be processed; should be\n" );
   printf( "           the saved output from the C pre-processor.\n" );
   printf( "           If not specified, uses standard input "
                       "(from input pipe).\n" );
   printf( "      -o <out_file>: output file name for the generated "
                       "man page.\n" );
   printf( "           If not specified, tries to generate it from "
                       "the input file name\n" );
   return;
}

/* ************************************************************************* *
 *  Private Function showFuncList
 *
 *  DESCRIPTION
 *     Prints the functions that were found in processing to standard
 *     output.  Used for debugging.
 *
 *  RETURN VALUES
 *     None (void).
 * ************************************************************************* */

static void showFuncList( FILE *ffp )
{
   funcDefType      *funcPtr;
   int               idx = 0;

   if( ffp == (FILE *)NULL )
      return;

   funcPtr = gblFuncList;
   while( funcPtr )
   {
      fprintf( ffp, "Function %d\n", idx++ );
      printValList( ffp, funcPtr->spec->storage );
      printValList( ffp, funcPtr->spec->type );
      printValList( ffp, funcPtr->func->decl );
      printParmList( ffp, funcPtr->func->parms );
      funcPtr = funcPtr->funcNext;
   }
   return;
}

/* ************************************************************************* *
 *  Private Function dumpFunc
 *
 *  DESCRIPTION
 *     Dump the function specification to a temporary file for a
 *     perl script to subsequently process.
 *
 *  RETURN VALUES
 *     None (void).
 * ************************************************************************* */

static void dumpFunc( char *dumpFile )
{
   FILE             *fp;
   funcDefType      *funcPtr;
   tokListType      *tokPtr;
   parmListType     *parmPtr;
   int               plvl;
   int               parmCnt;
   int               pidx;
   int               funcPtrLvl;


   fp = fopen( dumpFile, "w" );

/*
 *  Each function from the parser gets an entry in the function synopsis
 *  section.
*/
   funcPtr = gblFuncList;
   while( funcPtr )
   {
      tokPtr = funcPtr->func->decl;
      fprintf( fp, "FCT: " );
      while( tokPtr )
      {
         funcPtrLvl = tokPtr->ptrLvl;
         fprintf( fp, "%s\n", tokPtr->tok );
         tokPtr = tokPtr->tokNext;
      }
      if( funcPtr->spec != (typeSpecType *)NULL )
         tokPtr = funcPtr->spec->storage;
      else
         tokPtr = (tokListType *)NULL;
      fprintf( fp, "STG:" );
      while( tokPtr )
      {
         for( plvl = 0; plvl < tokPtr->ptrLvl; plvl++ )
         {
            fprintf( fp, "*" );
         }
         fprintf( fp, " %s ", tokPtr->tok );
         tokPtr = tokPtr->tokNext;
      }
      fprintf( fp, "\n" );
      if( funcPtr->spec != (typeSpecType *)NULL )
         tokPtr = funcPtr->spec->type;
      else
         tokPtr = (tokListType *)NULL;
      fprintf( fp, "TYP: " );
      if( tokPtr == (tokListType *)NULL )
         fprintf( fp, "\n" );
      while( tokPtr )
      {
         for( plvl = 0; plvl < tokPtr->ptrLvl; plvl++ )
         {
            fprintf( fp, "*" );
         }
         fprintf( fp, "%s", tokPtr->tok );
         if( funcPtrLvl > 0 )
         {
            fprintf( fp, " " );
            for( plvl = 0; plvl < funcPtrLvl; plvl++ )
               fprintf( fp, "*" );
         }
         fprintf( fp, "\n" );
         tokPtr = tokPtr->tokNext;
      }

      parmPtr = funcPtr->func->parms;

      pidx = 0;
      parmCnt = 0;
      while( parmPtr )
      {
         if( ( pidx % 2 ) == 0 )
            fprintf( fp, "PRM: " );
         dumpParm( fp, parmPtr, pidx, SCALAR );
         parmPtr = parmPtr->parmNext;
         if( ( pidx++ % 2 ) != 0 && parmPtr )
            fprintf( fp, "\n" );
      }

      fprintf( fp, "\n" );
      fprintf( fp, "HDRCMT:\n" );
      if( funcPtr->hdrComment != (char *)NULL )
         fprintf( fp, "%s", funcPtr->hdrComment );
      fprintf( fp, "ENDC:\n" );
      fprintf( fp, "INTCMT:\n" );
      if( funcPtr->intComment != (char *)NULL )
         fprintf( fp, "%s", funcPtr->intComment );
      fprintf( fp, "ENDC:\n" );
      funcPtr = funcPtr->funcNext;
   }

   fclose( fp );
   return;
}

/* ************************************************************************* *
 *  Private Function dumpParm
 *
 *  DESCRIPTION
 *     Dump the parameter definition to the temporary file for the
 *     perl script to subsequently process.  For scalar parameters, the
 *     type and name are printed; for function pointers, the name is
 *     parenthesized and the function parameter types are added as a list.
 *
 *  RETURN VALUES
 *     None (void).
 * ************************************************************************* */

static void dumpParm( FILE          *fp,
                      parmListType  *parmPtr,
                      int            pidx,
                      parmClassType  class )
{
   int               plvl;
   tokListType      *tokPtr;
   parmListType     *funcParmPtr;

   plvl    = 0;

         if( parmPtr->parmClass == SCALAR )
         {
            tokPtr = parmPtr->parm.scalarParm;
            while( tokPtr )
            {
               for( plvl = 0; plvl < tokPtr->ptrLvl; plvl++ )
                  fprintf( fp, "*" );
               if( ( pidx % 2 ) == 0 )
               {
                  if( tokPtr->tok )
                     fprintf( fp, "%s ", tokPtr->tok );
               }
               else if( tokPtr->tok )
               {
                  if( class == SCALAR )
                      fprintf( fp, " NAM: " );
                  fprintf( fp, "%s", tokPtr->tok );
               }
               tokPtr = tokPtr->tokNext;
               if( tokPtr )
                  fprintf( fp, " " );
            }
         }
         else if( parmPtr->parmClass == FUNC )
         {
            tokPtr = parmPtr->parm.funcParm->decl;
            fprintf( fp, " FUNC: " );
            fprintf( fp, "(" );
            while( tokPtr )
            {
               for( plvl = 0; plvl < tokPtr->ptrLvl; plvl++ )
                  fprintf( fp, "*" );
               fprintf( fp, "%s", tokPtr->tok );
               tokPtr = tokPtr->tokNext;
               if( tokPtr )
                  fprintf( fp, " " );
            }
            fprintf( fp, ") " );
            funcParmPtr = parmPtr->parm.funcParm->parms;
            fprintf( fp, "( " );
            while( funcParmPtr )
            {
               dumpParm( fp, funcParmPtr, pidx, FUNC );
               funcParmPtr = funcParmPtr->parmNext;
               if( funcParmPtr )
                  fprintf( fp, ", " );
            }
            fprintf( fp, " )" );
/*
            if( funcParmPtr->parmClass == SCALAR )
            {
               funcTokPtr = funcParmPtr->parm.scalarParm;
               while( funcTokPtr )
               {
                  for( plvl = 0; plvl < funcTokPtr->ptrLvl; plvl++ )
                     fprintf( fp, "*" );
                  if( ( pidx % 2 ) == 0 )
                     fprintf( fp, "%s ", funcTokPtr->tok );
                  else
                     fprintf( fp, " NAM: %s", funcTokPtr->tok );
                  funcTokPtr = funcTokPtr->tokNext;
               }
               funcParmPtr = funcParmPtr->parmNext;
               if( ( pidx++ % 2 ) != 0 && funcParmPtr )
                  fprintf( fp, "\n" );
            }
*/
         }
}

/* ---------------------------------------------------------------------- */

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (External Interface Functions)                           */
/*                                                                           */
/* ************************************************************************* */

/**************************************************************************
 * Public Function Prototypes
 **************************************************************************/

char *srcdocMainCopyright( void );
char *srcdocMainId( void );
char *srcdocMainUsage( void );
char *stripID( char * );

/* ************************************************************************* *
 *  Public Function main
 *
 *  DESCRIPTION
 *     Processes command line arguments, invokes the source code
 *     parser and dumps function definitions and comments to an
 *     intermediate file.  This should be called with the name of the
 *     intermediate file; if it isn't, it uses the source code file
 *     name to generate the name.  If no source code file is given, it
 *     bails out.
 *
 *     A perl program reads the intermediate file to generate documentation.
 *
 *     Usage is
 *        srcdoc [-h -v -z -m -f <source_file> -o <dump_file>
 *
 *     Inputs
 *        argc - the number of command line arguments
 *        argv - the command line args
 *        environ - the process environment
 *
 *  RETURN VALUES
 *     int - program status.  Parser status is used.
 *
 * ************************************************************************* */

int main( int argc, char **argv, char **environ )
{
   char           *options = "hvzmf:o:";
   int             opt;
   char           **envPtr;
   char           *inFile   = (char *)NULL;
   char           *outFile  = (char *)NULL;
   char           *fileRoot = (char *)NULL;
   char            tmpFile[0x100];
   char           *tmpFilePtr;
   int             idx;
   int             parseStat = 0;

   while( ( opt = getopt( argc, argv, options ) ) >= 0 )
   {
      switch( opt )
      {
         case 'h' :
            doHelp();
            exit( parseStat );
            break;
         case 'v' :
            printf( "Verbose option\n" );
            gblVbs = 1;
            break;
         case 'z' :
            printf( "%s\n", srcdocMainId() );
            printf( "%s\n", srcdocMainCopyright() );
            printf( "%s\n", srcdocMainUsage() );
            exit( parseStat );
            break;
         case 'm' :
            stgMode = 1;
            break;
         case 'f' :
            inFile = optarg;
            break;
         case 'o' :
            outFile = optarg;
            break;
         default :
            printf( "Not an option -%c\n", (char)opt );
            break;
      }
   }

   if( gblVbs == 1)
   {
      envPtr = environ;
      printf( "srcdoc-I-Environment:" );
      while( *envPtr )
      {
         printf( "%s ", *envPtr );
         envPtr++;
      }
      printf( "\n" );
   }

   if( inFile != (char *)NULL )
   {
      srcextractin = fopen( inFile, "r" );
      parseFileName = inFile;
      strcpy( tmpFile, inFile );
      tmpFilePtr = tmpFile;
      fileRoot = strsep( &tmpFilePtr, "." );
   }
   else
   {
      parseFileName = "stdin";
      srcextractin = stdin;
   }
   if( outFile == (char *)NULL && fileRoot != (char *)NULL )
   {
      outFile = (char *)malloc( strlen( fileRoot ) + 5 );
      strcpy( outFile, fileRoot );
      strcat( outFile, ".dmp" );
   }
   else if( outFile != (char *)NULL && fileRoot == (char *)NULL )
   {
      strcpy( tmpFile, outFile );
      tmpFilePtr = tmpFile;
      fileRoot = strsep( &tmpFilePtr, "." );
   }
   else if( outFile == (char *)NULL && fileRoot == (char *)NULL )
   {
      printf( "srcdoc-W-fileNames - Output file not specified or deduced\n" );
      printf( "     from input; set to 'default.dmp'.\n" );
      fileRoot = "stdin";
      outFile  = "default.dmp";
   }

   if( gblVbs == 1 )
      printf( "Process input file %s, dump file %s\n",
              parseFileName, outFile );

   if( outFile == (char *)NULL )
   {
      printf( "srcdoc-E-outputFile - Output file not specified or deduced\n" );
      exit( parseStat );
   }

   if( srcextractin != NULL )
   {
      parseStat = srcextractparse();
      if( parseStat != 0 )
         printf( "Parser failed: %d (%08X)\n", parseStat, parseStat );

      fclose( srcextractin );
      dumpFunc( outFile );
      if( gblVbs == 1 )
         showFuncList( stdout );
   }

   if( gblVbs == 1 )
   {
      printf( "Program defined types: %d\n", numTypeDefs );
      for( idx = 0; idx < numTypeDefs; idx++ )
         printf( "Type %d: %s\n", idx, typeList[idx] );
   }

   exit( parseStat );
}

/* *************************************************************************
 *
 *  Public Functions srcdocMainCopyright, srcdocMainId,
 *               and srcdocMainUsage
 *
 *  DESCRIPTION
 *     Module Identification functions used to print file ID's with the
 *     command line '-z' switch.
 *
 *  RETURN VALUES
 *     Pointer to the specified strings
 *
 * ************************************************************************* */

char *srcdocMainCopyright( void )
{
   return( cpyrght );
}

char *srcdocMainId( void )
{
   return( stripID( ident ) );
}

char *srcdocMainUsage( void )
{
   return( usage );
}

/***************************************************************************
 * Public Function stripID
 *
 * DESCRIPTION
 *    Strips RCS keywords from the ID string for display
 *
 * RETURN VALUES
 *    Pointer to ID String
 *
 ***************************************************************************/

char *stripID( char *idstr )
{
   char   *idStrCopy;
   char   *idPtr;
   char   *cpyPtr;
   int     idLen;
   boolean keywd;

   idLen = strlen( idstr ); 
   idStrCopy = (char *)malloc( idLen + 1);
   if( idStrCopy != (char *)NULL )
      strcpy( idStrCopy, idstr );
   else
      return( idstr );
   idPtr = idstr;
   cpyPtr = idStrCopy;
   keywd = FALSE;
   while( *cpyPtr != '\0' )
   {
      if( ! keywd )
         if( *cpyPtr == '$' )
            keywd = TRUE;

      if( keywd )
      {
         if( *cpyPtr == ' ' )
         {
            keywd = FALSE;
            cpyPtr++;
         }
         else
            cpyPtr++;
      }
      else
      {
         if( *cpyPtr == ',' && *(cpyPtr+1) == 'v' )
            cpyPtr += 2;
         else
            *idPtr++ = *cpyPtr++;
      }
   }
   *idPtr = '\0';
   free( idStrCopy );
   return( idstr );
}

#ifdef UNDEF
------------------------ BEGIN ILLEGITIMATE STUFF HERE ------------------------
------------------------- END ILLEGITIMATE STUFF HERE -------------------------
#endif /* UNDEF */
