#! /usr/bin/perl -w

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'srcdoc.pl,v control version 1.16 2002/01/05 02:35:29'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "Script for making man page outlines from source code.";


#  srcdoc.pl - program for generating man pages from source code
#
#  Copyright (c) 2000, 2001, Gordon B. Aaseng
#                      Glendale, AZ USA

#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  

#
#  FILE NAME
#
#  srcdoc.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This perl program scans C source code, finds externally available functions
#  and creates a standard format man page from the source code, comments and
#  user inputs.  It tries to figure out the needed include files, gets the
#  function parameters and gets user comments associated with the functions.
#  It then allows you to edit (with 'gvim', for now) the extracted information
#  to add, clean up, and embellish as you see fit.  It then formats all the
#  information into a man page, section 3.
#  
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  srcdoc.pl,v
#  Revision 1.16  2002/01/05 02:35:29  gordon
#  Changed license to GPL.
#
#  Revision 1.15  2001/12/31 04:35:23  gordon
#  Changed directory and file handling to process tilde substitution
#  correctly.
#
#  Revision 1.14  2001/12/28 05:57:20  gordon
#  Added options for selecting storage class, comment extraction selection
#  and extraction update policy.  Reorganized option handling to properly
#  prioritize defaults, init file and command line options.
#
#  Revision 1.13  2001/11/30 13:01:09  gordon
#  Added HTML formatting capabilities.  Added functions in other source
#  files; this file calls those functions and sets up some globals.
#
#  Revision 1.12  2001/01/20 13:35:51  gordon
#  Moved the DisplaOpen variable setting before calling DisplayUserIf
#
#  Revision 1.11  2000/12/17 14:54:23  gordon
#  Corrected an error when not using an initialization file.
#
#  Revision 1.10  2000/12/17 14:10:00  gordon
#  Added a DisplayOpen variable that PrintStat uses to determine if it can
#  write to the GUI or should use stdout.
#
#  Revision 1.9  2000/12/16 17:35:08  gordon
#  Added CM display information for all files used.
#
#  Revision 1.8  2000/12/16 05:32:53  gordon
#  Added checks for Tk installation, correct setting of INC for finding
#  required scripts.
#
#  Revision 1.7  2000/12/14 05:49:23  gordon
#  Major revision.  Added perl/Tk user interface; reorganized several
#  functions.  Split into multiple files.
#
#  Revision 1.3  2000/06/27 05:11:06  gordon
#  Last version in this lib.  Moving to mkdocs lib.
#
#  Revision 1.2  2000/06/24 03:55:49  gordon
#  Added major sections for reading and formatting ASCII text files that are pree-formatted as man-like pages.
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#  BUGS AND ENHANCEMENTS
#
#  FIXME:  Add capability to cut the edited documentation back into the
#          source code file preceding the associated function.
#  FIXME:  Might change to use a database for the storage of function
#          information, instead of intermediate files like .dmp and .mtp.
#          Could use the linux db routines, or mySql.  That could be set
#          up as an option, so that more complex installation and setup
#          wouldn't be mandatory but if you have it, it seems cleaner.
#  FIXME:  Function name references should include a link to other man pages
#          if they exist.  It will need to maintain a listing of known man
#          pages.  The messy part of this is that as additional man pages are
#          generated, it should update the references, which would mean
#          regenerating the HTML pages for those functions.  On the other
#          hand, it could assume the HTML pages will be generated for all
#          functions and generate the refs as it goes.

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

#  Unless this is installed in a place in @INC, the following is used
#  to find the perl scripts used by this main program.  SRCDOCHOME must
#  be defined in the environment, which is done in a shell script that
#  is in the search path.

if( defined( $ENV{SRCDOCHOME} ) ) { push( @INC, $ENV{SRCDOCHOME} ); }

use subs qw( sd_findINC );

if( ! sd_findINC( "Tk" ) ) {
   printf( "Tk Not Found.  Check the perl/Tk installation as described\n" );
   printf( "in the INSTALL file.\n" );
   exit( 0 );
}

use Tk;

use Getopt::Std;

use strict;

require Tk::Menubar;
require Tk::Menu::Item;
require Tk::FileSelect;

if( ! sd_findINC( "statwin.pl" ) ) {
   printf( "Required script files not found.  Set \$SRCDOCHOME to the\n" );
   printf( "directory that contains the 'srcdoc' perl source code\n" );
   exit(0);
}

require 'statwin.pl';
require 'srcdocui.pl';
require 'config.pl';
require 'srcproc.pl';
require 'manproc.pl';
require 'printhtml.pl';

#  These are constants; they maybe should be declared in such a way as to
#  make them guaranteed constant.  They are initialized near the end of the
#  file, after all the subroutines.

use vars qw( $True $False $Undef $Edit $Generate $Label $Normal $Bold $Code );
use vars qw( $Name $Synopsis $Return $Description $Bugs $See_Also $Credits );
use vars qw( $Example $Functions $Author $RawFmt $ProcFmt $View $Editman );
use vars qw( $Minline $Ellipsis $Hyphen $Spacechar $Colonchar $Tabchar );
use vars qw( $No_tagsep $AllCmt $HdrCmt $IntCmt $SrcCmtAlways );
use vars qw( $NewCmtAlways $EdtCmtAlways $StgPublic $StgStatic $StgAll );

#  Declare global variables

use vars qw( $SrcDir $DispSrcDir $TagFile $DispTagFile $ManDir $DispManDir );
use vars qw( $InitFile $DispInitFile $UseGui @SrcFiles $Verbose $TextFmt );
use vars qw( $EditPage $Cflags $TextFile $SrcFile $CmtFlag );
use vars qw( $ReComment $TabFmt $TabStr $TabSize $OptList $RcStartDir );
use vars qw( %FuncList %keylist %namelist @sys_includes @lcl_includes );
use vars qw( %manrc_desc @GblFuncnames %RcSettings $ManFormat $HtmlFormat );
use vars qw( $StatWin $DisplayOpen $IndexFile $DispIndexFile $TmpDir );
use vars qw( $DispHtmlDir $HtmlDir %PageIndexList %StorageMode $StgModes );

#  Some counters to see how many times functions are executed.

my $printman_cnt    = 0;
my $printdesc_cnt   = 0;
my $fmtword_cnt     = 0;
my $hashkeys_cnt    = 0;
my $parseproto_cnt  = 0;
my $readtext_cnt    = 0;
my $readdump_cnt    = 0;
my $processcmt_cnt  = 0;
my $readsrc_cnt     = 0;
my $editfile_cnt    = 0;
my $writedesc_cnt   = 0;
my $fmtrefs_cnt     = 0;
my $resetlabels_cnt = 0;
my $indentcode_cnt  = 0;
my $reindent_cnt    = 0;
my $mkdocgui_cnt    = 0;
my $psel_cnt        = 0;
my $mkdoccmd_cnt    = 0;
my $initdesc_cnt    = 0;
my $initfunc_cnt    = 0;
my $savefunc_cnt    = 0;
my $getlabel_cnt    = 0;


#  Declare option variables, which are set by getopts

our $opt_c          = "";
our $opt_d          = "";
our $opt_D          = "";
our $opt_h          = "";
our $opt_i          = "";
our $opt_m          = "";
our $opt_n          = "";
our $opt_s          = "";
our $opt_t          = "";
our $opt_u          = "";
our $opt_v          = "";
our $opt_w          = "";
our $opt_z          = "";

#  These are default values for man page headings, used when no other
#  information is provided, either in an initialization file or from the
#  user interface.

my $ManSource    = "SRC";
my $ManHeading   = "Progarmmer's Manual";
my $ManSection   = 3;

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
#  First, the list of all functions contained in this file.
#
# ***************************************************************

use subs qw( Usage CmInfo ReadDump ProcessCmt ReadSource EditFile WriteDesc );
use subs qw( WriteSection PrintMan FormatRefs ResetLabels IndentCode Reindent );
use subs qw( PrintDesc FormatWord HashKeys DisplayUserIf PrintHtmlMan );
use subs qw( UpdateAuthor PrintHtmlIndex TildeSub EditFuncList );
use subs qw( ReadRcFile SetDefaultCfg ReadCfgHeader ReadCfgAuthor InitDesc );
use subs qw( PrintLineDesc InitFunc SaveFunc GetLabel ReadTextFmt ParseProto );
use subs qw( PrintExecStats ExtractSrc );
use subs qw( ConfigCmInfo mpCmInfo DisplayCmInfo spCmInfo StatwinCmInfo );
use subs qw( HtmlCmInfo );

# ***************************************************************
#
#  Subroutine Usage
#
#  Print script help message
#
# ***************************************************************

# $OptList = 'hzwvnc:d:i:D:t:s:u:';
sub Usage {
   printf( STDOUT "Script synopsis:\n" );
   printf( STDOUT "Usage: srcdoc [-h -z -w -v -n -c <cmt_source>\n"         );
   printf( STDOUT "              -d <text_file> -D <CPP_Directives>\n"      );
   printf( STDOUT "              -i <init_file> -s <tab_size> "             );
   printf( STDOUT "              -t <tag_file>\n"                           );
   printf( STDOUT "              -u <update_policy> <src>\n"                );
   printf( STDOUT "   where\n"                                              );
   printf( STDOUT "      -h : Print this synopsis\n"                        );
   printf( STDOUT "      -z : Print CM ID and info, then exit\n"            );
   printf( STDOUT "      -n : Don't use the graphical user interface\n"     );
   printf( STDOUT "      -w : Write/edit the function descriptions\n"       );
   printf( STDOUT "      -v : Set verbose mode\n"                           );
   printf( STDOUT "      -c <cmt_source> : Specify comment source, either\n");
   printf( STDOUT "              function headers or all comments, where:\n");
   printf( STDOUT "          [H]eader - Use the comments from the\n"        );
   printf( STDOUT "              function headers only\n"                   );
   printf( STDOUT "          [A]ll - Use the comments from both the\n "     );
   printf( STDOUT "              function header and internal comments\n"   );
   printf( STDOUT "      -d <text_file> : Specify the name of a text "      );
   printf( STDOUT "document to be formatted\n"                              );
   printf( STDOUT "      -D <CPP_Directives> : Pass the (quoted) directives\n"
         );
   printf( STDOUT "              directly to the C Pre-Processor\n"         );
   printf( STDOUT "      -i <init_file> : use <init_file> for setting"      );
   printf( STDOUT "options\n"                                               );
   printf( STDOUT "      -s <tab_size> : Use <tab_size> for tab character\n");
   printf( STDOUT "              substitution when reading inputs; defaults");
   printf( STDOUT " to 4\n"                                                 );
   printf( STDOUT "      -t <tag_file> : Use <tag_file> for the editor\n"   );
   printf( STDOUT "              tags file\n"                               );
   printf( STDOUT "      -u <update_policy> : Specify how to update a man " );
   printf( STDOUT "page, where:\n"                                          );
   printf( STDOUT "          [S]ource - Always generate man page from the\n");
   printf( STDOUT "                     source code file\n"                 );
   printf( STDOUT "          [N]ewest - Generate page from the souce code\n");
   printf( STDOUT "                     if it is newer than the edited\n"   );
   printf( STDOUT "                     intermediate file, otherwise use\n" );
   printf( STDOUT "                     the intermediate file\n"            );
   printf( STDOUT "          [E]dited - Always use the edited intermed-\n"  );
   printf( STDOUT "                     iate file for the man page\n"       );
   printf( STDOUT "      <src> : The source code files to be documented\n"  );

   return;
}

# ***************************************************************
#
#  Subroutine CmInfo
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub CmInfo {
   my $vcheck;

   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   printf( STDOUT "\n" );
   ConfigCmInfo;
   printf( STDOUT "\n" );
   mpCmInfo;
   printf( STDOUT "\n" );
   DisplayCmInfo;
   printf( STDOUT "\n" );
   spCmInfo;
   printf( STDOUT "\n" );
   StatwinCmInfo;
   printf( STDOUT "\n" );
   HtmlCmInfo;
   printf( STDOUT "\n" );

   return;
}

# ***************************************************************
#
#  Subroutine ReadDump
#
#  Description: This routine reads an intermediate file extracted
#               from source code files, containing function definitions
#               and the comments preceding the functions.  The
#               data is organized into an array of function descriptions
#               including the storage class, data type, name, and
#               source code comments, and for each parameter the data
#               type and name.  The data structure is organized as:
#
#               STGCLASS - a scalar value, static or NULL
#               TYPE     - a scalar value, possibly multiple words and '*'s
#               PARAMS:  - a list of elements containing 
#                  TYPE  - a string of 1 to several words
#                  NAME  - a single word for scalars
#                    -or-
#                  FUNC  - a name and parameter list for function pointers
#               COMMENTS - an arbitrarily long string spanning lines
#
# ***************************************************************

sub ReadDump {
   my $dfile   = shift;  # The dump file name
   my $sfile   = shift;  # The source file name

   my $inCmt   = $False;
   my $comment = "";
   my $cmttype;
   my $linekey;
   my $prmName;
   my $prmType;
   my $linetxt;
   my $fctname;
   my $rawline;
   my $line;
   my %prmElem;
   my $dupfunc = $False;
   my $fctdesc = {};
   my @prmList = ();
 
   $readdump_cnt++;

   open( DFIL, $dfile ) or die "Couldn't open $dfile\n";

   if( $Verbose eq $True ) {
      printf( "ReadDump: Start processing %s\n", $dfile );
   }

   $sfile =~ s/\/.*\///;

   FUNC: while( <DFIL> ) {
#  Reads a line from the file and splits the line at the first ':'
#  to get the keyword that describes what information follows.
#  If the function has already been defined, skip over it.

      $rawline = $_;
      chomp;
      $line = $_;
      ($linekey, $linetxt) = ($line =~ /(\S*)\:(.*)/);
      if( $dupfunc eq $True ) {
         if( !defined( $linekey ) || $linekey ne "FCT" ) { next FUNC; }
         else { $dupfunc = $False; }
      }

#  If the function comments section has been started, continue reading
#  comment lines until the 'ENDC' keyword is found.
#  Filter out separator lines, like all asterisks, hyphens, and equal signs.
#  Also strip off leading white space and asterisks or pound signs.

      if( $inCmt eq $True ) {
         if( !defined( $linekey ) || ($linekey ne "ENDC") ) {
            if( $rawline =~ /^\s*(\/\*)*\*{2,}(\*\/)*\s*$/ ) { next FUNC; }
            if( $rawline =~ /^\s*(\/\*)*\-{2,}(\*\/)*\s*$/ ) { next FUNC; }
            if( $rawline =~ /^\s*(\/\*)*\={2,}(\*\/)*\s*$/ ) { next FUNC; }
            if( !($rawline =~ /^$/) ) {
               $rawline =~ s/^\s*\*+(\s+)/$1/;
            }
            $comment = $comment . $rawline;
            next FUNC;
         }
      }
#  All lines other than comments start with a keyword:  FCT, STG, TYP,
#  PRM or CMT.  The data is cleaned up, stripping leading or trailing
#  whitespace, then saved in a function description data structure.
      if( !defined( $linekey ) && ( $inCmt eq $False ) ) {next FUNC; }

      if( defined( $linetxt ) ) {
         $linetxt =~ s/^\s*//;
         $linetxt =~ s/\s*$//;
      }

      if( $linekey eq "FCT" ) {
         $fctname = $linetxt;
   #  Does a check to see if a function is defined in multiple places in the
   #  selected source code.  Rejects the subsequent entries, so defaults to
   #  the first occurrence found and prints a message reporting the duplicate.
         if( exists( $FuncList{$fctname} ) ) {
            PrintStat( "File $sfile: Dup Func $fctname ignored", 'advise' );
            $dupfunc = $True;
            next FUNC;
         }
         $fctdesc = {};
         $fctdesc->{datafmt}     = $RawFmt;
         $fctdesc->{description} = "";
         $fctdesc->{summary}     = "";
         $fctdesc->{ret_desc}    = "";
         $fctdesc->{bug_desc}    = "";
         $fctdesc->{sysincl}     = ();
         $fctdesc->{lclincl}     = ();
         $fctdesc->{params}      = ();
         $fctdesc->{stg}         = "";
         $fctdesc->{refs}        = "";
         $fctdesc->{example}     = "";
         $fctdesc->{credits}     = "";
         $fctdesc->{author}      = "";
         $fctdesc->{funclist}    = "";
         $fctdesc->{intcmt}      = "";
         $fctdesc->{srcfile}     = $dfile;
         $FuncList{$fctname}     = $fctdesc;
         push( @GblFuncnames, $fctname );
         next FUNC;
      }
      if( $linekey eq "STG" ) {
         if( defined( $linetxt ) ) {
            $fctdesc->{stg} = $linetxt;
         }
         next FUNC;
      }
      if( $linekey eq "TYP" ) {
         if( defined( $linetxt ) ) {
            $fctdesc->{typ} = $linetxt;
         }
         else { $fctdesc->{typ} = ""; }
         next FUNC;
      }
#  For parameters, the line is split into the parameter type and parameter
#  name as indicated by keyword 'NAM:' or 'FUNC:'.  The parameters are put
#  into an array of the parameters and a reference to the array is put in the
#  function definition $fctdesc->{params}.

      if( $linekey eq "PRM" ) {
         if( $linetxt =~ /NAM/ ) {
            ($prmType, $prmName) = ($linetxt =~ /(.*) NAM\: (.*)/);
         } elsif( $linetxt =~ /FUNC/ ) {
            ($prmType, $prmName) = ($linetxt =~ /(.*) FUNC\: (.*)/);
         }  
         if( !defined( $prmType ) ) {
            $prmType = $linetxt;
         }
         $prmType =~ s/^\s*//;
         $prmType =~ s/\s*$//;
         if( defined( $prmName ) ) {
            $prmName =~ s/^\s*//;
            $prmName =~ s/\s*$//;
         }
         else { $prmName = ""; }
         $prmElem{ptype} = $prmType;
         $prmElem{pname} = $prmName;
         push @prmList, { %prmElem };
         next FUNC;
      }
#  When keyword CMT is found, begin reading the function description from
#  the file, and also finish copying the parameter list into the permanent
#  function data structure and reset the temporary parameter collector to
#  empty so it's ready for the next one.
      if( $linekey eq "HDRCMT" ) {
         $fctdesc->{params} = [ @prmList ];
         $#prmList = -1;
         $inCmt = $True;
         $comment = "";
         $cmttype = $HdrCmt;
         next FUNC;
      }
      if( $linekey eq "INTCMT" ) {
         $inCmt = $True;
         $comment = "";
         $cmttype = $IntCmt;
         next FUNC;
      }
      if( $linekey eq "ENDC" ) {
         if( $cmttype eq $HdrCmt ) {
            ProcessCmt( $comment, $fctdesc );
         } else {
            $fctdesc->{intcmt} = $comment;
         }
         $inCmt = $False;
         next FUNC;
      }
   }
   close( DFIL );

   return;
}

# ***************************************************************
#
#  Subroutine ProcessCmt
#
#  Description:  Process the description block extracted from the source
#                code.  If the description contains 'man' keywords, split
#                the function header into the appropriate 'man' page sections.
#                If it's unformatted text, it all goes under the DESCRIPTION
#                section.
#
# ***************************************************************

sub ProcessCmt {
   my  $comment  = shift;
   my  $fctdesc  = shift;

   my  @lines;
   my  $line;
   my  $section  = $Description;
   my  $linekey;
   my  $sep      = "\n";
   my  $dopara;
   my  $codefrag = $False;
   my  $retdesc  = "";
   my  $desc     = "";
   my  $summary  = "";
   my  $bugs     = "";
   my  $refs     = "";
   my  $credits  = "";
   my  $author   = "";
   my  $example  = "";
   my  $blankline = $False;
   my  $functionlist;
   my  $ftype    = "SUBR";
   my  $para     = $False;

   $processcmt_cnt++;

   @lines = split( '\n', $comment );

   CMTLINE:  foreach $line (@lines) {
      if( $line =~ /\S+/ and $line =~ /^[ \*]*$/ ) { $line = ""; }

      if( $line eq "" ) {
         if( $blankline eq $True ) {
            next CMTLINE;
         } else {
            $blankline = $True;
         }
      } else { $blankline = $False; }

      if( $line =~ /^\s*<SUBSYSTEM/ or $line =~ /^\s*<\/SUBSYSTEM>/ ) {
         $ftype = "SUBS";
         next CMTLINE;
      }
      if( ( $line =~ /^\s*<SUBROUTINE/ or $line =~ /^\s*<\/SUBROUTINE>/ ) ) {
         $ftype = "SUBR";
         next CMTLINE;
      }

      ($linekey) = ($line =~ /^\s*(\w*\s?\w*).*/);

      if( $linekey eq "NAME" )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq $Description )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq "RETURN VALUE" )
         { $section = $Return;   $sep = ""; next CMTLINE; }
      if( $linekey eq "RETURN VALUES" )
         { $section = $Return;   $sep = ""; next CMTLINE; }
      if( $linekey eq $Bugs )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq "SEE ALSO" )
         { $section = $See_Also; $sep = ""; next CMTLINE; }
      if( $linekey eq $Credits )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq $Author )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq $Example )
         { $section = $linekey;  $sep = ""; next CMTLINE; }
      if( $linekey eq $Functions )
         { $section = $linekey;  $sep = ""; next CMTLINE; }

      if( $line =~ /<CODE>/ ) {
         $codefrag = $True;
         $line = "<CODE>";
      }

      if( $codefrag eq $True ) {
         if( $line =~ /<\/CODE>/ ) { $line = "<\/CODE>"; }
      }

      if( $section eq $Name ) {
         if( $line =~ /^.*\-\s*\w+/ ) {
            ($summary) = ($line =~ /^.*\-\s*(.*)/ );
         }
         next CMTLINE;
      }
      if( $section eq $Description ) {
         $desc = $desc . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Return ) {
         $retdesc = $retdesc . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Bugs ) {
         $bugs = $bugs . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $See_Also ) {
         $refs = $refs . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Credits ) {
         $credits = $credits . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Author ) {
         $author = $author . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Example ) {
         $example = $example . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
      if( $section eq $Functions ) {
         $functionlist = $functionlist . $sep . $line;
         $sep = "\n";
         next CMTLINE;
      }
   }
   $fctdesc->{summary}     = $summary;
   $fctdesc->{description} = $desc;
   $fctdesc->{example}     = $example;
   $fctdesc->{ret_desc}    = $retdesc;
   $fctdesc->{bug_desc}    = $bugs;
   $fctdesc->{refs}        = $refs;
   $fctdesc->{credits}     = $credits; 
   $fctdesc->{author}      = $author; 
   $fctdesc->{funclist}    = $functionlist;
   $fctdesc->{blktype}     = $ftype;

   return;
}

# ***************************************************************
#
#  Subroutine ReadSource
#
#  Description:  Reads the source code file and extracts preprocessor
#                directives 'include' and 'define'.  The includes will
#                be used to start the 'synopsis' section listing of include
#                files.  The defined names will be saved in a list, but
#                how they'll be used isn't right yet clear.
#
# ***************************************************************

sub ReadSource {
   my $filename = $_[0];

   my $incl;
   my @srcline;

   $readsrc_cnt++;

   open( SRCFILE, "<$filename" ) or
                     die "Couldn't open source file $filename\n" ;

   @lcl_includes = ();
   @sys_includes = ();

   while( <SRCFILE> ) {
      chomp;
      @srcline = split( '\s+', $_ );
      if( $#srcline >= 0 && $srcline[0] eq "#include" ) {
         if( $srcline[1] =~ /\<.*\>/ ) {
            push( @sys_includes, $srcline[1] );
         }
         elsif( $srcline[1] =~ /\".*\"/ ) {
            push( @lcl_includes, $srcline[1] );
         }
      }
   }
   if( $Verbose eq $True ) {
      for $incl (@sys_includes) { printf( "Sys Include: $incl\n" ); }
      for $incl (@lcl_includes) { printf( "Lcl Include: $incl\n" ); }
   }
   close( SRCFILE );
   return;
}

# ***************************************************************
#
#  Subroutine EditFile
#
#  Description:  Edit and create man page for the specified function.
#
# ***************************************************************

sub EditFile {
#  Input Parameters
   my $fct      = shift;
   my $edit_man = shift;
   my $srcfile  = shift;

#  Variables local to this function
   my $desc_file = $TmpDir . '/' . $fct . ".mtp";

   $editfile_cnt++;
   if( $Verbose eq $True ) {
      printf( "EditFile: Call ReadTextFmt on fct %s, file %s\n",
              $desc_file, $srcfile );
   }
   if( $ReComment eq $SrcCmtAlways ) { WriteDesc( $fct, $desc_file ); }
   elsif( $ReComment eq $NewCmtAlways and -e $desc_file
                                      and -C $srcfile lt -C $desc_file ) {
      WriteDesc( $fct, $desc_file );
   } elsif( ! -e $desc_file ) {
      WriteDesc( $fct, $desc_file );
   }
   if( $edit_man eq $Edit ) {
      my $editcmd = "gvim +\"set tag=$TagFile\" $desc_file";
      qx { $editcmd };
   }
   ReadTextFmt( $desc_file, $srcfile );

   if( $Verbose eq $True ) {
      printf( "EditFile: Call PrintMan( %s )\n", $fct );
   }
   if( $ManFormat eq $True )  { PrintMan( $fct ); }
   if( $HtmlFormat eq $True ) { PrintHtmlMan( $fct ); }

   PrintStat( "Generated Man Page for $fct", 'advise' );

   return;
}

# ***************************************************************
#
#  Subroutine WriteDesc
#
#  Description:  Writes the function definitions out to an edit file.
#                Depending on the comment flag specified on the command
#                line, the source code comments are written along with the
#                description previously written.
#                The function DESCRIPTION, RETURN VALUES, and BUGS
#                sections of the man page are outlined for the developer
#                to fill in.
#
# ***************************************************************

sub WriteDesc {
   my $fcts      = $_[0];
   my $tmpfile   = $_[1];

   my $func_desc;
   my $typename;
   my @cmtlines;
   my $cmtline;
   my $srccmt;
   my $internalcmt;
   my @tmplist;   
   my $idx;
   my $stgclass;
   my $incl;

   $writedesc_cnt++;

   $func_desc = $FuncList{$fcts};

   open( TMPDESC, ">$tmpfile" ) or die "Couldn't open $tmpfile";
 
   $stgclass = $func_desc->{stg};
   $typename = $func_desc->{typ};
   if( $func_desc->{blktype} eq "SUBR" ) {
      printf( TMPDESC "<SUBROUTINE NAME=\"%s\">\n", $fcts );
   } else { printf( TMPDESC "<SUBSYSTEM NAME=\"%s\">\n", $fcts ); }
   printf( TMPDESC "\nNAME\n" );
   printf( TMPDESC "\t%s - ", $fcts );
   if( $func_desc->{summary} eq "" ) {
      printf( TMPDESC "<Enter a brief description>\n" );
   }
   else {
      printf( TMPDESC "%s\n\n", $func_desc->{summary} );
   }

   printf( TMPDESC "\n" );
   printf( TMPDESC "SYNOPSIS\n" );

#  If the #includes were read from a previously edited function document
#  (*.mtp file), use the list there, otherwise, use the list extracted from
#  the source code file.

   if( $#{$func_desc->{sysincl}} < 0 ) {
      for $incl (@sys_includes) { printf( TMPDESC "\t#include $incl\n" ); }
   }
   else {
      for $incl (@{$func_desc->{sysincl}}) {
         printf( TMPDESC "\t#include $incl\n" );
      }
   }
   if( $#{$func_desc->{lclincl}} < 0 ) {
      for $incl (@lcl_includes) { printf( TMPDESC "\t#include $incl\n" ); }
   }
   else {
      for $incl (@{$func_desc->{lclincl}}) {
         printf( TMPDESC "\t#include $incl\n" );
      }
   }

   if( $func_desc->{blktype} eq "SUBR" ) {
      printf( TMPDESC "\n" );
      printf( TMPDESC "\t" );
      if( $stgclass ne "" ) {
         printf( TMPDESC "%s ", $stgclass );
      }
      printf( TMPDESC "%s %s (", $typename, $fcts  );

      if( $#{$func_desc->{params}} >= 0 ) {
         printf( TMPDESC "\n" );
         @tmplist = @{ $func_desc->{params} };
         for $idx (0 .. $#tmplist) {
            printf( TMPDESC "\t\t  %s", $tmplist[$idx]->{ptype} );
            if( defined( $tmplist[$idx]->{pname} ) ) {
               printf( TMPDESC " %s", $tmplist[$idx]->{pname} );
            }
            if( $idx < $#tmplist ) { printf( TMPDESC ",\n" ); }
         }
         printf( TMPDESC "\n\t\t);\n" );
      } else { printf( TMPDESC " );\n" ); }
   }

   printf( TMPDESC "\n" );
   if( $func_desc->{datafmt} eq $ProcFmt ) {
      WriteSection( \*TMPDESC, $func_desc->{description},  "DESCRIPTION"   );
      if( defined( $func_desc->{intcmt} ) ) {
         WriteSection( \*TMPDESC, $func_desc->{intcmt},       ""           );
      }
      WriteSection( \*TMPDESC, $func_desc->{ret_desc},     "RETURN VALUES" );
      WriteSection( \*TMPDESC, $func_desc->{example},      "EXAMPLE"       );
      WriteSection( \*TMPDESC, $func_desc->{bug_desc},     "BUGS"          );
      WriteSection( \*TMPDESC, $func_desc->{credits},      "CREDITS"       );
      WriteSection( \*TMPDESC, $func_desc->{author},       "AUTHOR"        );
      WriteSection( \*TMPDESC, $func_desc->{refs},         "SEE ALSO"      );
      WriteSection( \*TMPDESC, $func_desc->{funclist},     "FUNCTIONS"     );
      goto WRITEDESC_EXIT;
   }

   printf( TMPDESC "DESCRIPTION\n" );
   $srccmt = $func_desc->{description};
   @cmtlines = split( '\n', $srccmt );
   for $cmtline (@cmtlines) {
      printf( TMPDESC "%s\n", $cmtline );
   }
   if( $CmtFlag eq $AllCmt ) {
      $internalcmt = $func_desc->{intcmt};
      @cmtlines = split( '\n', $internalcmt );
      for $cmtline (@cmtlines) {
         printf( TMPDESC "%s\n", $cmtline );
      }
   }

   printf( TMPDESC "\n" );
   if( $func_desc->{ret_desc} ne "" ) {
      printf( TMPDESC "RETURN VALUES\n" );
      printf( TMPDESC "%s\n", $func_desc->{ret_desc} );
   } else {
      printf( TMPDESC "RETURN VALUES\n" );
      printf( TMPDESC
              "\t%s() returns type %s\n\t<Enter additional description>\n",
              $fcts, $typename );
   }

   if( $func_desc->{example} ne "" ) {
      printf( TMPDESC "EXAMPLE\n" );
      printf( TMPDESC "%s\n", $func_desc->{example} );
   }

   if( $func_desc->{bug_desc} ne "" ) {
      printf( TMPDESC "BUGS\n" );
      printf( TMPDESC "%s\n", $func_desc->{bug_desc} );
   }

   if( $func_desc->{credits} ne "" ) {
      printf( TMPDESC "CREDITS\n" );
      printf( TMPDESC "%s\n", $func_desc->{credits} );
   }

   if( $func_desc->{author} ne "" ) {
      printf( TMPDESC "AUTHOR\n" );
      printf( TMPDESC "%s\n", $func_desc->{author} );
   }

   if( $func_desc->{refs} ne "" ) {
      printf( TMPDESC "SEE ALSO\n" );
      printf( TMPDESC "%s\n", $func_desc->{refs} );
   }

   WRITEDESC_EXIT:
   if( $func_desc->{blktype} eq "SUBR" ) {
      printf( TMPDESC "\n</SUBROUTINE>\n" );
   } elsif( $func_desc->{blktype} eq "SUBS" ) {
      printf( TMPDESC "\n</SUBSYSTEM>\n" );
   }
   close( TMPDESC );
}

# ***************************************************************
#
#  Subroutine WriteSection
#
#  Description:  Writes a text section from the formatted data structure
#                to an intermediate (*.mtp) file.
#
# ***************************************************************

sub WriteSection {
   my $file       = $_[0];
   my @fmtdesc    = @{$_[1]};
   my $sec_name   = $_[2];

   my $paragraph;       # Hash reference to the paragraph.
   my @words;
   my $word;
   my $linelen    = 0;
   my $sp_idx;
   my $indent_len;
   my $offset_len;
   my $max_len;
   my $tag_char;
   my $tabcnt;

   $printdesc_cnt++;

   if( $#fmtdesc < 0 ) { return; }
#  Now process each paragraph, defined as a hash including text, indent,
#  line spacing, length and paragraph type.

   Reindent( \@fmtdesc );

   printf( $file "%s\n", $sec_name ) if( $sec_name ne "" );
   FMTLINE: foreach $paragraph (@fmtdesc) {

      $indent_len = $paragraph->{indent} + 4;

      for $sp_idx (1..$indent_len) { printf( $file " " ); }

      if( $paragraph->{fmttype} eq $Label ) {
         $tag_char = $paragraph->{tagsep};
         printf( $file "%s", $paragraph->{label} );
         $offset_len = $paragraph->{taglen} - length( $paragraph->{label} );
         if( $tag_char eq $Ellipsis or $tag_char eq $Spacechar ) {
            printf( $file " " );
            for $sp_idx (1..$offset_len + 2) {
               printf( $file "%s", $tag_char );
            }
            printf( $file " " );
            $indent_len = $indent_len + $paragraph->{taglen} + 4;
         } elsif( $tag_char eq $Tabchar ) {
            $tabcnt = int( ($offset_len + 2) / 8 ) + 1;
            for $sp_idx (1..$tabcnt) { printf( $file "\t" ); }
            $indent_len = $indent_len + $paragraph->{taglen} + $tabcnt * 8;
            $indent_len = int( $indent_len / 8 ) * 8;
         } elsif( $tag_char eq $Colonchar ) {
            printf( $file ":" );
            $tabcnt = int( ($offset_len + 1) / 8 ) + 1;
            for $sp_idx (1..$tabcnt) { printf( $file "\t" ); }
            $indent_len = $indent_len + $paragraph->{taglen} + $tabcnt * 8;
            $indent_len = int( $indent_len / 8 ) * 8;
         } elsif( $tag_char eq $Hyphen ) {
            printf( $file " -" );
            for $sp_idx (1..$offset_len) { printf( $file " " ); }
            printf( $file " " );
            $indent_len = $indent_len + $paragraph->{taglen} + 4;
         } else {
            printf( $file " " );
            $indent_len = $indent_len + $paragraph->{taglen} + 1;
         }

         $linelen = 0;
      } else { $linelen = 0; }

      if( $paragraph->{fmttype} eq $Code ) {
         my @codelines = split( "\n", $paragraph->{text} );
         my $codeline;
         printf( $file "<CODE>\n" );
         foreach $codeline (@codelines) {
            for $sp_idx (1..$indent_len) { printf( $file " " ); }
            printf( $file "%s\n", $codeline );
         }
         for $sp_idx (1..$indent_len) { printf( $file " " ); }
         printf( $file "</CODE>\n" );
         next FMTLINE;
      }

      $max_len = 78 - $indent_len;

      @words = split( ' ', $paragraph->{text} );
      foreach $word (@words) {
         if( $linelen + length( $word ) > $max_len ) {
            printf( $file "\n" );

            for $sp_idx (1..$indent_len) { printf( $file " " ); }
            $linelen = length( $word );
         } elsif( $linelen > 0 ) {
            printf( $file " " );
            $linelen += length( $word ) + 1;
         } else { $linelen = length( $word ); }
         printf( $file "%s", $word );
      }
      
      printf( $file "\n" );
      if( $paragraph->{spaces} > 0 ) { printf( $file "\n" ); }
   }
   return;
}

# ***************************************************************
#
#  Subroutine PrintMan
#
#  Description:  Print a man page for a function from the information
#                obtained from the source code and the developer's
#                inputs.  It does a bunch of troff formatting, using
#                bold and italics to highlight function names, data types
#                and section headings.
#
# ***************************************************************

sub PrintMan {
   my $fcts = $_[0];

   my $hdrname;
   my $func;
   my @prmlist;
   my $idx;
   my $indent;
   my $typename;
   my $stgclass;
   my $totlen;
   my $fname;
   my $inclkey;
   my $incl;
   my @lcl_incl_list;
   my @sys_incl_list;
   my $timebuf;
   my $manpage;
   my $shrt_hdr;
   my $sec;
   my $min;
   my $hour;
   my $mday;
   my $mon;
   my $year;
   my $wday;
   my $yday;
   my $isdst;
   my $hdrlen;

   $printman_cnt++;

   ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime;
   $year += 1900;
   $mon += 1;
   $sec = 0; $min = 0; $hour = 0; $wday = 0; $yday = 0; $isdst = $False;
   $timebuf = $year . "/" . $mon . "/" . $mday;

   if( %manrc_desc ) {
      if( $manrc_desc{header}{manual} ne "" ) {
         $ManHeading = $manrc_desc{header}{manual};
      }
      if( $manrc_desc{header}{section} > 0 ) {
         $ManSection = $manrc_desc{header}{section};
      }
      if( $manrc_desc{header}{source} ne "" ) {
         $ManSource = $manrc_desc{header}{source};
      }
   }

#  Open a man page file, print headers

   $manpage = $fcts . "." . $ManSection;
   $manpage = $ManDir . "/" . $manpage;
   $func = $FuncList{$fcts};
   $hdrname = $fcts;
   $hdrlen = int( ( 54 - length( $ManHeading ) ) / 2 );
   if( length( $hdrname ) > $hdrlen ) {
      ($shrt_hdr) = ($hdrname =~ /(\w{1,$hdrlen}).+/ );
      $hdrname = $shrt_hdr . "..";
   }


   open( MANP, ">$manpage" ) or die "Couldn't open man page file $manpage";
   printf( MANP ".TH %s %d  %s \"%s\" \"%s\"\n",
            $hdrname, $ManSection, $timebuf, $ManSource, $ManHeading );
   printf( MANP ".SH NAME\n" );
   if( length( $fcts ) + length( $func->{summary} ) > 57 ) {
      printf( MANP ".TP %d\n", length( $fcts ) + 3 );
   }
   printf( MANP "%s", $fcts );
   if( $func->{summary} ne "" ) {
      printf( MANP " \\- %s",
              $func->{summary} );
   }
   printf( MANP "\n" );

#  Print the SYNOPSIS section header.  List the include files needed and the
#  function prototype(s).

   @sys_incl_list = @{$func->{sysincl}};
   @lcl_incl_list = @{$func->{lclincl}};
   if( $func->{blktype} eq "SUBR"  or
          $#sys_incl_list >= 0   or
          $#lcl_incl_list >= 0 )    {
      printf( MANP ".SH SYNOPSIS\n" );
   }
   for $incl (@sys_incl_list) {
       printf( MANP ".B #include %s\n", $incl );
       printf( MANP ".br\n" );
   }

#  The quotes around the local file name need to be handled using *(lq *(rq 
#  - it's a perl thing, what can I say.

   for $incl (@lcl_incl_list) {
       ($inclkey, $fname) = ( $incl =~ /(.*)\s*\"(.*)\"/ );
       printf( MANP ".B #include %s\\*(lq%s\\*(rq\n", $inclkey, $fname );
       printf( MANP ".br\n" );
   }

#  Print the function prototype.  It needs bold for the name and types,
#  and italic for the parameter names.  If it's a long line, it should
#  be broken up and indented nicely, but if it all fits on one line, then
#  it's best left on one line.

   if( $func->{blktype} eq "SUBR" ) {
      my $longparm = 0;
      my $parmlen;
      my $multiln;
      my $preindent = $False;

      printf( MANP ".br\n" );
      printf( MANP ".sp\n" );
      printf( MANP ".BI \"" );

      $stgclass = $func->{stg};
      $typename = $func->{typ};
      $totlen = length( $stgclass ) + 1;
      $totlen += length( $typename ) + 1;
      $totlen += length( $fcts ) + 1;
      @prmlist = @{ $func->{params} };

      for $idx (0 .. $#prmlist ) {
         $parmlen  = length( $prmlist[$idx]->{ptype} ) + 1;
         $parmlen += length( $prmlist[$idx]->{pname} ) + 1;
         $totlen  += $parmlen;
         if( $parmlen > $longparm ) { $longparm = $parmlen; }
      }
      if( $totlen > 55 ) {
         $multiln = $True;
         $indent = length( $stgclass ) + length( $typename ) +
                   length( $fcts ) + 3;
         if( ($indent + $longparm) > 55 ) {
            $preindent = $True;
            if( $longparm <= 58 ) { $indent = (56 - $longparm); }
            else { $indent = 3; }
         }
      } else { $multiln = $False; }

      if( $stgclass ne "" ) { printf( MANP "%s ", $stgclass ); }
      printf( MANP "%s %s( ", $typename, $fcts  );
      if( $multiln eq $True ) {
         for $idx (0 .. $#prmlist) {
            if( $preindent eq $True and $idx == 0 ) {
               printf( MANP "\"\n.RS %d\n.BI \"", $indent );
            }
            printf( MANP "%s ", $prmlist[$idx]->{ptype} );
            if( $prmlist[$idx]->{pname} ne "" ) {
               my $parmname = $prmlist[$idx]->{pname};
               if( $parmname =~ /\(+\*+\w*\)+/ ) {
                  my $pre;
                  my $fname;
                  my $fparms;
                  ($pre, $fname, $fparms) = ($parmname =~ /(\W*)(\w*)(.*)/);
                  printf( MANP "%s\" %s \"%s", $pre, $fname, $fparms );
               } else {
                  printf( MANP "\" %s \"", $parmname );
               }
            }
            if( $idx eq 0 and $preindent eq $False ) {
               printf( MANP ",\"\n.RS %d\n.BI \"", $indent );
               $preindent = $False;
            }
            elsif( $idx < $#prmlist ) { printf( MANP ",\"\n.br\n.BI \"" ); }
            else {
               printf( MANP " );\"\n" );
               printf( MANP ".br\n.RE\n" );
            }
         }
      } else {
         for $idx (0 .. $#prmlist) {
            printf( MANP "%s", $prmlist[$idx]->{ptype} );
            if( $prmlist[$idx]->{pname} ne "" ) {
               printf( MANP " \" %s \"", $prmlist[$idx]->{pname} );
            }
            if( $idx < $#prmlist ) { printf( MANP ", " ); }
            else { printf( MANP " );\n.br\n" ); }
         }
      }
   }
#  Now prints the description, return values and bugs, if any are specified.
#  The PrintDesc subroutine is called to look at each word to see if it
#  is an 'interesting' word that needs highlighting. 

   if( @{ $func->{description} } ) {
      printf( MANP ".SH DESCRIPTION\n" );
      PrintDesc( $func->{description}, $fcts );
   }
   if( @{ $func->{example} } ) {
      printf( MANP ".SH EXAMPLE\n" );
      PrintDesc( $func->{example}, $fcts );
   }
   if( @{ $func->{ret_desc} } ) {
      printf( MANP ".SH RETURN VALUES\n" );
      PrintDesc( $func->{ret_desc}, $fcts );
   }
   if( @{ $func->{bug_desc} } ) {
      printf( MANP ".SH BUGS\n" );
      PrintDesc( $func->{bug_desc}, $fcts );
   }
   if( @{ $func->{credits} } ) {
      printf( MANP ".SH CREDITS\n" );
      PrintDesc( $func->{credits}, $fcts );
   }
   if( @{ $func->{funclist} } ) {
      printf( MANP ".SH FUNCTIONS\n" );
      PrintDesc( $func->{funclist}, $fcts );
   }
   if( @{ $func->{refs} } ) {
      printf( MANP ".SH SEE ALSO\n" );
      FormatRefs( $func );
      PrintDesc( $func->{refs}, $fcts );
   }
   if( @{ $func->{author} } ) {
      printf( MANP ".SH AUTHOR\n" );
      PrintDesc( $func->{author}, $fcts );
   } elsif( %manrc_desc ) {
      printf( MANP ".SH AUTHOR\n" );
     
      PrintDesc( $manrc_desc{author}, $fcts );

   } else {
      my $passwdent;
      my $username = $ENV{USER};
      my @pwstuff = getpwnam($username);
      my $author_name = $pwstuff[6];
 
      printf( MANP ".br\n" );
      printf( MANP ".SP\n" );
      printf( MANP ".SH AUTHOR\n" );
      printf( MANP "%s\n", $author_name );
      printf( MANP ".PP\n" );
   }
   close( MANP );
}

# ***************************************************************
#
#  Subroutine FormatRefs
#
#  Description:  For cases where references ("SEE ALSO") are written as
#                a column list, reformat them as a comma-separated list
#                to comply with Unix standard man page formats.
#
# ***************************************************************

sub FormatRefs {
   my $func = $_[0];

   my @reflist = @{$func->{refs}};
   my $ref;
   my $idx = 0;
   my @newref = ();
   my $reftext = "";
   my $first_indent = 0;
   my %newhash;

   $fmtrefs_cnt++;

#  Create a new hash and initialize it to all NULLs

   InitDesc( \%newhash );

#  For each reference entry, see if it has a single reference entry on a
#  line.  If it does, concatenate the entry with previous reference entries.
#  If an entry that isn't a man page reference is found, save the concatenated
#  list and then save the new entry as is.  Then initialize a new hash entry
#  to be used for subsequent man page references.

   foreach $ref ( @reflist ) {
      $ref->{text} =~ s/^\s*//;
      if( $ref->{text} =~ /^\w+\(.*\)\W*$/ ) {
         if( $idx == 0 ) {
            $reftext = $ref->{text};
            $first_indent = $ref->{indent};
         }
         else { $reftext = $reftext . ", " . $ref->{text}; }
         $idx++;
      }
      else {
         if( $reftext ne "" ) {
            $newhash{text} = $reftext;
            $newhash{fmttype} = $Normal;
            $newhash{indent}  = $ref->{indent};
            $newhash{linelen} = length( $reftext );
            $newhash{spaces}  = $ref->{spaces};
            push( @newref, { %newhash } );
            push( @newref, { %$ref } );
            $reftext = "";
            $idx = 0;
            InitDesc( \%newhash );
         }
      }
   }

#  When the end of the list is found, if some man page references have
#  been concatenated, save the concatenated list in a paragraph hash.
#  Then replace the original array of hashes with the new array of hashes.
#  Set the paragraph space value to 1 for the last entry, regardless of what
#  the entries are.

   if( $reftext ne "" ) {
      $newhash{text} = $reftext;
      $newhash{fmttype} = $Normal;
      $newhash{indent}  = $first_indent;
      $newhash{linelen} = length( $reftext );
      $newhash{spaces}  = 1;
      push( @newref, { %newhash } );
   }
   $func->{refs} = [ @newref ];
   $func->{refs}[$#newref]->{spaces} = 1;
   return;
}

# ***************************************************************
#
#  Subroutine ResetLabels
#
#  Description:  This function resets all the indents to a standard size.
#  For NORMAL paragraphs, sets a base indent, then for each indent level,
#  adds a tab size.
#
#  For LABELED paragraphs, needs to look at all the LABELED paragraphs in
#  sequence and figure out what the indent should be for all of them.
#  For CODE sections, needs to see if any outdenting is needed.
#
# ***************************************************************

sub ResetLabels {
   my $section  = shift;
   my $first    = shift;
   my $listlen  = shift;

   my $max      = 0;
   my $lidx;
   my $last     = $first + $listlen;
   my $maxline  = 0;
   my $numlabel = $True;
   my $minindent = 0xFFFF;

   $resetlabels_cnt++;

#  Make a first pass to get the longest label and the number of words in
#  the text.  Also, see if all the labels are numeric, because they'll be
#  right-justified.

   foreach $lidx ( $first .. $last ) {
      my  @linewords = split( ' ', $section->[$lidx]->{text} );

      if( !($section->[$lidx]->{label} =~ /^\-*\d+/ ) ) {
         $numlabel = $False;
      }
      if( $section->[$lidx]->{taglen} > $max ) {
         $max = $section->[$lidx]->{taglen};
      }
      if( $#linewords > $maxline ) { $maxline = $#linewords; }
      if( $section->[$lidx]->{indent} < $minindent ) {
         $minindent = $section->[$lidx]->{indent};
      }
   }

#  Make a second pass to set all the label sizes (taglen) to the same
#  size.  If there's only one word in the text, it's really a double column
#  and the tag format is left blank; otherwise use a default BOLD format.

   if( $numlabel eq $True ) {
      my $tabdiff;

      foreach $lidx ( $first .. $last ) {
         $tabdiff = $max - $section->[$lidx]->{taglen};
         $section->[$lidx]->{indent} = $minindent + $tabdiff;
         $section->[$lidx]->{tagfmt} = "I";
      }
   } else {
      foreach $lidx ( $first .. $last ) {
         $section->[$lidx]->{indent} = $minindent;
         $section->[$lidx]->{taglen} = $max;
         if( $maxline < 1 ) {
            $section->[$lidx]->{tagfmt} = "";
         } else {
            $section->[$lidx]->{tagfmt} = "B";
         }
      }
   }
   return $minindent;
}

# ***************************************************************
#
#  Subroutine IndentCode
#
#  Description:  Sets the indentation for CODE paragraphs by finding
#                the minimum indent in the source and measureing the
#                longest line.  Sets the indent for the whole section
#                by finding the best fit.  If the lines are short enough
#                it uses the indentation in the original text, otherwise
#                it shifts the lines to the left to prevent line
#                wrapping, using a hard-coded line length.
#
# ***************************************************************

sub IndentCode {
   my $code_seg = $_[0];

   my @codelines;
   my $indent;
   my $text;
   my @int_codelines = ();
   my $int_codeline;
   my $min_indent    = 0xFFFF;
   my $max_text      = 0;
   my $predent;
   my $new_codetext  = "";
   my $line;
   my $cutstr;
   my $cutfmt;

   $indentcode_cnt++;

   @codelines = split( '\n', $code_seg->{text} );

   foreach $line (@codelines) {
      ($indent, $text) = ( $line =~ /^(\s*)(.*)/ );
      $indent =~ s/\t/$TabStr/g;
      $int_codeline = $indent . $text;
      push( @int_codelines, $int_codeline );
      if( length( $indent ) < $min_indent and length( $text ) > 0 ) {
         $min_indent = length( $indent );
      }
      if( length( $text ) > $max_text ) {
         $max_text = length( $text );
      }
   }
   $predent = 64 - $max_text;
   if( $predent < 0 ) {
      if( $predent < -7 ) { $predent = -7; }
      $code_seg->{indent} = $predent;
   } 
   $cutfmt = sprintf( "\%%%ds", $min_indent );
   $cutstr = sprintf( $cutfmt, " " );
   foreach $line( @int_codelines ) {
      $line =~ s/^$cutstr//;
      $new_codetext .= $line . "\n";
   }
   $code_seg->{text} = $new_codetext;
   return;
}

# ***************************************************************
#
#  Subroutine Reindent
#
#  Description:  Normalizes the indentation of a section.  Since
#                indentation may use spaces or tabs, and might be
#                inconsistent, this function looks for changes in
#                indentation and sets a standard tab size, currently
#                hard-coded to 4 spaces.
#                This function also checks sequences of LABELED
#                paragrpahs and gets the maximum length of the LABEL,
#                which is then used to set a second tab following the
#                paragraph LABEL.
#                It also checks for conditions where there are
#                multiple paragraphs associated with the LABEL.  This
#                is done by checking if the row indent is close to
#                (within 4 spaces) of the paragraph indent plus the
#                LABEL length, starting at the 3rd line after the
#                PARAGRAPH label below.
#
# ***************************************************************

sub Reindent {
   my @section = @{$_[0]};

   my $para;
   my $prev_indent  = $section[0]->{indent};
   my $norm_indent  = 0;
   my $prev_type    = $Normal;
   my $prev_tagsep  = $No_tagsep;
   my $pidx         = 0;
   my $startidx;
   my $endidx       = -1;
   my $lidx;
   my $label_len;
   my $maxtag       = 0;
   my $Tabsize      = 4;

   $reindent_cnt++;

   PARAGRAPH:  foreach $para (@section) {
      if( $para->{fmttype} ne $Label and $prev_type eq $Label ) {
         if( $para->{indent} > ( $prev_indent + $maxtag - 4 ) ) {
            $para->{fmttype} = $Label;
            $endidx++;
            $para->{indent} = $norm_indent;
            $pidx++;
            $para->{tagsep} = $prev_tagsep;
            next PARAGRAPH;
         }
         else {
            $norm_indent = ResetLabels( \@section, $startidx, $endidx );
            $endidx = -1;
            $maxtag = 0;
         }
      }

      if( $para->{indent} > $prev_indent ) {
         $norm_indent += $Tabsize;
      } elsif( $para->{indent} < $prev_indent ) {
         $norm_indent -= $Tabsize;
      }
      $prev_indent = $para->{indent};
      $para->{indent} = $norm_indent;

      if( $para->{fmttype} eq $Label ) {
         if( $endidx < 0 ) { $startidx = $pidx; }
         $endidx++;
         if( length( $para->{label} ) > $maxtag ) {
            $maxtag = length( $para->{label} );
         }
         $prev_tagsep = $para->{tagsep};
      } elsif( $para->{fmttype} eq $Code ) {
         IndentCode( $para );
      }
      $prev_type = $para->{fmttype};
      $pidx++;
   }
 
   if( $endidx >= 0 ) { ResetLabels( \@section, $startidx, $endidx ); }
   return;
}

# ***************************************************************
#
#  Subroutine PrintDesc
#
#  Description:  Processes a block of text for a 'man' page section.
#                Uses previously identified paragraph attributes to set
#                indents and line spacing.  For each word, sets bold and
#                italic highlights for function, data type and parameter
#                names.
#
# ***************************************************************

sub PrintDesc {
   my @fmtdesc    = @{$_[0]};
   my $fctname    = $_[1];

   my $paragraph;       # Hash reference to the paragraph.
   my $paratext;
   my $previndent = 0;
   my @words;
   my $word;
   my $filename;
   my $punc;
   my $fmtline;
   my $prevspace  = 1;
   my $linelen    = 0;
   my @decomp;
   my $fmt;
   my $linesep    = "";

   $printdesc_cnt++;

#  Now process each paragraph, defined as a hash including text, indent,
#  line spacing, length and paragraph type.

   Reindent( \@fmtdesc );

   $previndent = $fmtdesc[0]->{indent};

   FMTLINE: foreach $paragraph (@fmtdesc) {

      $paratext = $paragraph->{text};
      $paratext =~ s/\\/\\e/g;

      if( $paragraph->{indent} > $previndent ) {
         printf( MANP "%s.RS %d\n",
                 $linesep, $paragraph->{indent} - $previndent );
         $previndent = $paragraph->{indent};
         $linesep = "";
      } elsif( $paragraph->{indent} < $previndent ) {
         printf( MANP "%s.RS %d\n",
                 $linesep, $paragraph->{indent} - $previndent );
         $previndent = $paragraph->{indent};
         $linesep = "";
      }

      if( $paragraph->{fmttype} eq $Normal ) {
         printf( MANP "%s.PP\n", $linesep );
         $linesep = "";
      } elsif( $paragraph->{fmttype} eq $Label ) {
         printf( MANP "%s.TP %d\n", $linesep, $paragraph->{taglen} + 2 );
         my $fmt;
         my $tagword = $paragraph->{label};
         if( defined( $keylist{$tagword} ) ) { $fmt = $keylist{$tagword}; }
         elsif( defined( $namelist{$fctname}{$tagword} ) ) {
            $fmt = $namelist{$fctname}{$tagword};
         } else { $fmt = $paragraph->{tagfmt}; }
         if( $fmt ne "" ) {
            printf( MANP ".%s ", $fmt );
         }
         if( $paragraph->{label} ne "" ) {
            printf( MANP "%s\n", $paragraph->{label} );
         } else {
            printf( MANP "\\ \n" );
         }
         $linesep = "";
      } elsif( $paragraph->{fmttype} eq $Code ) {
         printf( MANP "\n.PP\n" );
         if( $paragraph->{indent} < 0 ) {
            printf( MANP ".RS %d\n", $paragraph->{indent} );
         }
         printf( MANP ".nf\n" );
         printf( MANP "%s", $paratext );
         printf( MANP ".fi\n" );
         if( $paragraph->{indent} < 0 ) {
            printf( MANP ".RE\n" ); 
         }
         $linesep = "";
         next FMTLINE;
      }

      @words = split( ' ', $paratext );

   #  Here it gets messy.  First, any trailing punctuation has to come off,
   #  because we don't want to highlight that.  Then the remainder needs to
   #  be separated into the alphanumeric parts so each word can be checked to
   #  see if it should be highlighted.  Associated C syntax, like '()', '*'
   #  needs to be highlighted with the word, but other punctuation shouldn't.
   #  Slashes need to separate words so we can deal with things like 'and/or',
   #  but path names need to be treated as an entity, and they should be
   #  highlighted also.

      WORD: for $word ( @words ) {

         if( $word =~ /^\.{0,2}\/\w+/ ) {
            ($filename, $punc) = ($word =~ /^(\.{0,2}\/[\/\w\.]+)(\W*)$/ );
            if( defined( $punc ) and $punc ne "" ) {
               $fmtline = ".IR $filename $punc";
            } else { $fmtline = ".I $filename"; }
         } else {
            @decomp = split( '\b', $word );

            if( $#decomp > 0 ) {
               $fmtline = FormatWord( \@decomp, $fctname, "MAN" );
            }
            else { $fmtline = ""; }
         }

         if( $fmtline ne "" ) {
            printf( MANP "%s%s\n", $linesep, $fmtline );
            $linesep = "";
            $linelen = 0;
            next WORD;
         }
         if( defined( $keylist{$word} ) ) { $fmt = $keylist{$word}; }
         elsif( defined( $namelist{$fctname}{$word} ) ) {
            $fmt = $namelist{$fctname}{$word}; }
         else { $fmt = ""; }
         if( $fmt ne "" ) { 
            printf( MANP "%s.%s %s\n", $linesep, $fmt, $word );
            $linelen = 0;
            $linesep = "";
            next WORD;
         } else {
      #  Print the word.  If the line is filled up, print a newline and
      #  start fresh.  Otherwise, print a space before the word unless it's
      #  the start of a line.
            my $wsep;
            if( $linelen == 0 ) { $wsep = ""; }
            else { $wsep = " "; }
            $linelen += length( $word ) + 1;
            if( $linelen > 70 ) {
               printf( MANP "\n" );
               $linelen = length( $word ) + 1;
               $wsep = "";
            }
            printf( MANP "%s%s", $wsep, $word );
            $linesep = "\n";
         }
      }
      if( $paragraph->{spaces} != $prevspace ) {
         printf( MANP "%s.PD %s\n", $linesep, $paragraph->{spaces} );
         $linesep = "";
         $prevspace = $paragraph->{spaces};
      }
      $linelen = 0;
   }
   printf( MANP "$linesep" );
   return;
}

# ***************************************************************
#
#  Subroutine FormatWord
#
#  Description:  Split a word into alphanumeric and other characters,
#                and construct a line of text to be printed in the man
#                page.  Function names and keywords are highlighted, with
#                associated function syntax '()', '*', '&' highlighted
#                along with the word.
#
# ***************************************************************

sub FormatWord {
   my @decomp    = @{$_[0]};
   my $fctname   = $_[1];
   my $fmtmode   = $_[2];

   my $fmtline   = "";
   my $idx;
   my $subword;
   my $highlight = "";
   my $hlfound = $False;
   my @newlist;
   my @lexelems;
   my @plist;
   my $elem;
   my @fmtlist;
   my @parens;
   my $boldcount = 0;
   my $italcount = 0;
   my $romcount  = 0;
   my $ff;
   my $type       = "";
   my $fmt        = "";
   my $text       = "";
   my $opentype   = 'N';
   my $htype      = '';
   my $prev_htype = '';
   my $newtype    = $False;
   my $lqsub;
   my $rqsub;
   my $qsub;
   my $subparen;
   my %fctkeylist;


   if( $fmtmode eq "MAN" ) {
      $lqsub      = "\\*(lq";
      $rqsub      = "\\*(rq";
   } elsif( $fmtmode eq "HTML" ) {
      $lqsub = "&#8220;";
      $rqsub = "&#8221;";
   } else {
      $lqsub = "\"";
      $rqsub = "\"";
   }
   $qsub = $lqsub;

   $fmtword_cnt++;

   #  Find out if any of the split-out alpha words are keywords, and
   #  if they are, find out what format they use and what position within the
   #  word they are found.

   if( defined( $namelist{$fctname} ) ) {
      %fctkeylist = %{$namelist{$fctname}};
   }

   foreach $subword (@decomp) {
      @parens = [];
      if( defined( $keylist{$subword} ) ) {
         push( @plist, [($subword, "FUNC")] );
         $hlfound = $True;
      } elsif( defined( $fctkeylist{$subword} ) ) {
         push( @plist, [($subword, "PARM")] );
         $hlfound = $True;
      }
      else {
         @lexelems = split( /([\(\)\*\&]+)/, $subword );
         foreach $subparen ( @lexelems ) {
            if( $subparen ne "" ) {
               if( $subparen =~ /\(\)/ )   { @parens = ($subparen, "SPRN" );  }
               elsif( $subparen =~ /\(/ )  { @parens = ($subparen, "OPRN" );  }
               elsif( $subparen =~ /\)/  ) { @parens = ($subparen, "CPRN" );  }
               else                        { @parens = ($subparen, "OTHR" );  }
               push( @plist, [@parens] );
            }
         }
      }
   }
   if( $hlfound eq $False ) { return ""; }
#
#  Now, go through the list of lexical elements and associate the function
#  syntax with associated key words, so parentheses and references ('*', '&')
#  are highlighted with the function or parameter name.
# 
   $type = "";
   $text = "";
   $opentype = 'N';
   $htype = '';
   $prev_htype = '';

   foreach $elem (@plist) {
      $prev_htype = $htype;
      $text = @$elem[0];
      if( $text =~ /\"/ ) {
         $text =~ s/\"/$qsub/g;
         if( $qsub eq $lqsub ) { $qsub = $rqsub; }
         else { $qsub = $lqsub; }
      }
      $type = @$elem[1];
      if   ( $type eq "FUNC" ) {
         $fmt = 'B';
         $htype = 'B';
         if( $opentype eq 'U' ) {
            $opentype = $htype;
            $newtype = $False;
         }
         elsif( $opentype eq $htype ) { $newtype = $False; }
         elsif( $prev_htype eq '' )   { $newtype = $False; }
         elsif( $prev_htype ne $htype ) { $newtype = $True; }
         else { $newtype = $False; }

         if( $newtype eq $True ) {
            push( @newlist, $highlight );
            push( @fmtlist, $prev_htype );
            $highlight = $text;
         }
         else { $highlight = $highlight . $text; }
      }
      elsif( $type eq "PARM" ) {
         $fmt = 'I';
         $htype = 'I';
         if( $opentype eq 'U' ) {
            $opentype = $htype;
            $newtype = $False;
         }
         elsif( $opentype eq $htype ) { $newtype = $False; }
         elsif( $prev_htype eq '' )   { $newtype = $False; }
         elsif( $prev_htype ne $htype ) { $newtype = $True; }
         else { $newtype = $False; }

         if( $newtype eq $True) {
            push( @newlist, $highlight );
            push( @fmtlist, $prev_htype );
            $highlight = $text;
         }
         else { $highlight = $highlight . $text; }
      }
      elsif( $type eq "SPRN" ) {
         $fmt = '';
         $htype = $prev_htype;
         $highlight = $highlight . $text;
         $opentype = 'N';
      }
      elsif( $type eq "OPRN" ) {
         $fmt = '';
         $htype = $prev_htype;
         if( $prev_htype eq '' ) { $opentype = 'U'; }
         else { $opentype = $htype; }
         $highlight = $highlight . $text;
      }
      elsif( $type eq "CPRN" ) {
         $fmt = '';
         if( $opentype ne 'N' ) { $htype = $opentype; }
         else { $htype = $prev_htype; }
         if( $htype eq $prev_htype ) {
            $highlight = $highlight . $text;
         } else {
            push( @newlist, $highlight );
            push( @fmtlist, $prev_htype );
            $highlight = $text;
         }
      }
      elsif( $type eq "OTHR" ) {
         $fmt = 'R';
         $htype = 'R';
         if( $opentype eq 'U' ) {
            $opentype = $htype;
            $newtype = $False;
         }
         elsif( $opentype eq $htype ) { $newtype = $False; }
         elsif( $prev_htype eq '' )   { $newtype = $False; }
         elsif( $prev_htype ne $htype ) { $newtype = $True; }
         else { $newtype = $False; }

         if( $newtype eq $True ) {
            push( @newlist, $highlight );
            push( @fmtlist, $prev_htype );
            $highlight = $text;
         } else { $highlight = $highlight . $text; }
      }
      else { printf( "Hoser mode\n" ); }

   }

   if( $highlight ne "" ) {
       push( @newlist, $highlight );
       push( @fmtlist, $htype );
   }

   $fmtline = ".";
   foreach $ff (@fmtlist) {
      if( $ff eq 'B' ) {
         if( $boldcount == 0 ) { $fmtline = $fmtline . $ff; }
         $boldcount++;
      }
      elsif( $ff eq 'I' ) {
         if( $italcount == 0 ) { $fmtline = $fmtline . $ff; }
         $italcount++;
      }
      elsif( $ff eq 'R' ) {
         if( $romcount == 0 ) { $fmtline = $fmtline . $ff; }
         $romcount++;
      }
      else { printf( "Format Type %s not valid\n", $ff ); }
   }

   if( $boldcount > 0 and $italcount > 0 and $romcount > 0 ) {
      $fmtline = "";
      if( $#fmtlist != $#newlist ) {
         printf( "Fmts: %d, Items: %d\n", $#fmtlist, $#newlist ); }
      for $idx ( 0 .. $#fmtlist ) {
         $fmtline = $fmtline . "\\f" . $fmtlist[$idx] . $newlist[$idx];
      }
   }
   elsif( $boldcount == 0 and $italcount == 0 ) { $fmtline = ""; }
   else {
      my  $hl;

      foreach $hl (@newlist) {
         $fmtline = $fmtline . ' ' . $hl;
      }
   }

   return $fmtline;

}

# ***************************************************************
#
#  Subroutine HashKeys
#
#  Description:  This procedure puts all the names into two hash tables,
#                one for function names and data types, that will be
#                printed in bold, and one for parameter names which will
#                be printed in italics.  Actually, in 'man' they'll just
#                be printed in different colors, but in an HTML conversion
#                they'll have the groff meaning.  The hash key is the name
#                of the item, and the value is the print type bold or italic.
#                A hash is used so I don't have to do lookup loops through a
#                list on every word in the input, but rather can go
#                straight to a hash table.  The hash table is globally
#                available.
#
# ***************************************************************

sub HashKeys {

   my %fctnames;

   my $temp;
   my $type;
   my $stgclass;
   my $idx;
   my $name;
   my @typelist;
   my @tmplist;
   my $fct;

   $hashkeys_cnt++;

   FCT: for $fct (keys %FuncList ) {
      %fctnames = ();
      if( !defined( $keylist{$fct} ) ) {
         $keylist{$fct} = "B";
      }
      $temp = $FuncList{$fct};
      if( !defined( $temp->{typ} ) ) { next FCT; }
      $type = $temp->{typ};
      $type =~ s/ *\*$//;
      if( !defined( $keylist{$type} ) ) { $keylist{$type} = "B"; }

      if( $temp->{stg} ne "" ) {
         $stgclass = $temp->{stg};
         if( !defined( $keylist{$stgclass} ) ) {
            $keylist{$stgclass} = "B";
         }
      }
      @tmplist = @{ $temp->{params} };
      for $idx (0 .. $#tmplist) {
         $type = $tmplist[$idx]->{ptype};
         @typelist = split( '\b', $type ); 
         foreach $type (@typelist ) {
            if( $type =~ /\w+/ ) {
               if( !defined( $keylist{$type} ) ) {
                  $keylist{$type} = "B";
               }
            }
         }
         if( $tmplist[$idx]->{pname} ne "" ) {
            my @nametoklist;

            $name = $tmplist[$idx]->{pname};
            @nametoklist = split( '\b', $name );
            foreach $name( @nametoklist ) {
               if( $name =~ /\w+/ ) {
                  if( !defined( $fctnames{$name} ) ) { $fctnames{$name} = "I"; }
               }
            }
         }
      }
      $namelist{$fct} = { %fctnames };
   }
   return;
}

# ***************************************************************
#
#  Subroutine UpdateAuthor
#
#  Description:  Saves the author information into the run-time
#                configuration data structure.
#
# ***************************************************************

sub UpdateAuthor {
   my $author_text = $_[0];
   my $author_desc = $_[1];

   my $manrc_text   = "";
   my $para_mark    = $False;
   my $line;
   my %linedesc   = ();

   my @para = split( "\n", $author_text );

   InitDesc( \%linedesc );

   MANRC_LINE: foreach $line ( @para ) {
      if( $line =~ /^\s*$/ ) {
         $linedesc{spaces} = 1;
         $para_mark = $True;
         next MANRC_LINE;
      }
#  Start a new paragraph by storing the last one and initializaing the
#  collector element 'linedesc'.
      if( $para_mark eq $True ) {
         $linedesc{linelen} = length( $linedesc{text} );
         $linedesc{fmttype} = $Normal;
         if( $linedesc{text} ne "" ) { push( @$author_desc, { %linedesc } ); }
         InitDesc( \%linedesc );
         $para_mark = $False;
      }

      if( $linedesc{text} ne "" ) {
         $linedesc{text} = $linedesc{text} . ' ' . $line;
      } else { $linedesc{text} = $line; }
   }

   $linedesc{spaces} = 1;
   $linedesc{linelen} = length( $linedesc{text} );
   $linedesc{fmttype} = $Normal;
   if( $linedesc{text} ne "" ) { push( @$author_desc, { %linedesc } ); }

   return;
}

# ***************************************************************
#
#  Subroutine EditFuncList
#
#  Description:  Calls the edit function for each source code function
#                when running without the GUI.
#
# ***************************************************************

sub EditFuncList {
   my  $fct;
   my  $tagwd;
   my  $fct_stg;

   $mkdoccmd_cnt++;

   for $fct ( keys %FuncList ) {
      $fct_stg = $FuncList{$fct}->{stg};
      if( $fct_stg ne "static" ) { $fct_stg = "public"; }
      if( $StorageMode{$fct_stg} eq $True ) {
         EditFile( $fct, $EditPage, $FuncList{$fct}->{srcfile} ); }

      if( $Verbose eq $True ) {
         for $tagwd (keys %keylist) {
            printf( "Keylist: (%s): %s\n", $tagwd, $keylist{$tagwd}  );
         }
      }
   }
   return;
}

# ***************************************************************
#
#  Subroutine ReadRcFile
#
#  Description:  Reads the process initialization file which contains
#                author and header information.  Stores the data
#                in a file-scoped hash.
#
#  Supporting Functions:
#                ReadCfgAuthor
#                ReadCfgHeader
#
# ***************************************************************

sub ReadRcFile {

   my $manrc_line;
   my $manrc_sect;
   my @manrc_para;
   my %manrc_header = ();
   my $line;
   my $value;
   my $expanded_value;
   my @valuelist;
   my $lclinitfile;

   $RcSettings{author}     = $False;
   $RcSettings{source}     = $False;
   $RcSettings{manual}     = $False;
   $RcSettings{section}    = $False;
   $RcSettings{srcdir}     = $False;
   $RcSettings{tagfile}    = $False;
   $RcSettings{mandir}     = $False;
   $RcSettings{htmldir}    = $False;
   $RcSettings{indexfile}  = $False;
   $RcSettings{fmtsel}     = $False;
   $RcSettings{stgmode}    = $False;
   $RcSettings{cmtflag}    = $False;
   $RcSettings{updatemode} = $False;

   $lclinitfile = $InitFile;
   TildeSub( \$lclinitfile );
   if( $lclinitfile eq "" ) { return; }

   if( ! -e $lclinitfile ) {
      PrintStat( "Initialization file $lclinitfile not found\n", 'warning' );
      return;
   }

   $manrc_desc{author}  = ();  # an array of paragraph descriptors
   $manrc_desc{header}  = ();  # a hash of header data

   $manrc_header{manual}  = "";
   $manrc_header{source}  = "";
   $manrc_header{section} = 3;

   open( MANRC, $lclinitfile ) or do {
                 PrintStat( "Couldn't open $lclinitfile", 'warning' );
                 return;
               };

   MANRC_LINE: while( <MANRC> ) {
      chop;
      $line = $_;
      if( $line =~ /<Author>/ ) {
         ReadCfgAuthor( \*MANRC, \@manrc_para );
         $RcSettings{author} = $True;
         next MANRC_LINE;
      } elsif( $line =~ /<ManHeader/ ) {
         ReadCfgHeader( \*MANRC, \%manrc_header );
         next MANRC_LINE;
      } elsif( $line =~ /<SrcDir/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $expanded_value = $value;
         TildeSub( \$expanded_value );
         if( -e $expanded_value ) {
            $SrcDir = $expanded_value;
            $DispSrcDir = $value;
            $RcSettings{srcdir} = $True;
         }
         next MANRC_LINE;
      } elsif( $line =~ /<ManDir/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $expanded_value = $value;
         TildeSub( \$expanded_value );
         if( -e $expanded_value ) {
            $ManDir = $expanded_value;
            $DispManDir = $value;
            $RcSettings{mandir} = $True;
         } else {
            PrintStat( "Reading $lclinitfile, Man Directory $value not found",
                       'warning' );
         }
         next MANRC_LINE;
      } elsif( $line =~ /<HtmlDir/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $expanded_value = $value;
         TildeSub( \$expanded_value );
         if( -e $expanded_value ) {
            $HtmlDir = $expanded_value;
            $DispHtmlDir = $value;
            $RcSettings{htmldir} = $True;
         } else {
            PrintStat( "Reading $lclinitfile, HTML Directory $value not found",
                       'warning' );
         }
         next MANRC_LINE;
      } elsif( $line =~ /<TagFile/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $expanded_value = $value;
         TildeSub( \$expanded_value );
         if( -e $expanded_value ) {
            $TagFile = $expanded_value;
            $DispTagFile = $value;
            $RcSettings{tagfile} = $True;
         }
         next MANRC_LINE;
      } elsif( $line =~ /<IndexFile/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $expanded_value = $value;
         TildeSub( \$expanded_value );
         $IndexFile = $expanded_value;
         $DispIndexFile = $value;
         $RcSettings{indexfile} = $True;
         next MANRC_LINE;
      } elsif( $line =~ /<PageFormat/ ) {
         my $fmtspec;
         $ManFormat = $False;
         $HtmlFormat = $False;
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         (@valuelist) = split( /[, ]+/, $value );
         foreach $value (@valuelist) {
            if( uc($value) eq "MAN" )  { $ManFormat = $True;  }
            if( uc($value) eq "HTML" ) { $HtmlFormat = $True; }
         }
         $RcSettings{fmtsel} = $True;
      } elsif( $line =~ /<StorageMode/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         (@valuelist) = split( /[, ]+/, $value );
         $StorageMode{$StgPublic} = $False;
         $StorageMode{$StgStatic} = $False;
         foreach $value (@valuelist) {
            if( lc($value) eq $StgPublic ) { $StorageMode{public} = $True; }
            if( lc($value) eq $StgStatic ) { $StorageMode{static} = $True; }
            if( lc($value) eq $StgAll    ) { $StorageMode{public} = $True;
                                             $StorageMode{static} = $True; }
         }
         $RcSettings{stgmode} = $True;
      } elsif( $line =~ /<Comment/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $value = uc( $value );
         if( $value =~ /($HdrCmt)|($IntCmt)|($AllCmt)/ ) {
            $CmtFlag = $value;
            $RcSettings{cmtflag} = $True;
         }
      } elsif( $line =~ /<UpdateMode/ ) {
         ($value) = ( $line =~ /.*\"(.*)\".*/ );
         $value = uc( $value );
         if( $value =~ /(SRC)|(EDIT)|(NEW)/ ) {
            $ReComment = $value;
            $RcSettings{updatemode} = $True;
         }
      }
   }

   $manrc_desc{author} = [ @manrc_para ];
   $manrc_desc{header} = { %manrc_header };

   close( MANRC );

   return;
}

# ***************************************************************
#
#  Subroutine SetDefaultCfg
#
#  Description:  If no configuration or preference file is found, set
#                default vaules for author name, man section, source and
#                manual title.  The author is obtained from the
#                environment 'USER'; section 3 and generic title
#                information is set.
#
# ***************************************************************

sub SetDefaultCfg {
   my %manrc_header = ();
   my %linedesc;
   my @author_desc  = ();
   my $passwdent;
   my $username = $ENV{USER};
   my @pwstuff = getpwnam($username);
   my $author_name = $pwstuff[6];

   $manrc_desc{author}  = ();
   $manrc_desc{header}  = ();

   UpdateAuthor( $author_name, \@author_desc );

   $manrc_header{manual}  = "Programmer's Manual";
   $manrc_header{source}  = "";
   $manrc_header{section} = 3;

   $manrc_desc{author} = [ @author_desc ];
   $manrc_desc{header} = { %manrc_header };

   $RcStartDir = $ENV{HOME};

   return;
}

# ***************************************************************
#
#  Subroutine ReadCfgHeader
#
#  Description:  Reads man page header data from the configuration file.
#
# ***************************************************************

sub ReadCfgHeader {
   my $fh  = $_[0];
   my $header = $_[1];

   my $line;
   my $item;

   HDR_LINE: while( <$fh> ) {
      chomp;
      $line = $_;

      if( $line =~ /MANUAL/ ) {
         ($item) = ($line =~ /\"(.*)\"/ );
         $header->{manual} = $item;
         $RcSettings{manual} = $True;
      } elsif( $line =~ /SOURCE/ ) {
         ($item) = ($line =~ /\"(.*)\"/ );
         $header->{source} = $item;
         $RcSettings{source} = $True;
      } elsif( $line =~ /SECTION/ ) {
         ($item) = ($line =~ /\=\s*\"*(\d*)\"*/ );
         $header->{section} = $item;
         $RcSettings{section} = $True;
      } elsif( $line =~ /<\/ManHeader>/ ) {
         last HDR_LINE;
      }
   }

   return;
}

# ***************************************************************
#
#  Subroutine ReadCfgAuthor
#
#  Description:  Reads author information from the preference configuration
#                file and stores it in an array of paragraph descriptor
#                hashes as used for collecting man page text.
#
# ***************************************************************

sub ReadCfgAuthor {
   my $fh = $_[0];
   my $author_desc = $_[1];

   my $manrc_text   = "";
   my $para_mark    = $False;
   my $line;
   my %linedesc   = ();

   InitDesc( \%linedesc );

   MANRC_LINE: while( <$fh> ) {
      chomp;
      $line = $_;
      $line =~ s/^\s*//;

      if( $line =~ /<\/Author>/ ) {
         last MANRC_LINE;
      }
      if( $line =~ /^\s*$/ ) {
         $linedesc{spaces} = 1;
         $para_mark = $True;
         next MANRC_LINE;
      }
#  Start a new paragraph by storing the last one and initializaing the
#  collector element 'linedesc'.
      if( $para_mark eq $True ) {
         $linedesc{linelen} = length( $linedesc{text} );
         $linedesc{fmttype} = $Normal;
         if( $linedesc{text} ne "" ) { push( @$author_desc, { %linedesc } ); }
         InitDesc( \%linedesc );
         $para_mark = $False;
      }

      if( $linedesc{text} ne "" ) {
         $linedesc{text} = $linedesc{text} . ' ' . $line;
      } else { $linedesc{text} = $line; }
   }

   $linedesc{spaces} = 1;
   $linedesc{linelen} = length( $linedesc{text} );
   $linedesc{fmttype} = $Normal;
   if( $linedesc{text} ne "" ) { push( @$author_desc, { %linedesc } ); }

   return;
}

# ***************************************************************
#
#  Subroutine InitDesc
#
#  Description:  Initializes the text paragraph data structure for each
#                paragraph, creating the hash fields and setting default
#                values.
#
# ***************************************************************

sub InitDesc {
   my $vhash = $_[0];

   $initdesc_cnt++;

   $vhash->{text}    = "";
   $vhash->{linelen} = 0;
   $vhash->{spaces}  = 0;
   $vhash->{indent}  = 0;
   $vhash->{taglen}  = 0;
   $vhash->{label}   = "";
   $vhash->{fmttype} = $Undef;
   $vhash->{tagfmt}  = "";
   $vhash->{tagsep}  = $No_tagsep;

   return;
}

# ***************************************************************
#
#  Subroutine PrintLineDesc
#
#  Description:  Prints a paragraph descriptor to stdout for debugging.
#
# ***************************************************************

sub PrintLineDesc {
   my $prnline = $_[0];
   printf( "Text:    %s\n", $prnline->{text}    );
   printf( "Length:  %d\n", $prnline->{linelen} );
   printf( "Spaces:  %d\n", $prnline->{spaces}  );
   printf( "Indent:  %d\n", $prnline->{indent}  );
   printf( "Tag Len: %d\n", $prnline->{taglen}  );
   printf( "Label:   %s\n", $prnline->{label}   );
   printf( "Format:  %s\n", $prnline->{fmttype} );
   printf( "Tag Fmt: %s\n", $prnline->{tagfmt}  );
   return;
}

# ***************************************************************
#
#  Subroutine InitFunc
#
#  Description:  Initializes the data structures for a function,
#                including all the arrays of paragraph descriptors,
#                to NULL values.
#
# ***************************************************************

sub InitFunc {
   my  $funcproto    = shift;
   my  $brief        = shift;
   my  $retdesc      = shift;
   my  $desc         = shift;
   my  $bugs         = shift;
   my  $refs         = shift;
   my  $credits      = shift;
   my  $author       = shift;
   my  $example      = shift;
   my  $functionlist = shift;
   my  $nameline     = shift;

   $initfunc_cnt++;

   $$nameline               = 0;
   $$funcproto              = "";
   $$brief                  = "";
   @$retdesc                = ();
   @$desc                   = ();
   @$bugs                   = ();
   @$refs                   = ();
   @$credits                = ();
   @$author                 = ();
   @$example                = ();
   @$functionlist           = ();

   @lcl_includes            = ();
   @sys_includes            = ();
   return;
}

# ***************************************************************
#
#  Subroutine SaveFunc
#
#  Description:  When a function segment is complete, adds the data
#                elements to the hash for the function.  First clears
#                any values, then adds the collected data to the hash.
#                This way assures that every element has a value even
#                if it's NULL.
#
# ***************************************************************

sub SaveFunc {
   my $funcproto = $_[0];
   my $line      = $_[1];
   my @pvt_sys_hdrs = @{$_[2]};
   my @pvt_lcl_hdrs = @{$_[3]};
   my @retdesc      = @{$_[4]};
   my @desc         = @{$_[5]};
   my @bugs         = @{$_[6]};
   my @refs         = @{$_[7]};
   my @credits      = @{$_[8]};
   my @author       = @{$_[9]};
   my @example      = @{$_[10]};
   my @functionlist = @{$_[11]};
   my $fctname      = $_[12];
   my $brief        = $_[13];
   my $srcfile      = $_[14];

   my $newfunc;

   $savefunc_cnt++;
   $newfunc = ();
   $newfunc->{datafmt}     = $ProcFmt;
   $newfunc->{description} = ();
   $newfunc->{summary}     = ();
   $newfunc->{ret_desc}    = ();
   $newfunc->{bug_desc}    = ();
   $newfunc->{sysincl}     = ();
   $newfunc->{lclincl}     = ();
   $newfunc->{params}      = ();
   $newfunc->{stg}         = "";
   $newfunc->{refs}        = ();
   $newfunc->{example}     = ();
   $newfunc->{credits}     = ();
   $newfunc->{author}      = ();
   $newfunc->{funclist}    = ();
   $newfunc->{srcfile}     = $srcfile;
   $newfunc->{intcmt}      = ();

   $FuncList{$fctname}     = $newfunc;
   $newfunc->{summary}     = $brief;
   if( $funcproto ne "" ) {
      ParseProto( $funcproto, $fctname, $newfunc );
   }
   $newfunc->{sysincl}     = [ @pvt_sys_hdrs ];
   $newfunc->{lclincl}     = [ @pvt_lcl_hdrs ];
   $newfunc->{ret_desc}    = [ @retdesc      ];
   $newfunc->{description} = [ @desc         ];
   $newfunc->{bug_desc}    = [ @bugs         ];
   $newfunc->{refs}        = [ @refs         ];
   $newfunc->{credits}     = [ @credits      ];
   $newfunc->{author}      = [ @author       ];
   $newfunc->{example}     = [ @example      ];
   $newfunc->{funclist}    = [ @functionlist ];
   if( $line =~ /^<\/SUBSYSTEM>/ ) { $newfunc->{blktype} = "SUBS"; }
   else { $newfunc->{blktype} = "SUBR"; }
   return;
}

# ***************************************************************
#
#  Subroutine GetLabel
#
#  Description:  Extracts single-word paragraph headings, or labels,
#                from the source text.  Looks for:
#                   Ellipsis ( ... )
#                   Hyphen
#                   Number followed by '.' or ')'
#                   Colon followed by white space
#                   Tab
#                following the first word of the line.  Measures the
#                length of the label and puts the paragraph text into
#                the 'text' field of the paragraph descriptor.
#
# ***************************************************************

sub GetLabel {
   my $labelline = shift;
   my $linedesc  = shift;

   my $tag;
   my $rest;
   my $tagsep;

   $getlabel_cnt++;

   if( $$labelline =~ /^\S+\s+\.\.+\W*\s+/ ) {
      ($tag, $rest) = ( $$labelline =~ /^(\S+)\s+\.\.+\s*(.*)\s*$/ );
      $tagsep = $Ellipsis;
   } elsif( $$labelline =~ /^\S+\s+\-\s+/ ) {
      ($tag, $rest) = ( $$labelline =~ /^(\S+)\s+\-\s+(.*)/ );
      $tagsep = $Hyphen;
   } elsif( $$labelline =~ /^\s*\d+[\)\.]\s+/ ) {
      ($tag, $rest) = ( $$labelline =~ /^\s*(\d+[\)\.])\s+(.*)\s*$/ );
      $tagsep = $Spacechar;
   } elsif( $$labelline =~ /^\s*\S+:\s+/ ) {
      ($tag, $rest) = ( $$labelline =~ /^\s*(\S+):\s*(.*)\s*$/ );
      $tagsep = $Colonchar;
   } elsif( $$labelline =~ /^\S+\t/ ) {
      ($tag, $rest) = ( $$labelline =~ /^(\S+)\s+(.*)\s*$/ );
      $tagsep = $Tabchar;
   } else {
      $tag = "";
      $rest = $$labelline;
      $tagsep = $No_tagsep;
      printf( "GetLabel() Error: Didn't find any label in %s\n", $$labelline );
   }

   $linedesc->{label} = $tag;
   $linedesc->{taglen} = length( $tag );
   $linedesc->{tagsep} = $tagsep;
   $$labelline = $rest;
   return;
}

# ***************************************************************
#
#  Subroutine ReadTextFmt
#
#  Description:  Top-level routine for processing the original text
#                from a man page-like format.  Finds function description
#                sections, and the 'man' page sections for each function,
#                such as 'Description', 'Return Values', etc.  Puts the
#                function descriptions into a hash table by function
#                name.  Does the first pass of fomatting, stripping out
#                extra white space, and keeping track of line spacing and
#                indentation.
#
# ***************************************************************

sub ReadTextFmt {
   my  $textfile = shift;
   my  $srcfile  = shift;

   my  $brief;
   my  $fctname;
   my  $section = "";
   my  $para    = $False;
   my  $codefrag   = $False;
   my  $shortline  = $False;

   my  $indent     = "";
   my  $idnt_size  = -1;
   my  $prev_idnt  = 0;
   my  $indnt_cnt  = 0;
   my  $orig_idnt  = 0;
   my  $no_func    = $True;
   my  $idx;
   my  %linedesc   = ();
   my  $line       = "";
   my  $text       = "";
   my  $funcproto  = "";
   my  $contline;
   my  $nameline   = 0; 
   my  $linekey;
   my  $linelen;
   my  $firstpass  = $False;

   my  $textdesc;
   my  $newdesc;
   my  @desc;
   my  @intdesc;
   my  @retdesc;
   my  @bugs;
   my  @refs;
   my  @credits;
   my  @author;
   my  @example;
   my  @functionlist;

   $readtext_cnt++;

   if( $#GblFuncnames < 0 ) { $firstpass = $True; }

#  Read the text file into the data structures defined in the
#  other sections.
   open( TEXTFILE, "<$textfile" ) or
               die "Couldn't open text file $textfile\n";

   InitDesc( \%linedesc );

   LINE: while( <TEXTFILE> ) {
      chomp;
      $line = $_;

#  If a line is empty, set a $para flag so that the next text line starts
#  a new paragraph.  This filters multiple blank lines to a single blank.
      if( $line =~ /^\s*$/ and $codefrag eq $False ) {
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }

#  Start a new paragraph by storing the last one and initializaing the
#  collector element 'linedesc'.
      if( $para eq $True ) {
         if( $linedesc{text} ne "" ) { push( @$textdesc, { %linedesc } ); }
         InitDesc( \%linedesc );
         $textdesc = $newdesc;
         $para = $False;
      }

#  If at the end of a function section, put all the collected information into
#  the function data structure.
      if( $line =~ /^<\/SUBROUTINE>/ or $line =~ /^<\/SUBSYSTEM>/ ) {
         $linedesc{spaces} = 1;
         if( $linedesc{text} ne "" ) { push( @$textdesc, { %linedesc } ); }
         InitDesc( \%linedesc );

         SaveFunc( $funcproto, $line, \@sys_includes,
                   \@lcl_includes, \@retdesc, \@desc, \@bugs, \@refs,
                   \@credits, \@author, \@example, \@functionlist, $fctname,
                   $brief, $srcfile );

         if( $firstpass eq $True ) {
            push( @GblFuncnames, $fctname );
         }
         $section = "";
         $no_func = $True;
         next LINE;
      }

      ($linekey) = ($line =~ /(^\w*).*/);

#  If at the start of a function section, initialize the collection data
#  structures.
#  By default, start with DESCRIPTION.  Anything not under any other
#  heading will go here.

      if( ( $line =~ /^<SUBROUTINE/ ) or ( $line =~ /^<SUBSYSTEM/ ) ) {
         if( $line =~ /^<SUBSYSTEM/ ) {
            ($fctname) = ( $line =~ /<SUBSYSTEM.*\"(.*)\"/ );
         } elsif(  $line =~ /^<SUBROUTINE/ ) {
            ($fctname) = ( $line =~ /<SUBROUTINE.*\"(.*)\"/ );
         }
         InitFunc( \$funcproto,
                   \$brief, \@retdesc, \@desc, \@bugs, \@refs, \@credits,
                   \@author, \@example, \@functionlist, \$nameline );
         $no_func = $False;
         $section = $Description;
         $newdesc = \@desc;
         $textdesc = \@desc;
         next LINE;
      }
      if( $no_func eq $True ) { next LINE; }
#  Handle code fragments as raw text.  Look for '<CODE>' to identify code.
#  The indent gets set to the last value so that when the code segment ends,
#  the indent comparison is done to the last non-code paragraph.
      if( $line =~ /<CODE>/ ) {
         $codefrag = $True;
         $linedesc{fmttype} = $Code;
         $linedesc{indent}  = $idnt_size;
         next LINE;
      }

#  First see if the line starts a new function section by looking for a
#  keyword at the start of a line.
#  NAME starts a new section, so initialize all the function data structures.
#  They get filled in as the function section is processed.
      if( $linekey eq $Name ) {
         $section = $Name;
         $newdesc = [];
         next LINE;
      }
          # One of my first users had this misspelling, so I let it be.
      if( $linekey eq "SYNOPSYS" or $linekey eq $Synopsis ) {
         $section = $Synopsis;
         $newdesc = [];
         next LINE;
      }
      if( $linekey eq $Return ) {
         $section = $Return;
         $newdesc = \@retdesc;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Description ) {
         $section = $Description;
         $newdesc = \@desc;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Bugs ) {
         $section = $Bugs;
         $newdesc = \@bugs;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq "SEE" ) {
         $section = $See_Also;
         $newdesc = \@refs;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Credits ) {
         $section = $Credits;
         $newdesc = \@credits;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Example ) {
         $section = $Example;
         $newdesc = \@example;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Functions ) {
         $section = $Functions;
         $newdesc = \@functionlist;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }
      if( $linekey eq $Author ) {
         $section = $Author;
         $newdesc = \@author;
         $linedesc{spaces} = 1;
         $para = $True;
         next LINE;
      }

#  OK, it wasn't a keyword.  The section was saved the last time a section
#  keyword was seen, so process the line according to the section we're in.

      if( $section eq $Name ) {
         if( $nameline++ == 0 ) {
            if( $line =~ /\-.+/ ) {
               ($fctname, $brief) = ( $line =~ /\s*(\w*)\s*\-\s*(.*$)/ );
            } else {
               ($fctname) = ( $line =~ /\s*(\w*).*/ ); 
               $brief = "";
            }
         } else {
            ($contline) = ( $line =~ /^\s*(.*$)/ );
            if( defined( $brief ) and $brief ne "" ) {
               $brief = $brief . ' ' . $contline;
            }
         }
         next LINE;
      }
      if( $section eq $Synopsis ) {
         my  @srcline;

         @srcline = split( ' ', $line );
         if( $#srcline >= 0 && $srcline[0] eq "#include" ) {
            if( $srcline[1] =~ /\<.*\>/ ) {
               push( @sys_includes, $srcline[1] );
            }
            elsif( $srcline[1] =~ /\".*\"/ ) {
               push( @lcl_includes, $srcline[1] );
            }
         }
         else {
            $line =~ s/^\s*//;
            $line =~ s/\s*$//;
            $funcproto = $funcproto . ' ' . $line;
         }
         next LINE;
      }

#  Now strip off leading white space, put the text into the {text}
#  segment, and the indent, converted to spaces, in the {indent} element.
#  If it's known that the line is a part of the paragraph, add the text to
#  the existing text.  Otherwise, start a new text segment.

      if( $codefrag ne $True ) {
         ($indent, $text) = ($line =~ /(^\s*)(.*)\s*$/ );
         $indent =~ s/\t/$TabStr/g;
         $prev_idnt = $idnt_size;
         $idnt_size = length( $indent );

   #  See if one of the paragraph label syntaxes is present, and if so,
   #  set the paragraph type to LABELED and start a new paragraph.
   #  The "Word: " form of paragraph labels only applies to the first
   #  line of a paragraph, while others can be used to define a new
   #  paragraph.  If it's a quoted string, don't do the LABEL checks.

         if( ( $text =~ /^\S+\s+\.\.+\W*\s+/ ) or
             ( $text =~ /^\S+\s+\-\s+/ )       or
             ( $text =~ /^\s*\d+[\)\.]\s+/ )   or
             ( $text =~ /^\S+\t/ )             or
             ( !( $text =~ /^\".*\"/ ) and
               $text =~ /^\s*\S+:\s+/  and
               ( $prev_idnt != $idnt_size or $linedesc{text} eq "" ) ) )
         {
            if( $linedesc{text} ne "" ) {
               push( @$textdesc, { %linedesc } );
               InitDesc( \%linedesc );
            }
            GetLabel( \$text, \%linedesc );
            $linedesc{fmttype} = $Label;
            $linedesc{indent} = $idnt_size;
         } elsif( $text =~ /.*\.{2,}\s*\w+/ ) {
   #  If it isn't labeled, but there is an ellipsis, it may be an index
   #  or table of contents line.  That gets processed here as a labeled
   #  paragraph.
            my $idxline;
            my $idxlabel;
            my $elp_len;
            my $i;

            if( $idnt_size < 8 ) { $idnt_size = 8; }

            ($idxline, $idxlabel) = split( /\s*\.{2,}\s*/, $text, 2 );
            $linedesc{fmttype} = $Label;
            $linedesc{indent}  = $idnt_size;
            $linedesc{label}   = $idxline . ' ';

            $elp_len = 64 - ( length( $idxline ) + length( $idxlabel ) +
                              $idnt_size + 2 );

            for $i (0 .. $elp_len ) {
               $linedesc{label} .= '.';
            }

            $linedesc{taglen}  = 64 - ( length( $idxlabel ) + $idnt_size );
            $linedesc{tagsep} = $No_tagsep;

            $text              = $idxlabel;
            $para              = $True;
         }

   #  Get the length of the line, and if the line is less than 'Minline'
   #  (50 characters, currently defined), start a new paragraph.

         $linelen = length( $text );
         if( $linedesc{fmttype} ne $Label and $linelen < $Minline ) {
            $para = $True;
         }
         $linedesc{linelen} = $linedesc{linelen} + $linelen;

   #  If the indent is different than the last line, and it's not already
   #  identified as a labeled paragraph, close the last paragraph and start
   #  a new one.
   #  An additional check of the previous line is done to see if there's
   #  a label of the form 'word word ... :  text'.  It'll be declared a
   #  label if the indent on the current line is about the same as the
   #  position of the colon.

         if( $prev_idnt != $idnt_size and $linedesc{fmttype} ne $Label ) {
            if( $linedesc{text} ne "" ) {
               my $relabel = $False;
               if( $linedesc{text} =~ /:\s+\w+/ ) {
                  my $pretext;
                  my $posttext;
                  my $preindent;
                  ($pretext, $posttext) = ( $linedesc{text} =~ /(.*):\s+(.*)/ );
                  $preindent = length( $pretext ) + $linedesc{indent};
                  if( $preindent > $idnt_size - 4 and
                      $preindent < $idnt_size + 4 ) {
                     $linedesc{fmttype} = $Label;
                     $linedesc{tagsep}  = $Colonchar;
                     $linedesc{label}   = $pretext;
                     $linedesc{text}    = $posttext . ' ' . $text;
                     $linedesc{taglen}  = length( $pretext );
                     $linedesc{linelen} = length( $linedesc{text} );
                     $para = $False;
                     $relabel = $True;
                  }
               }
               if( $relabel eq $False ) {
                  push( @$textdesc, { %linedesc } );
                  InitDesc( \%linedesc );
               }
            }
            if( $linedesc{fmttype} ne $Label ) {
               $linedesc{text} = $text;
               $linedesc{indent} = $idnt_size;
            }
         } else {
      #  It wasn't an indent change, (unless it's a LABELED paragraph), so
      #   add the text to the paragraph.  If it's not LABELED, set the
      #   indent size to this line's indent.

            $linedesc{text} = $linedesc{text} . ' ' . $text;
            if( $linedesc{fmttype} ne $Label ) {
               $linedesc{indent} = $idnt_size;
            }
         }
      #  If we got here without the format type being set, default it to
      #  NORMAL.

         if( $linedesc{fmttype} eq $Undef ) { $linedesc{fmttype} = $Normal; }
      } else {
   #  It's a CODE fragment, so preserve line spacing and indentation.
         if( $line =~ /<\/CODE>/ ) {
            $codefrag = $False;
            $para = $True;
         } elsif( $linedesc{text} ne "" ) {
            $linedesc{text} = $linedesc{text} . "\n" . $line;
         } else { $linedesc{text} = $line; }
      }
   }
   if( $no_func eq $False ) {
      SaveFunc( $funcproto, $line, \@sys_includes,
                \@lcl_includes, \@retdesc, \@desc, \@bugs, \@refs,
                \@credits, \@author, \@example, \@functionlist, $fctname,
                $brief, $srcfile );
      PrintStat( "Function $fctname malformed; attempt recovery", 'warning' );
   }
   close( TEXTFILE );

   return;
}

# ***************************************************************
#
#  Subroutine ParseProto
#
#  Description:  Parses function prototypes into the function 
#                name, parameters, and return type.
#
# ***************************************************************

sub ParseProto {
   my $proto = $_[0];
   my $name  = $_[1];
   my $funcdef = $_[2];

   my @prmList = ( );
   my @paramlist;
   my %prmElem;
   my $prmName;
   my $prmType;
   my $params;
   my $param;
   my @partlist;
   my $tok;
   my @decltoks;
   my $decl;

   my $ftype = "";

   $parseproto_cnt++;

#  The word immediately before the first open paren is the function name.
#  The word(s) preceding the name are the return type, and if the first word
#  is extern or static, those are the storage class.
#  After the open paren, the last word before a comma is the parameter name,
#  and tokens preceding the name are the parameter type.  The last parameter
#  is followed by a close paren instead of a comma.

#  Split the function prototype at 'funcname ('.  Preceding the name is
#  the return type and maybe a storage class (static, extern).

   ($decl, $params) = split( /$name\s*\(/, $proto, 2 );
   @decltoks = split( ' ', $decl );

#  First process the return type; it may be multiple tokens, and there may
#  be a storage class at the start.  If there are multiple tokens, concatenate
#  them into a space-separated entity.

   foreach $tok (@decltoks) {
      if( $tok eq "static" or $tok eq "extern" ) {
         $funcdef->{stg} = $tok;
      } else {
         if( $ftype ne "" ) { $ftype = $ftype . ' ' . $tok; }
         else { $ftype = $tok; }
         $funcdef->{stg} = "";
      }
   }
   $funcdef->{typ} = $ftype;

#  Lop off any trailing punctuation, then make an array of the parameters
#  and find the type and name.  First separate parameters at the commas,
#  then for each parameter, split into space-separated tokens.  The last
#  one is the parameter name, and all the others are the type specification.
#  The type is concatenated into a space-separated string.
#  The name and type are put into a two-element hash, and each of these
#  parameter hashes are put into an array.  When done, the array is assigned
#  to the function hash.
#  For function pointer parameters, the type is the part before the function
#  construct (*func)(...).  In this context, the 'name' parameter is the
#  function definition.

   @paramlist = ( );
   $params =~ s/\)\s*\;*$//;
   @partlist  = split( /(\w+\s*\(.*\)\s*\(.*\))/, $params );
   foreach $param (@partlist) {
      my @parms;
      my $pp;

      if( !( $param =~ /\(.*\)/ ) ) {
         $param =~ s/^\s+//;
         $param =~ s/\s+$//;
         $param =~ s/\s+/ /g;
         @parms = split( /,/, $param );
         foreach $pp (@parms) {
            if( $pp ne "" ) { push( @paramlist, $pp ); }
         }
      }
      else { push( @paramlist, $param ); }
   }
   
   foreach $param (@paramlist) {
      my @paramtoks;
      my $idx;

      $prmName = "";
      $prmType = "";
      if( $param =~ /\(\*.*\)\s*\(.*\)/ ) {
         ($prmType, $prmName) = ($param =~ /(\w*)\s*(\(.*\))/ );
      } else {
         @paramtoks = split( ' ', $param );
         if( $#paramtoks < 0 ) {
            printf( "\tNo params\n" );
         } elsif( $#paramtoks == 0 ) {
            $prmType = $paramtoks[0];
         } else {
            $prmType = $paramtoks[0];
            for $idx (1 .. $#paramtoks - 1) {
               $prmType = $prmType . ' ' . $paramtoks[$idx];
            }
            $prmName = $paramtoks[$#paramtoks];
         }
      }
      $prmElem{ptype} = $prmType;
      $prmElem{pname} = $prmName;
      push @prmList, { %prmElem };
         
   }
   $funcdef->{params} = [ @prmList ];

   return;
}

# ***************************************************************
#
#  Subroutine PrintExecStats
#
#  Description:  Used for printing the count of the number of times each
#                function was called.  This is executed when Verbose
#                mode is selected from the command line.
#
# ***************************************************************

sub PrintExecStats {

   printf( "Execution Counts:\n" );
   printf( "\tPrintMan:\t%5d\n",       $printman_cnt     );
   printf( "\tPrintDesc:\t%5d\n",      $printdesc_cnt    );
   printf( "\tFormatWord:\t%5d\n",     $fmtword_cnt      );
   printf( "\tHashKeys:\t%5d\n",       $hashkeys_cnt     );
   printf( "\tParseProto:\t%5d\n",     $parseproto_cnt   );
   printf( "\tReadTextFmt:\t%5d\n",    $readtext_cnt     );
   printf( "\tReadDump:\t%5d\n",       $readdump_cnt     );
   printf( "\tProcessCmt:\t%5d\n",     $processcmt_cnt   );
   printf( "\tReadSource:\t%5d\n",     $readsrc_cnt      );
   printf( "\tEditFile:\t%5d\n",       $editfile_cnt     );
   printf( "\tWriteDesc:\t%5d\n",      $writedesc_cnt    );
   printf( "\tFormatRefs:\t%5d\n",     $fmtrefs_cnt      );
   printf( "\tResetLabels:\t%5d\n",    $resetlabels_cnt  );
   printf( "\tIndentCode:\t%5d\n",     $indentcode_cnt   );
   printf( "\tReindent:\t%5d\n",       $reindent_cnt     );
   printf( "\tDisplayUserIf:\t%5d\n",  $mkdocgui_cnt     );
   printf( "\tGetSelection:\t\t%5d\n", $psel_cnt         );
   printf( "\tEditFuncList:\t%5d\n",   $mkdoccmd_cnt     );
   printf( "\tInitDesc:\t%5d\n",       $initdesc_cnt     );
   printf( "\tInitFunc:\t%5d\n",       $initfunc_cnt     );
   printf( "\tSaveFunc:\t%5d\n",       $savefunc_cnt     );
   printf( "\tGetLabel:\t%5d\n",       $getlabel_cnt     );

   return;
}

# ***************************************************************
#
#  Subroutine ExtractSrc
#
#  Description:  For each source code file specified on the command line
#                or selected from the user interface, extract documentation
#                from the the source code file.  The C pre-compiler is
#                executed, and then 'srcextract' is run on the pre-compiler
#                output.  The results are written to intermediate files
#                which are then read into hash data structures.  The
#                source code is read directly to extract the names of
#                include files and pre-compiler directives, and this
#                information is added to the hashes.  A pass through the
#                extracted data is made to get a list of key words,
#                including data types and variable names.  The collected
#                data is used throughout the program to format and edit
#                the documentation to generate man pages.  Some checks
#                are made to see if a source file has already been
#                processed and is current with respect to the source.
#
# ***************************************************************

sub ExtractSrc {
   my $srcfile;
   my $dmpfile;
   my $cppfile;
   my $cmdstr;

   if( $#SrcFiles >= 0 ) {
      SRCFILE: foreach $srcfile (@SrcFiles) {
         TildeSub( \$srcfile );
         if( -e $srcfile ) {
            ($dmpfile) = ( $srcfile =~ /(.*)(\..*)/ );
            $dmpfile = $dmpfile . ".dmp";
            $cppfile = $dmpfile;
            $cppfile =~ s/\.dmp/\.cpp/;
            $dmpfile =~ s/.*\///;
            $dmpfile = $TmpDir . '/' . $dmpfile;
            if( !(-e $dmpfile) or ( -C $srcfile lt -C $dmpfile ) ) {
               if( $Verbose eq $True ) {
                  printf( "Parsing the source code file $srcfile\n" );
               }
               $cmdstr = "/usr/bin/gcc -E -C -DLINUX -DI586 $Cflags $srcfile >"
                         . "$cppfile";
               system( $cmdstr );
               $cmdstr = "srcextract -f $cppfile -o $dmpfile";
               system( $cmdstr );
               if( $? != 0 ) {
                  PrintStat( "Parser failed: file $srcfile: Error $?",
                             'warning' );
                  if( -e $dmpfile ) { unlink $dmpfile; }
                  next SRCFILE;
               } else { unlink $cppfile; }
            }
            else {
               if( $Verbose eq $True ) {
                  printf( "Srcdoc dump file $dmpfile is up to date\n" );
               }
            }

            ReadSource( $srcfile );
            ReadDump( $dmpfile, $srcfile );
            HashKeys;
         }
      }
   }
   return;
}

sub sd_findINC
{
 my $file = join('/',@_);
 my $dir;
 $file  =~ s,::,/,g;
 foreach $dir (@INC)
  {
   my $path;
   return $path if (-e ($path = "$dir/$file"));
  }
 return undef;
}

# ***************************************************************
#
#  Main script body
#
# ***************************************************************

#  Set values used as constants.

$False        = "FALSE";
$True         = "TRUE";
$Undef        = "UNDEF";
$Edit         = "EDIT";
$Generate     = "GEN";
$Label        = "LABELED";
$Normal       = "NORMAL";
$Bold         = "BOLD";
$Code         = "CODE";
$Name         = "NAME";
$Synopsis     = "SYNOPSIS";
$Return       = "RETURN";
$Description  = "DESCRIPTION";
$Bugs         = "BUGS";
$See_Also     = "SEE_ALSO";
$Credits      = "CREDITS";
$Example      = "EXAMPLE";
$Functions    = "FUNCTIONS";
$Author       = "AUTHOR";
$RawFmt       = "RAW";
$ProcFmt      = "PROC";
$View         = "VIEW";
$Editman      = "EMAN";
$Minline      = 50;
$Ellipsis     = ".";
$Hyphen       = "-";
$Spacechar    = " ";
$Colonchar    = ":"; 
$Tabchar      = "\t";
$No_tagsep    = "";
$AllCmt       = "ALL";
$HdrCmt       = "HDR";
$IntCmt       = "INT";
$SrcCmtAlways = "SRC";
$NewCmtAlways = "NEW";
$EdtCmtAlways = "EDIT";
$StgPublic    = "public";
$StgStatic    = "static";
$StgAll       = "all";

#  Initialize global variables

$ManDir        = ".";
$DispManDir    = "";
$HtmlDir       = ".";
$DispHtmlDir   = "";
$SrcDir        = ".";
$DispSrcDir    = "";
$TmpDir        = ".";
$TagFile       = "";
$DispTagFile   = "";
$IndexFile     = "";
$DispIndexFile = "";
$InitFile      = "";
$DispInitFile  = "";
@SrcFiles      = ();
$Verbose       = $False;
$TextFmt       = $False;
$Cflags        = "";
$TextFile      = "";
$SrcFile       = "";
$TabFmt        = "";
$TabStr        = "";
$TabSize       = 8;
$OptList       = "";
$RcStartDir    = $ENV{HOME};
$DisplayOpen   = $False;
@GblFuncnames  = ();
%FuncList      = ();
$EditPage      = $False;
$Verbose       = $False;
$CmtFlag       = $HdrCmt;
$UseGui        = $True;
$TextFmt       = $False;
$ReComment     = $EdtCmtAlways;
$ManFormat     = $True;
$HtmlFormat    = $False;
$StgModes      = "";
$StorageMode{$StgPublic} = $True;
$StorageMode{$StgStatic} = $False;

#
#  Process command line options.  Precedence is to use command line setting
#  if present, if not, use the initialization file, and any items not set,
#  use the default settings.
#

$OptList = 'hzwvnc:d:i:D:m:t:s:u:';

getopts( $OptList );

if( $opt_h ) { Usage; $opt_h = ""; exit( 0 );         }
if( $opt_z ) { CmInfo; $opt_z = ""; exit( 0 );        }
if( $opt_i ) { $InitFile = $opt_i;                    }

if( $#ARGV >= 0 ) { push( @SrcFiles, @ARGV ); }
$SrcDir = qx/pwd/;
chomp $SrcDir;

$TabFmt = sprintf( "\%%%ds", $TabSize );
$TabStr = sprintf( $TabFmt, " " );


if( $InitFile eq "" or ( ! -e $InitFile ) ) {
   my $alt_rcfile   = "\.srcdocrc";
   my $alt_rcpath   = $ENV{HOME};
   if( -e "$alt_rcpath/$alt_rcfile" ) {
      $InitFile = "$alt_rcpath/$alt_rcfile";
      $RcStartDir = $alt_rcpath;
   } elsif( -e "./$alt_rcfile" ) {
      $InitFile = "./$alt_rcfile";
      $RcStartDir = qx/pwd/;
      $RcStartDir =~ s/\s*$//;
   } else { SetDefaultCfg; }
} else {
   if( $InitFile =~ /^\s*\~/ ) {
      $InitFile =~ s/\~/$ENV{HOME}/;
   } elsif( $InitFile !~ /^\// ) {
      my $initdir = qx/pwd/;
      chomp $initdir;
      $InitFile = $initdir . '/' . $InitFile;
   }
}

ReadRcFile;

if( $opt_c ) { $CmtFlag = $opt_c;                     }
if( $opt_d ) { $TextFmt = $True; $TextFile = $opt_d;  }
if( $opt_D ) { $Cflags = $opt_D;                      }
if( $opt_n ) { $UseGui = $False;                      }
if( $opt_m ) { $StgModes = $opt_m;                    }
if( $opt_s ) { $TabSize = $opt_s;                     }
if( $opt_t ) { $TagFile = $opt_t;                     }
if( $opt_u ) { $ReComment = $opt_u;                   }
if( $opt_v ) { $Verbose = $True; $opt_v = "";         }
if( $opt_w ) { $EditPage = $Edit;                     }

if( $UseGui eq $True ) { $EditPage = $Edit; }

if( $CmtFlag =~ /^[Aa]/ ) {
   $CmtFlag = $AllCmt;
} elsif( $CmtFlag =~ /^[Hh]/ ) {
   $CmtFlag = $HdrCmt;
}

if( $ReComment =~ /^[Ss]/ ) {
   $ReComment = $SrcCmtAlways;
} elsif ( $ReComment =~ /^[Ee]/ ) {
   $ReComment = $EdtCmtAlways;
} elsif( $ReComment =~ /^[Nn]/ ) {
   $ReComment = $NewCmtAlways;
} else {
   $ReComment = $EdtCmtAlways;
   PrintStat( "Invalid Update Option $ReComment; default to $EdtCmtAlways\n",
              'advise'
            );
}

if( $StgModes ne "" ) {
   if(    $StgModes =~ /^[Pp]/ ) { $StgModes = $StgPublic; }
   elsif( $StgModes =~ /^[Ss]/ ) { $StgModes = $StgStatic; }
   elsif( $StgModes =~ /^[Aa]/ ) { $StgModes = $StgAll;    }
   else { PrintStat( "Invalid Storage Option $StgModes; using $StgPublic",
                     'advise' );
          $StgModes = $StgPublic;
   }
   if( $StgModes eq $StgAll ) {
      $StorageMode{$StgPublic} = $True;
      $StorageMode{$StgStatic} = $True;
   } else {
      $StorageMode{$StgPublic} = $False;
      $StorageMode{$StgStatic} = $False;
      $StorageMode{$StgModes} = $True;
   }
}

if( $ENV{TMPDIR} ) {
   $TmpDir = $ENV{TMPDIR};
} elsif( -e "./tmp" ) {
   $TmpDir = "./tmp";
} elsif( -w "/tmp" ) {
   if( ! -d "/tmp/srcdoc" ) {
      system( "mkdir /tmp/srcdoc" );
   }
   $TmpDir = "/tmp/srcdoc";
} else { $TmpDir = "."; }

if( ! -e $TmpDir ) { $TmpDir = "."; }

if( $Verbose eq $True ) { printf( "Temp Directory set to $TmpDir\n" ); }

#  For each source code file, call the C preprocessor to create a temporary
#  file containing function descriptions.  Then call routines to read the
#  source code extract, edit the man page sections and print the man page.

if( $TextFmt eq $True ) {
   my $ff;

   $SrcFile = $TextFile;
   ReadTextFmt( $TextFile, $TextFile );
   HashKeys;
   if( $UseGui eq $True ) {
      $DisplayOpen = $True;
      DisplayUserIf;
   }
   foreach $ff (keys %FuncList) {
      if( $ManFormat eq $True )  { PrintMan( $ff );     }
      if( $HtmlFormat eq $True ) { PrintHtmlMan( $ff ); }
   }
   if( $Verbose eq $True ) {
      PrintExecStats;
   }
   exit( 0 );
}

ExtractSrc;

if( $UseGui eq $True ) {
   $DisplayOpen = $True;
   DisplayUserIf;
}
else {
   EditFuncList;
}

if( $Verbose eq $True ) {
   PrintExecStats;
}

# END OF LEGITIMATE SECTION
