#if( defined( SOLARIS ) )
static char   ident[] __attribute__((unused))
         = "%Z%%M%  control version %I%  %G% %U%";
#elif( defined( LINUX ) )
static char   ident[] __attribute__((unused))
         = "srcdocActions.c,v control version 1.6 2002/01/05 03:02:54";
#endif
static char   cpyrght[] __attribute__((unused))
         = "Copyright (c) 2000, Gordon B. Aaseng";
static char   usage[]  __attribute__((unused))
         = "Parser actions for srcdoc";

/*
 * copyright (c) 2000, 2001, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

/*
 *
 * FILE NAME
 *
 * srcdocActions.c
 *
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 *
 * FILE DESCRIPTION
 *
 * This file contains the parser action functions for the 'srcdoc'
 * parser.  The yacc rules invokes these functions for the pertinent
 * constructs found in the source code; the constructs of interest
 * are those that define functions.
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * srcdocActions.c,v
 * Revision 1.6  2002/01/05 03:02:54  gordon
 * Changed license to GPL.
 *
 * Revision 1.5  2001/12/18 13:12:59  gordon
 * Split comment collection data structures into header and function
 * internal types, and added allocation, copying and freeing processing.
 *
 * Revision 1.4  2001/02/21 12:25:37  gordon
 * Moved the freeDecl function to public scope; minor cleanup and added
 * checks in a couple functions also.
 *
 * Revision 1.3  2001/02/13 12:12:03  gordon
 * Added functions for K&R function declarations; added more print functions;
 * added some memory freeing functions; changed locally used functions to
 * static; and some miscellaneous cleanup.
 *
 * Revision 1.2  2000/08/01 05:32:07  gordon
 * Changed names from mkdoc to srcdoc, including file and variable names
 *
 * Revision 1.1  2000/07/24 12:13:31  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */


#define   _POSIX_SOURCE     1       /* posix.1             */
#ifndef   _POSIX_C_SOURCE
#define   _POSIX_C_SOURCE   199309  /* posix.1 and posix.4 */
#endif

/* ************************************************************************* */
/*                                                                           */
/* Included Files                                                            */
/*                                                                           */
/* ************************************************************************* */

/*
 * os specific header files.
 */
/*  (None.)  */

/*
 * standard c (ANSI) header files
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * posix header files
 */
/*  (None.)  */

/*
 * project specific header files.
 */

#include "srcdocLex.h"
#include "srcdocActions.h"

/* ************************************************************************* */
/*                                                                           */
/* External References                                                       */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

#define printd(x)  printf( "%s - %s\n", __FUNCTION__, (x) )

/* ************************************************************************* */
/*                                                                           */
/* Data Structures                                                           */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Public Global Variables                                                   */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Private (static) Global Variables                                         */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (Locally Used Functions)                                 */
/*                                                                           */
/* ************************************************************************* */


/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

static parmListType *transformKRSpec( funcDefType * );
static void          printKRFunc( typeSpecType *,
                                  declDefType  *,
                                  funcDefType  *
                                );
static void          printDecl( FILE *, declDefType * );
static void          freeTokList( tokListType * );

/**************************************************************************
 *  Private Function transformKRSpec
 *
 *  DESCRIPTION
 *     Converts proto-function definitions collected for K&R-type function
 *     declaration parameters into a parameter list for the function.
 *
 *     Memory is allocated for the returned parameter list.  The user is
 *     responsible for memory management and freeing the memory.
 *
 *  RETURN VALUES
 *     A pointer to a parmListType parameter list.
 *
 * ************************************************************************* */

static parmListType *transformKRSpec( funcDefType *kparms )
{
   parmListType    *parmtype;
   parmListType    *parmdecl;
   parmListType    *parm;
   parmListType    *parmlist;
   funcDefType     *fptr;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif

   fptr = kparms;
   parmlist = (parmListType *)NULL;

   while( fptr )
   {
      parmtype = createParmList( createDecl( fptr->spec->type ), SCALAR );
      parmdecl = createParmList( fptr->func, fptr->func->declClass );
      parm = addParmList( parmtype, parmdecl );


      parmlist = addParmList( parmlist, parm );
      fptr = fptr->funcNext;
   }
   return( parmlist );
}

/**************************************************************************
 *  Private Function freeTokList
 *
 *  DESCRIPTION
 *     Frees memory used by a token list.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void freeTokList( tokListType *junktok )
{
   tokListType  *ptr;
   tokListType  *old;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   ptr = junktok;
   while( ptr )
   {
      if( ptr->tok )
         free( ptr->tok );
      old = ptr;
      ptr = ptr->tokNext;
      free( old );
   }
   return;
}

/**************************************************************************
 *  Private Function printKRFunc
 *
 *  DESCRIPTION
 *     This function was used in testing the parser constructs for K&R
 *     function declarations.  It's left in the source code for additional
 *     testing if needed.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void printKRFunc( typeSpecType *kspec,
                         declDefType  *kdecl,
                         funcDefType  *kfunc
                       )
{
   funcDefType   *kptr = kfunc;
   typeSpecType  *kpspec;
   declDefType   *kpfunc;

   if( kspec )
   {
      printf( "K&R Function typespec\n" );
      if( kspec->type )    printValList( stdout, kspec->type );
      if( kspec->storage ) printValList( stdout, kspec->storage );
   }
   if( kdecl )
   {
      printf( "K&R Function Declaration\n" );
      printDecl( stdout, kdecl );
   }

   if( kptr ) printf( "K&R Function Parameters\n" );
   while( kptr )
   {
      kpspec = kptr->spec;
      kpfunc = kptr->func;

      if( kpspec->type )
      {
         printf( "K&R parameter spec->type\n" );
         printValList( stdout, kpspec->type );
      }
      if( kpspec->storage )
      {
         printf( "K&R parameter spec->storage\n" );
         printValList( stdout, kpspec->storage );
      }
      if( kpfunc )
      {
         printf( "K&R parameter decl\n" );
         printDecl( stdout, kpfunc );
      }
      kptr = kptr->funcNext;
   }
   return;
}

/**************************************************************************
 *  Private Function printDecl
 *
 *  DESCRIPTION
 *     Print a declaration definition data structure, consisting of a 
 *     class specification, a token list and a parameter list.  The token
 *     list is the function name, and the parameter list is the function
 *     parameters.  Each parameter consists of the data type, name, and address
 *     or pointer references, and these are printed for each parameter by
 *     the printParmList() function.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void printDecl( FILE *ffp, declDefType *declaration )
{
#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   if( declaration->decl ) {
      printValList( ffp, declaration->decl );
   }
   if( declaration->parms ) {
      printParmList( ffp, declaration->parms );
   }
   return;
}

/* ---------------------------------------------------------------------- */

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (External Interface Functions)                           */
/*                                                                           */
/* ************************************************************************* */

/**************************************************************************
 * Public Function addTypeDef
 *
 *  DESCRIPTION
 *     Add an identifier to the list of type definitions so that the
 *     lexical analyzer can determine if a token is a variable identifier
 *     or a type.  The input is a declaration definition, which could be a
 *     list of tokens (to handle multi-word names, like 'struct mytype'.
 *     The type definition will always be the last token in the list, and
 *     for type definitions, it should always be a single token.  The
 *     C grammar definition, though, uses the same rule for other things.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void addTypeDef( declDefType *idVal )
{
   tokListType   *ptr;
   char          *tok;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   if( idVal == NULL )
      return;

   ptr = idVal->decl;
   while( ptr )
   {
      tok = ptr->tok;
      ptr = ptr->tokNext;
   }
   typeList[numTypeDefs] = tok;
   if( gblVbs == 1 )
      printf( "PARSE: Add Type %d = %s\n", numTypeDefs, tok);
   numTypeDefs++;
   return;
}

/**************************************************************************
 *  Public Function addTokList
 *
 *  DESCRIPTION
 *     The addTokList() function adds tok to an existing token list toklist.
 *     If pos is 0, adds to the head of the list, otherwise adds to the end
 *     of toklist.
 *
 *     This function call createTokList which allocates memory, so the caller
 *     is responsible for freeing the memory when done.
 *
 *  RETURN VALUES
 *     A pointer to the head of the resultant list.
 *
 *     If the new list is null (it shouldn't be, but ...) return the existing
 *     list.
 *
 * ************************************************************************* */

tokListType *addTokList( char *tok, tokListType *toklist, int pos )
{
   tokListType      *ptr;
   tokListType      *prevptr;
   tokListType      *new;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   new = createTokList( tok );
   if( pos == 0 ) /* add to head of list */
   {
      new->tokNext = toklist;
      return( new );
   }
   else /* add to tail */
   {
      ptr = toklist;
      prevptr = toklist;
      while( ptr )
      {
         prevptr = ptr;
         ptr = ptr->tokNext;
      }
      if( prevptr )
         prevptr->tokNext = new;
      else
         toklist = new;

      return( toklist );
   }
}

/**************************************************************************
 *  Public Function createTokList
 *
 *  DESCRIPTION
 *     Create a tokListType structure, which is a linked list element
 *     consisting of a character string and a pointer to the next element.
 *     There is also a pointer level element, that is incremented for each
 *     pointer reference to the token.  Allocates memory for the new list
 *     element, setting the tokNext pointer to NULL and the pointer level to 0.
 *
 *     Note that memory is allocated, so function users need to free the
 *     memory when done.
 *
 *  RETURN VALUES
 *     A pointer to the created tokListType structure.
 *
 * ************************************************************************* */

tokListType *createTokList( char *tok )
{
   tokListType    *new;
#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
   printf( "\t%s: tok %s\n", __FUNCTION__, tok );
#endif

   new = (tokListType *)malloc( sizeof( tokListType ) );
   if( new == (tokListType *)NULL )
      return( new );
   new->tok = tok;
   new->ptrLvl = 0;
   new->tokNext = (tokListType *)NULL;
   return( new );
}

/**************************************************************************
 *  Public Function concatTokList
 *
 *  DESCRIPTION
 *     The concatTokList() function joins two token lists together.  The new
 *     list can be added to the existing toklist, either at the end or the head,
 *     as specified by pos.  If pos is 0, adds to the head of tlist,
 *     otherwise adds new to the end of tlist.
 *
 *  RETURN VALUES
 *     A pointer to the head of the resultant list.
 *
 * ************************************************************************* */

tokListType *concatTokList( tokListType *tlist,
                            tokListType *new,
                            int          pos
                          )
{
   tokListType *head;
   tokListType *ptr;
   tokListType *prevptr;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif

   switch( pos )
   {
      case 0 :
         ptr     = new;
         prevptr = new;
         head    = new;
         while( ptr )
         {
            prevptr = ptr;
            ptr = ptr->tokNext;
         }
         if( prevptr )
            prevptr->tokNext = tlist;
         else
            head = tlist;
         break;
      case 1 :
         ptr     = tlist;
         prevptr = tlist;
         head    = tlist;
         while( ptr )
         {
            prevptr = ptr;
            ptr = ptr->tokNext;
         }
         if( prevptr )
            prevptr->tokNext = new;
         else
            head = new;
         break;
      default :
         fprintf( stderr, "concatTokList: Invalid concat type %d\n", pos );
         head = tlist;
         break;
   }
   return( head );
}

/**************************************************************************
 *  Public Function concatVal
 *
 *  DESCRIPTION
 *     Concatenates two character strings into a single string with a space
 *     separating the two words.  Allocates memory for the new string, copies
 *     the two elements into the new string and frees the memory from the
 *     original strings tok1 and tok2.
 *
 *     This function is used by the parser to handle named structures and
 *     enumerations as a single token.
 *
 *     Memory is allocated by the function.  Users are responsible for freeing
 *     it when complete.
 *
 *  RETURN VALUES
 *     A pointer to the new concatenated token.
 *     Returns NULL if the malloc failed.
 *
 * ************************************************************************* */

char *concatVal( char *tok1, char *tok2 )
{
   char       *new;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   new = (char *)malloc( strlen( tok1 ) + strlen( tok2 ) + 2 );
   if( new == (char *)NULL )
      return( new );
   strcpy( new, tok1 );
   strcat( new, " " );
   strcat( new, tok2 );
   free( tok1 );
   free( tok2 );
   return( new );
}

/**************************************************************************
 *  Public Function createParm
 *
 *  DESCRIPTION
 *     Create a parameter, which is a union of a tokListType or a
 *     parmListType.  Passes the type in as a void pointer, and parmClass 
 *     specifies which type is used.  The parameter is cast to the correct
 *     type and is assigned to the union parameter.
 *
 *     Note that memory is allocated and users are responsible for freeing it.
 *
 *  RETURN VALUES
 *     A pointer to a parmType union, or NULL if the malloc fails.
 *
 * ************************************************************************* */

parmType *createParm( void *parm, parmClassType parmClass )
{
   parmType      *newParm;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   newParm = (parmType *)malloc( sizeof( parmType ) );
   if( newParm == (parmType *)NULL )
      return( newParm );
   switch( parmClass )
   {
      case SCALAR :
         newParm->scalarParm = (tokListType *)parm;
         break;
      case FUNC :
         newParm->funcParm   = (declDefType *)parm;
         break;
      default :
         newParm->scalarParm = (tokListType *)NULL;
         printf( "createParm: Invalid parmClass %d\n", parmClass );
         break;
   }
   return( newParm );
}

/**************************************************************************
 *  Public Function createParmList
 *
 *  DESCRIPTION
 *     Create an element of a parameter linked list from parameter parm.
 *     Allocates memory for a parmListType element, assigns parm to the
 *     element, and sets the parmNext pointer to NULL.
 *
 *     Note that memory is allocated and users are responsible for freeing it.
 *
 *  RETURN VALUES
 *     A pointer to a parmListType structure, or NULL if the malloc fails.
 *
 * ************************************************************************* */

parmListType *createParmList( declDefType *parm, parmClassType class )
{
   parmListType     *new;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   new = (parmListType *)malloc( sizeof( parmListType ) );
   if( new == (parmListType *)NULL )
      return( new );
   new->parmClass        = class;
   switch( class )
   {
      case SCALAR :
         new->parm.scalarParm  = parm->decl;
         break;
      case FUNC :
         new->parm.funcParm    = parm;
         break;
      default :
         new->parm.scalarParm  = (tokListType *)NULL;
         printf( "createParmList: Invalid parmClass %d\n", class );
         break;
   }
      
   new->parmNext = NULL;
   return( new );
}

/**************************************************************************
 *  Public Function addParmList
 *
 *  DESCRIPTION
 *     Adds a parmListType list segment to an existing parameter list.
 *     It adds the new parameter list to the end of the existing list.
 *
 *  RETURN VALUES
 *     A pointer to the head of the original list.  If the existing list
 *     is NULL, returns the new list pointer.
 *
 * ************************************************************************* */

parmListType *addParmList( parmListType *list, parmListType *new )
{
   parmListType    *ptr = list;
   parmListType    *prevptr = list;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif

   while( ptr )
   {
      prevptr = ptr;
      ptr = ptr->parmNext;
   }
   if( prevptr )
      prevptr->parmNext = new;
   else
      list = new;
   return( list );
}

/**************************************************************************
 *  Public Function createDecl
 *
 *  DESCRIPTION
 *     Create a function declaration data structure.  The structure consists
 *     of a name and parameter list.  This function initializes the data
 *     structure with the character string tok, converted to a tokListType.
 *     the parameter list is initialized to NULL.  The addToDecl() function
 *     is used to fill in other elements of the structure as the information
 *     is obtained by the parser.
 *
 *     This function allocates memory for the declaration data structure.
 *     Users are responsible for freeing the allocated memory.
 *
 *  RETURN VALUES
 *     A pointer to the created declaration data structure.
 *
 * ************************************************************************* */

declDefType *createDecl( tokListType *tok )
{
   declDefType    *new;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   new = (declDefType *)malloc( sizeof( declDefType ) );

   new->decl      = tok;
   new->declClass = SCALAR;
   new->parms     = (parmListType *)NULL;

   return( new );
}

/**************************************************************************
 *  Public Function addToDecl
 *
 *  DESCRIPTION
 *     Adds parameter list parms to the declaration definition decl.
 *
 *  RETURN VALUES
 *     A pointer to the declaration definition passed in as parameter decl.
 *
 * ************************************************************************* */

declDefType *addToDecl( declDefType *decl, parmListType *parms )
{
#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   decl->parms = addParmList( decl->parms, parms );
   return( decl );
}

/**************************************************************************
 *  Public Function createFunc
 *
 *  DESCRIPTION
 *     Create a function definition data structure.  Initialize the type
 *     specification and declaration.  Function comments that were collected
 *     by the lexical analyzer are attached to the data structure.  The
 *     function definition is an element of a linked list, and createFunc()
 *     initializes the list pointer to NULL.
 *
 *     This function allocates memory for the function definition data
 *     structure, so calling routines must free the memory when complete.
 *
 *  RETURN VALUES
 *     A pointer to the created function definition data structure.
 *
 * ************************************************************************* */

funcDefType *createFunc( typeSpecType *spec,
                         declDefType  *decl,
                         funcDefType  *KRFuncDef
                       )
{
   funcDefType     *funcspec;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   funcspec = (funcDefType *)malloc( sizeof( funcDefType ) );
   if( strlen( hdrCommentText ) > 0 )
   {
      funcspec->hdrComment = (char *)malloc( strlen( hdrCommentText ) + 1 );
      strcpy( funcspec->hdrComment, hdrCommentText );
      hdrCommentText[0] = '\0';
   }
   else
      funcspec->hdrComment = (char *)NULL;  
   intCommentText[intCmtIdx] = '\0';
   if( strlen( intCommentText ) > 0 )
   {
      funcspec->intComment = (char *)malloc( strlen( intCommentText ) + 1 );
      strcpy( funcspec->intComment, intCommentText );
      intCommentText[0] = '\0';
   }
   else
      funcspec->intComment = (char *)NULL;  
   hdrCmtIdx = 0;
   intCmtIdx = 0;

   funcspec->spec = spec;
   funcspec->func = decl;
   funcspec->funcNext = NULL;

   if( KRFuncDef )
   {
      int  lclVbs;

      funcspec->func->parms = transformKRSpec( KRFuncDef );
   /* These two lines are left for additional debugging of the K&R function
      processing.  Change the lclVbs flag to print the values of the data
      structures used in K&R functions.
   */
      lclVbs = 0;
      if( lclVbs > 0 ) printKRFunc( spec, decl, KRFuncDef );
   }
   
   return( funcspec );
}

/**************************************************************************
 *  Public Function addFunc
 *
 *  DESCRIPTION
 *     Adds a function definition to the global function list gblFuncList.
 *     If function new is NULL (it shouldn't be, but ...), return NULL.
 *     If gblFuncList is NULL, make new be the head of gblFuncList.
 *     Otherwise, add new to the end of gblFuncList.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void addFunc( funcDefType *new )
{
   funcDefType     *ptr;
   funcDefType     *prev;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   if( new == (funcDefType *)NULL )
      return;

   ptr = gblFuncList;
   if( ptr == NULL )
   {
      gblFuncList = new;
      return;
   }

   while( ptr )
   {
      prev = ptr;
      ptr = ptr->funcNext;
   } 
   prev->funcNext = new;
   return;
}

/**************************************************************************
 *  Public Function createTypeSpec
 *
 *  DESCRIPTION
 *     Create a type specification structure which consists of a storage
 *     class and a data type.  Storage class can be static, extern, etc.
 *
 *     The parameters are an identifier tok, which is converted to a
 *     tokListType, and a specType element indicating if the element is a
 *     storage class or a data type.
 *
 *     The function allocates memory for a typeSpecType structure, so calling
 *     routines are responsible for managing and freeing memory.
 *
 *  RETURN VALUES
 *     A pointer to the created typeSpecType data structure.
 *
 * ************************************************************************* */

typeSpecType *createTypeSpec( char *tok, specType t )
{
   typeSpecType     *spec;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
   printf( "\tToken: %s, %s\n", tok, t == STG ? "STG" : "TYP" );
#endif
   spec = (typeSpecType *)malloc( sizeof( typeSpecType ) );

   spec->storage = (tokListType *)NULL;
   spec->type = (tokListType *)NULL;
   
   switch( t )
   {
      case STG:
         spec->storage = createTokList( tok );
         break;
      case TYP:
         spec->type = createTokList( tok );
         break;
      default:
         break;
   }
   return( spec );
}

/**************************************************************************
 *  Public Function addTypeSpec
 *
 *  DESCRIPTION
 *     Adds information to an existing typeSpecType data structure.  This
 *     completes the data structure initialized by createTypeSpec.
 *     The string tok is the name to be added, spec is the existing data
 *     structure, and t specifies if the new token is a storage class or
 *     type specifier.
 *
 *  RETURN VALUES
 *     A pointer to the typeSpecType parameter spec.
 *
 * ************************************************************************* */

typeSpecType *addTypeSpec( char *tok, typeSpecType *spec, specType t )
{
#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
   printf( "\tToken: %s, %s\n", tok, t == STG ? "STG" : "TYP" );
#endif
   switch( t )
   {
      case STG:
         spec->storage = addTokList( tok, spec->storage, 0 );
         break;
      case TYP:
         spec->type = addTokList( tok, spec->type, 0 );
         break;
      default:
         break;
   }
   return( spec ); 
}

/**************************************************************************
 *  Public Function concatDecl
 *
 *  DESCRIPTION
 *     This function is used in K&R declarations to concatenate two
 *     declarations by grouping the declaration (a token list) and the
 *     parameter ilst.  This is used in the rule
 *        'init_declarator_list ',' init_declarator'
 *     However, simple test functions haven't ever triggered this rule so
 *     this is so far a placeholder so that it will do something possibly
 *     sane if this rule gets executed.  It may be that this doesn't ever
 *     get used for a function definition, as I haven't analyzed the grammar
 *     enough.
 *
 *  RETURN VALUES
 *     A pointer to a declDefType, which will be the first non-NULL parameter
 *     passed to the function.
 *
 * ************************************************************************* */

declDefType *concatDecl( declDefType *decl, declDefType *add )
{
   
#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
   printDecl( stdout, decl );
   printDecl( stdout, add );
#endif

   if( decl == (declDefType *)NULL )
      return( add );

   if( decl->declClass != add->declClass )
   {
      printf( "%s-W-declaration:  declaration class doesn't match, "
              "unknown implications (%d <-> %d)\n",
              __FUNCTION__, decl->declClass, add->declClass
            );
   }
   decl->decl = concatTokList( decl->decl, add->decl, 1 );
   decl->parms = addParmList( decl->parms, add->parms );
   return( decl );
}

/**************************************************************************
 *  Public Function createKRFunc
 *
 *  DESCRIPTION
 *     Uses the function definition structures to hold the K&R type
 *     parameter declarations.  When the function has been fully parsed,
 *     the parameter delarations are returned as part of the function body
 *     and then transformed into the parameter declarations.  It's sort of
 *     a confusing anomaly of the old-style C syntax.
 *
 *     This function allocates memory.  The user of the function is
 *     responsible for memory management and freeing of the used memory.
 *
 *  RETURN VALUES
 *     A pointer to a funcDefType that is created by the function, containing
 *     the type spec and declaration spec passed to the function.
 *
 * ************************************************************************* */

funcDefType *createKRFunc( typeSpecType *spec, declDefType *decl )
{
   funcDefType     *func;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   func = (funcDefType *)malloc( sizeof( funcDefType ) );
   func->hdrComment = (char *)NULL;
   func->intComment = (char *)NULL;
   func->spec       = spec;
   func->func       = decl;
   func->funcNext   = NULL;
   return( func );
}


/**************************************************************************
 *  Public Function addKRFunc
 *
 *  DESCRIPTION
 *     Create a list of proto-function defs for K&R declarations.  Adds the
 *     new proto-def to the end of the existing list and returns the list.
 *
 *  RETURN VALUES
 *     A pointer to a list of funcDefType.
 *
 * ************************************************************************* */

funcDefType *addKRFunc( funcDefType *list, funcDefType *new )
{
   funcDefType     *ptr;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   if( new == (funcDefType *)NULL )
      return( list );
   if( list == (funcDefType *)NULL )
      return( new );

   ptr = list;
   while( ptr->funcNext )
   {
      ptr = ptr->funcNext;
   }

   ptr->funcNext = new;
   return( list );
}

/**************************************************************************
 *  Public Function freeFuncDef
 *
 *  DESCRIPTION
 *     Frees memory of a funcDefType structure.  This is used by the parser
 *     when proto-function definitions are constructed for declarations that
 *     turn out not to be part of a function declaration.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void freeFuncDef( funcDefType *junk )
{
   typeSpecType *typeptr;
   funcDefType  *funcptr;
   funcDefType  *prevfunc;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   funcptr = junk;
   while( funcptr )
   {
      if( funcptr->hdrComment )
         free( funcptr->hdrComment );
      if( funcptr->intComment )
         free( funcptr->intComment );
      if( funcptr->spec )
      {
         typeptr = funcptr->spec;
         freeTokList( typeptr->type );
         freeTokList( typeptr->storage );
         free( typeptr );
      }
      if( funcptr->func )
      {
         freeDecl( funcptr->func );
      }
      prevfunc = funcptr;
      funcptr = funcptr->funcNext;
      free( prevfunc );
   }
   return;
}

/**************************************************************************
 *  Public Function freeDecl
 *
 *  DESCRIPTION
 *     Frees memory used by a declaration definition data structure.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void freeDecl( declDefType *junkdecl )
{
   parmListType *parm;
   parmListType *prev;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   freeTokList( junkdecl->decl );
   parm = junkdecl->parms;
   while( parm )
   {
      if( parm->parmClass == SCALAR )
      {
         freeTokList( parm->parm.scalarParm );
      }
      else if( parm->parmClass == FUNC )
      {
         freeDecl( parm->parm.funcParm );
      }
      else
         printf( "%s: Invalid parameter type\n", __FUNCTION__ );
      prev = parm;
      parm = parm->parmNext;
      free( prev );
   }
   return;
}

/**************************************************************************
 *  Public Function printParmList
 *
 *  DESCRIPTION
 *     Print a function parameter list to the FILE pointer ffp.  A function
 *     definition contains a parmListType structure with an element for each
 *     parameter.  Each parameter consists of the data type, name, and address
 *     or pointer references, and these are printed for each parameter by
 *     the printParmList() function.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void printParmList( FILE *ffp, parmListType *parms )
{
   parmListType     *parmPtr;
   int               i = 0;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   parmPtr = parms;
   while( parmPtr )
   {
      if( parmPtr->parmClass == SCALAR )
         printValList( ffp, parmPtr->parm.scalarParm );
      else
      {
         if( parmPtr->parm.funcParm )
         {
            printValList( ffp, parmPtr->parm.funcParm->decl );
            printParmList( ffp, parmPtr->parm.funcParm->parms );
         }
      }
      parmPtr = parmPtr->parmNext;
      i++;
   }
   return;
}

/**************************************************************************
 *  Public Function printValList
 *
 *  DESCRIPTION
 *     Prints a linked list of tokens to the FILE pointer ffp.  For each
 *     pointer level, prints a pointer reference ('*').
 *
 *     This function is used for debugging, or visibility into the program's
 *     operation.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

void printValList( FILE *ffp, tokListType *list )
{
   tokListType    *ptr;
   int             plvl;

#ifdef YACCDEBUG
   printf( "Func %s: %s %d\n", __FUNCTION__, __FILE__, __LINE__ );
#endif
   ptr = list;
   while( ptr )
   {
      for( plvl = 0; plvl < ptr->ptrLvl; plvl++ )
         fprintf( ffp, "*" );
      fprintf( ffp, "%s ", ptr->tok );
      ptr = ptr->tokNext;
   }
   fprintf( ffp, "\n" );
   return;
}

#ifdef UNDEF
------------------------ BEGIN ILLEGITIMATE STUFF HERE ------------------------
------------------------- END ILLEGITIMATE STUFF HERE -------------------------
#endif /* UNDEF */

