/*
 * copyright (c) 2000, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * --------------------------------------------------------------------------- 
 *
 * FILE NAME:  srcdocActions.h
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 *
 * FILE DESCRIPTION
 *
 * Function prototypes for srcdoc parser action functions.
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * srcdocActions.h,v
 * Revision 1.9  2002/01/05 03:04:53  gordon
 * Changed license to GPL.
 *
 * Revision 1.8  2001/12/18 13:11:05  gordon
 * Added variables to support separating comments into header and
 * function internal types.
 *
 * Revision 1.7  2001/02/21 12:26:40  gordon
 * Added freeDecl to the public scope.
 *
 * Revision 1.6  2001/02/13 12:10:47  gordon
 * Removed a few functions that are really private functions to srcdocActions.c
 *
 * Revision 1.5  2001/02/13 05:37:11  gordon
 * Added functions for processing K&R type declarations.
 *
 * Revision 1.4  2000/08/02 12:47:56  gordon
 * Corrected the prototype for srcextractparse() function.
 *
 * Revision 1.3  2000/08/01 05:32:51  gordon
 * Changed executable name to srcextract, and some variables that depend
 * on the program name for lex/yacc autocode.
 *
 * Revision 1.2  2000/08/01 04:54:41  gordon
 * Changed names from mkdoc to srcdoc, and all variables with mkdoc prefix.
 *
 * Revision 1.1  2000/07/24 05:43:16  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */

/* ************************************************************************* */
/*                                                                           */
/* Control of Multiple Includes and File Identification                      */
/*                                                                           */
/* ************************************************************************* */

#ifndef ID_srcdocActionsH

#if( defined( SOLARIS ) )
#define ID_srcdocActionsH "%Z%%M%  control version %I%  %G% %U%"
#elif( defined( LINUX ) )
#define ID_srcdocActionsH "srcdocActions.h,v control version 1.9 2002/01/05 03:04:53"
#else
#define ID_srcdocActionsH
#endif

/* ************************************************************************* */
/*                                                                           */
/* Executable Macro Definitions                                              */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Scalar Data Types                                                         */
/*                                                                           */
/* ************************************************************************* */

typedef enum { FRETVAL, FRETPTR, FRETDPTR, FRETTPTR, FRETADDR } funcRetType;
typedef enum { STG, TYP } specType;
typedef enum
   {
#undef	FALSE
#undef	TRUE
#undef	INDETERMINATE
   FALSE, TRUE, INDETERMINATE
   } boolean;

/* ************************************************************************* */
/*                                                                           */
/* Non-Scalar Data Structures                                                */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* External Variable References                                              */
/*                                                                           */
/* ************************************************************************* */

extern int             srcextractlineno;
extern int             funcStart;
extern char           *parseFileName;
extern char           *typeList[];
extern int             numTypeDefs;
extern int             gblVbs;
extern funcDefType    *gblFuncList;
extern char            hdrCommentText[];
extern char            intCommentText[];
extern int             hdrCmtIdx;
extern int             intCmtIdx;
extern cmtControlType  cmtMode;
extern FILE           *srcextractin;

/* ************************************************************************* */
/*                                                                           */
/* Function Prototypes                                                       */
/*                                                                           */
/* ************************************************************************* */

extern void          addTypeDef( declDefType * );
extern tokListType  *addTokList( char *, tokListType *, int );
extern tokListType  *createTokList( char * );
extern tokListType  *concatTokList( tokListType *, tokListType *, int );
extern void          printValList( FILE *, tokListType * );
extern void          printParmList( FILE *, parmListType * );
extern char         *concatVal( char *, char * );
extern parmType     *createParm( void *, parmClassType );
extern parmListType *createParmList( declDefType *, parmClassType );
extern parmListType *addParmList( parmListType *, parmListType * );
extern declDefType  *createDecl( tokListType * );
extern declDefType  *addToDecl( declDefType *, parmListType * );
extern funcDefType  *createFunc( typeSpecType *, declDefType *, funcDefType * );
extern void          addFunc( funcDefType * );
extern typeSpecType *addTypeSpec( char *, typeSpecType *, specType );
extern typeSpecType *createTypeSpec( char *, specType );
extern int           srcextractparse( void );

extern declDefType  *concatDecl( declDefType *, declDefType * );
extern void          freeFuncDef( funcDefType * );
extern void          freeDecl( declDefType * );
extern funcDefType  *createKRFunc( typeSpecType *, declDefType * );
extern funcDefType  *addKRFunc( funcDefType *, funcDefType * );

/*  This one is defined in <string.h> but only if USE_BSD is defined.  It's
 *  advertised as a better strtok(), but maybe it's not getting acceptance.
 *  I'm not quite sure why I used it instead of strtok(), but I'm too lazy
 *  to go check out the differences.
*/
extern char        *strsep( char **, char * );
#endif /* ID_srcdocActionsH */
