/*
 * copyright (c) 2000, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * --------------------------------------------------------------------------- 
 *
 * FILE NAME:  srcdocLex.h
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 * FILE DESCRIPTION
 *
 * Definitions for the srcdoc lexical analyzer
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * srcdocLex.h,v
 * Revision 1.9  2002/01/05 03:06:38  gordon
 * Changed license to GPL.
 *
 * Revision 1.8  2001/12/18 13:10:07  gordon
 * Added a typedef for comment control to support separating comments into
 * header and function internal.
 *
 * Revision 1.7  2000/08/01 05:33:44  gordon
 * Changed name of executable to srcextract, and some variables that are
 * generated based on the program name.
 *
 * Revision 1.6  2000/08/01 04:53:15  gordon
 * Changed name to srcdoc from mkdoc.
 *
 * Revision 1.5  2000/07/23 14:01:07  gordon
 * Added a parmClassType element to the declDefType to differentiate
 * between scalar parameters and function pointer parameters.
 *
 * Revision 1.4  2000/07/13 12:34:31  gordon
 * Corrected the reference to the union, now uses the union itself, not a
 * pointer to the union.
 *
 * Revision 1.3  2000/07/13 12:14:46  gordon
 * Changed parameter definition to a union to handle scalars or function pointers
 *
 * Revision 1.2  2000/07/09 05:49:40  gordon
 * Cleaned up headers, commented, removed obsolete code.
 *
 * Revision 1.1  2000/06/29 11:47:29  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */

/* ************************************************************************* */
/*                                                                           */
/* Control of Multiple Includes and File Identification                      */
/*                                                                           */
/* ************************************************************************* */

#ifndef ID_srcdocLexH

#if( defined( SOLARIS ) )
#define ID_srcdocLexH "%Z%%M%  control version %I%  %G% %U%"
#elif( defined( LINUX ) )
#define ID_srcdocLexH "srcdocLex.h,v control version 1.9 2002/01/05 03:06:38"
#else
#define ID_srcdocLexH
#endif

/* ************************************************************************* */
/*                                                                           */
/* Executable Macro Definitions                                              */
/*                                                                           */
/* ************************************************************************* */

#define srcextracterror(x)  ( fprintf( stderr, \
                              "Parser: %s at line %d, column %d, file %s\n", \
                              (x), srcextractlineno, column, parseFileName ) )

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Scalar Data Types                                                         */
/*                                                                           */
/* ************************************************************************* */

typedef enum { SCALAR, FUNC } parmClassType;
typedef enum { CMTHDR, CMTINT, CMTNONE } cmtControlType;

/* ************************************************************************* */
/*                                                                           */
/* Non-Scalar Data Structures                                                */
/*                                                                           */
/* ************************************************************************* */

typedef struct _tokList tokListType;
struct _tokList {
   int            ptrLvl;
   char          *tok;
   tokListType   *tokNext;
};

typedef union _parm parmType;

typedef struct _parmList parmListType;

typedef struct {
   parmClassType  declClass;
   tokListType   *decl;
   parmListType  *parms;
} declDefType;

union _parm {
   tokListType   *scalarParm;
   declDefType   *funcParm;
};

struct _parmList {
   parmClassType  parmClass;
   parmType       parm;
   parmListType  *parmNext;
};

typedef struct {
   tokListType   *type;
   tokListType   *storage;
} typeSpecType;

typedef struct _funcDef funcDefType;
struct _funcDef {
   char          *hdrComment;
   char          *intComment;
   typeSpecType  *spec;
   declDefType   *func;
   funcDefType   *funcNext;
} ;

typedef union {
   int                iVal;
   float              fVal;
   char              *sVal;
   tokListType       *tokList;
   typeSpecType      *typeVal;
   parmListType      *pValList;
   declDefType       *declVal;
   funcDefType       *funcVal;
#if ( defined( LINUX ) || defined( SOLARIS ) )
} YYSTYPE;
#else
} DLOGSTYPE;
#endif

/* ************************************************************************* */
/*                                                                           */
/* External Variable References                                              */
/*                                                                           */
/* ************************************************************************* */

extern int column;

/* ************************************************************************* */
/*                                                                           */
/* Function Prototypes                                                       */
/*                                                                           */
/* ************************************************************************* */

extern int   srcextractlex( void );
extern char *srcdocFilterLexId( void );
extern char *srcdocFilterLexCopyright( void );
extern char *srcdocFilterLexUsage( void );

#endif /* ID_srcdocLexH */
