%{
#if( defined( SOLARIS ) )
static char   ident[]   = "%Z%%M%  control version %I%  %G% %U%";
#elif( defined( LINUX ) )
static char   ident[]   = "srcdocLex.l,v control version 1.7 2002/01/05 03:08:34";
#endif
static char   cpyrght[] = "Copyright (c) 2000, Gordon B. Aaseng";
static char   usage[]   = "Lexical analyzer for man page tool";

/*
 * copyright (c) 2000, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/*
 *
 * FILE NAME
 *
 * srcdocLex.l
 *
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 * FILE DESCRIPTION
 *
 * This module contains the 'lex' definitions for the man page generator.
 * It looks for tokens that identify function definitions in C source code,
 * ignoring everything else.  It uses states to ignore function bodies once a
 * function header has been found.
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * srcdocLex.l,v
 * Revision 1.7  2002/01/05 03:08:34  gordon
 * Changed license to GPL.
 *
 * Revision 1.6  2001/12/18 13:05:14  gordon
 * Added a lexeme for the __extension__ construct.  Split comment extraction
 * into separate buffers for header comments and comments inside functions.
 *
 * Revision 1.5  2000/08/01 05:34:26  gordon
 * Changed name of executable to srcextract, and some variable names that
 * are based on the program name.
 *
 * Revision 1.4  2000/08/01 04:53:42  gordon
 * Changed name from mkdoc to srcdoc, and all variables with mkdoc prefix.
 *
 * Revision 1.3  2000/07/23 13:55:25  gordon
 * Added pseudo-keywords '__volatile__', 'inline' and 'asm'.  Returns the
 * string value for constants. Cleaned up headers and comments.
 *
 * Revision 1.2  2000/07/04 13:08:01  gordon
 * Added a switch to get only comments preceding the function definition, or
 * get all comments in the function.  Defaults to preceding the definition.
 *
 * Revision 1.1  2000/06/29 12:19:41  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */


#define   _POSIX_SOURCE     1       /* posix.1             */
#ifndef   _POSIX_C_SOURCE
#define   _POSIX_C_SOURCE   199309  /* posix.1 and posix.4 */
#endif

/* ************************************************************************* */
/*                                                                           */
/* Included Files                                                            */
/*                                                                           */
/* ************************************************************************* */

/*
 * os specific header files.
 */
/*  (None.)  */

/*
 * standard c (ANSI) header files
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>

/*
 * posix header files
 */
/*  (None.)  */

/*
 * project specific header files.
 */

#include "srcdocLex.h"
#include "srcdocParse.h"

/* ************************************************************************* */
/*                                                                           */
/* External References                                                       */
/*                                                                           */
/* ************************************************************************* */

extern int     funcStart;
extern int     topOfFile;
extern char   *strdup( );

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

#define  NOECHO    (0x0)
#define  TOKECHO   (0x1)
#define  TOKDUP    (0x2)
#define  CMTSIZE   (4096)

/* ************************************************************************* */
/*                                                                           */
/* Data Structures                                                           */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Public Global Variables                                                   */
/*                                                                           */
/* ************************************************************************* */

char *typeList[2048];
int   numTypeDefs = 0;
char  hdrCommentText[CMTSIZE];
char  intCommentText[CMTSIZE];
int   hdrCmtIdx  = 0;
int   intCmtIdx  = 0;
cmtControlType   cmtMode = CMTHDR;


/* ************************************************************************* */
/*                                                                           */
/* Private (static) Global Variables                                         */
/*                                                                           */
/* ************************************************************************* */

static int   prevstate[100];
static int   cmtlevel     = 0;
static int   prevstateidx = 0;
static int   brclevel     = 0;

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (Locally Used Functions)                                 */
/*                                                                           */
/* ************************************************************************* */


/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

static void pushstate( int, int );
static int popstate( void );
static void resetstate( int );
static void count( int );
static int check_type( void );

/* ---------------------------------------------------------------------- */
%}

%option  yylineno
%option  outfile = "srcdocLex.c"  
%option  prefix  = "srcextract"
%option  noyywrap
%option  nounput

%s COMMENT FUNCBODY PREPROC
%o 8000
%a 4000

D			[0-9]
L			[a-zA-Z_]
H			[a-fA-F0-9]
E			[Ee][+-]?{D}+
FS			(f|F|l|L)
IS			(u|U|l|L)*
opencmt                 "/*"
closecmt                "*/"
dummy                   "qwertyuiop"


%%

%{
   if( topOfFile )
   {
      topOfFile = 0;
      resetstate( INITIAL );
      BEGIN INITIAL;
   }
   if( funcStart == 1 )
   {
      brclevel++;
      pushstate( INITIAL, FUNCBODY );
      funcStart = 0;
   }
%}

<COMMENT>{opencmt}   { cmtlevel++; }
<COMMENT>{closecmt}  { if( --cmtlevel == 0 )
                       {
                          if( cmtMode == CMTHDR && hdrCmtIdx < (CMTSIZE-1) )
                          {
                             hdrCommentText[hdrCmtIdx] = '\n';
                             hdrCmtIdx++;
                          }
                          else if( cmtMode == CMTINT && intCmtIdx < (CMTSIZE-1))
                          {
                             intCommentText[intCmtIdx] = '\n';
                             intCmtIdx++;
                          }
                          BEGIN popstate();
                       }
                     }
<COMMENT>.|\n        { if( cmtMode == CMTHDR )
                       {
                          if( hdrCmtIdx < (CMTSIZE - 3) )
                          {
                             hdrCommentText[hdrCmtIdx] = yytext[0];
                             hdrCmtIdx++;
                          }
                          else
                          {
                             hdrCommentText[hdrCmtIdx] = '\n';
                             hdrCmtIdx++;
                             hdrCommentText[hdrCmtIdx] = '\0';
                             cmtMode = CMTNONE;
                          }
                       }
                       else if( cmtMode == CMTINT )
                       {
                          if( intCmtIdx < (CMTSIZE - 3) )
                          {
                             intCommentText[intCmtIdx] = yytext[0];
                             intCmtIdx++;
                          }
                          else
                          {
                             intCommentText[intCmtIdx] = '\n';
                             intCmtIdx++;
                             intCommentText[intCmtIdx] = '\0';
                             cmtMode = CMTNONE;
                          }
                       }
                     }

<INITIAL>"#"         { pushstate( INITIAL, PREPROC ); }
<INITIAL>{opencmt}   { cmtlevel++;
                       pushstate( INITIAL, COMMENT ); }
<INITIAL>"auto"      { count(TOKDUP);  return(AUTO); }
<INITIAL>"break"     { count(TOKECHO); return(BREAK); }
<INITIAL>"case"      { count(TOKECHO); return(CASE); }
<INITIAL>"char"      { count(TOKDUP); return(CHAR); }
<INITIAL>"const"     { count(TOKDUP); return(CONST); }
<INITIAL>"__const"   { count(TOKDUP); return(CONST); }
<INITIAL>"continue"  { count(TOKECHO); return(CONTINUE); }
<INITIAL>"default"   { count(TOKECHO); return(DEFAULT); }
<INITIAL>"do"        { count(TOKECHO); return(DO); }
<INITIAL>"double"    { count(TOKDUP); return(DOUBLE); }
<INITIAL>"else"      { count(TOKECHO); return(ELSE); }
<INITIAL>"enum"      { count(TOKDUP); return(ENUM); }
<INITIAL>"extern"    { count(TOKDUP);  return(EXTERN); }
<INITIAL>"float"     { count(TOKDUP); return(FLOAT); }
<INITIAL>"for"       { count(TOKECHO); return(FOR); }
<INITIAL>"goto"      { count(TOKECHO); return(GOTO); }
<INITIAL>"if"        { count(TOKECHO); return(IF); }
<INITIAL>"int"       { count(TOKDUP); return(INT); }
<INITIAL>"long"      { count(TOKDUP); return(LONG); }
<INITIAL>"register"  { count(TOKDUP);  return(REGISTER); }
<INITIAL>"return"    { count(TOKECHO); return(RETURN); }
<INITIAL>"short"     { count(TOKDUP); return(SHORT); }
<INITIAL>"signed"    { count(TOKDUP); return(SIGNED); }
<INITIAL>"sizeof"    { count(TOKECHO); return(SIZEOF); }
<INITIAL>"static"    { count(TOKDUP);  return(STATIC); }
<INITIAL>"struct"    { count(TOKDUP); return(STRUCT); }
<INITIAL>"switch"    { count(TOKECHO); return(SWITCH); }
<INITIAL>"typedef"   { count(TOKDUP);  return(TYPEDEF); }
<INITIAL>"union"     { count(TOKDUP); return(UNION); }
<INITIAL>"unsigned"  { count(TOKDUP); return(UNSIGNED); }
<INITIAL>"void"      { count(TOKDUP); return(VOID); }
<INITIAL>"volatile"  { count(TOKDUP); return(VOLATILE); }
<INITIAL>"__volatile__"  { count(TOKDUP);  return(VOLATILE);  }
<INITIAL>"while"         { count(TOKECHO); return(WHILE);     }
<INITIAL>"__attribute__" { count(TOKECHO); return(ATTRIBUTE); }
<INITIAL>"__inline__"    { count(TOKDUP);  return(INLINE);    }
<INITIAL>"inline"        { count(TOKDUP);  return(INLINE);    }
<INITIAL>"__asm__"       { count(TOKECHO); return(ASM);       }
<INITIAL>"asm"           { count(TOKECHO); return(ASM);       }
<INITIAL>"typeof"        { count(TOKECHO); return(TYPEOF);    }
<INITIAL>"__typeof"      { count(TOKECHO); return(TYPEOF);    }
<INITIAL>"__extension__" { count(TOKECHO); return(EXTENSION); }

<INITIAL>{L}({L}|{D})* { count(NOECHO); return(check_type()); }

<INITIAL>0[xX]{H}+{IS}? { count(TOKDUP); return(CONSTANT); }
<INITIAL>0{D}+{IS}?     { count(TOKDUP); return(CONSTANT); }
<INITIAL>{D}+{IS}?      { count(TOKDUP); return(CONSTANT); }
<INITIAL>'(\\.|[^\\'])+' { count(TOKDUP); return(CONSTANT); }

<INITIAL>{D}+{E}{FS}?    { count(TOKDUP); return(CONSTANT); }
<INITIAL>{D}*"."{D}+({E})?{FS}?   { count(TOKDUP); return(CONSTANT); }
<INITIAL>{D}+"."{D}*({E})?{FS}?   { count(TOKDUP); return(CONSTANT); }

<INITIAL>\"(\\.|[^\\"])*\"   { count(TOKECHO); return(STRING_LITERAL); }

<INITIAL>\.\.\.              { count(TOKDUP); return(ELIPSIS); }
<INITIAL>">>="               { count(TOKECHO); return(RIGHT_ASSIGN); }
<INITIAL>"<<="               { count(TOKECHO); return(LEFT_ASSIGN); }
<INITIAL>"+="                { count(TOKECHO); return(ADD_ASSIGN); }
<INITIAL>"-="                { count(TOKECHO); return(SUB_ASSIGN); }
<INITIAL>"*="                { count(TOKECHO); return(MUL_ASSIGN); }
<INITIAL>"/="                { count(TOKECHO); return(DIV_ASSIGN); }
<INITIAL>"%="                { count(TOKECHO); return(MOD_ASSIGN); }
<INITIAL>"&="                { count(TOKECHO); return(AND_ASSIGN); }
<INITIAL>"^="                { count(TOKECHO); return(XOR_ASSIGN); }
<INITIAL>"|="                { count(TOKECHO); return(OR_ASSIGN); }
<INITIAL>">>"                { count(TOKECHO); return(RIGHT_OP); }
<INITIAL>"<<"                { count(TOKECHO); return(LEFT_OP); }
<INITIAL>"++"                { count(TOKECHO); return(INC_OP); }
<INITIAL>"--"                { count(TOKECHO); return(DEC_OP); }
<INITIAL>"->"                { count(TOKECHO); return(PTR_OP); }
<INITIAL>"&&"                { count(TOKECHO); return(AND_OP); }
<INITIAL>"||"                { count(TOKECHO); return(OR_OP); }
<INITIAL>"<="                { count(TOKECHO); return(LE_OP); }
<INITIAL>">="                { count(TOKECHO); return(GE_OP); }
<INITIAL>"=="                { count(TOKECHO); return(EQ_OP); }
<INITIAL>"!="                { count(TOKECHO); return(NE_OP); }
<INITIAL>";"                 { count(TOKECHO); return(';'); }
<INITIAL>"{"                 { count(TOKECHO); return('{'); }
<INITIAL>"}"                 { count(TOKECHO); return('}'); }
<INITIAL>","                 { count(TOKECHO); return(','); }
<INITIAL>":"                 { count(TOKECHO); return(':'); }
<INITIAL>"="                 { count(TOKECHO); return('='); }
<INITIAL>"("                 { count(TOKECHO); return('('); }
<INITIAL>")"                 { count(TOKECHO); return(')'); }
<INITIAL>"["                 { count(TOKECHO); return('['); }
<INITIAL>"]"                 { count(TOKECHO); return(']'); }
<INITIAL>"."                 { count(TOKECHO); return('.'); }
<INITIAL>"&"                 { count(TOKECHO); return('&'); }
<INITIAL>"!"                 { count(TOKECHO); return('!'); }
<INITIAL>"~"                 { count(TOKECHO); return('~'); }
<INITIAL>"-"                 { count(TOKECHO); return('-'); }
<INITIAL>"+"                 { count(TOKECHO); return('+'); }
<INITIAL>"*"                 { count(TOKECHO); return('*'); }
<INITIAL>"/"                 { count(TOKECHO); return('/'); }
<INITIAL>"%"                 { count(TOKECHO); return('%'); }
<INITIAL>"<"                 { count(TOKECHO); return('<'); }
<INITIAL>">"                 { count(TOKECHO); return('>'); }
<INITIAL>"^"                 { count(TOKECHO); return('^'); }
<INITIAL>"|"                 { count(TOKECHO); return('|'); }
<INITIAL>"?"                 { count(TOKECHO); return('?'); }

<INITIAL>[ \t\v\n\f]         { count(NOECHO); }
<INITIAL>.                   { /* ignore bad characters */ }

<FUNCBODY>"{"   { brclevel++; }
<FUNCBODY>"}"   { if( --brclevel == 0 )
                     {
                       BEGIN popstate();
                       return( yytext[0] );
                     }
                } 
<FUNCBODY>.|\n  ;
<FUNCBODY>{dummy}  { /* This eliminates a compiler warning for unused func */
                     yy_flex_realloc(NULL, 0); REJECT; }

<PREPROC>\\\n   ;
<PREPROC>\n     { BEGIN popstate(); }
<PREPROC>.      ;

%%

#ifdef LEXDEBUG
static char   *states[] = { "INITIAL", "COMMENT", "FUNCBODY" };
#endif

/* *************************************************************************
 *
 *  Private function pushstate
 *
 *  DESCRIPTION
 *     Save the previous lex state (oldstate) on a stack, and set the lex
 *     state to newstate. 
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void pushstate( int oldstate, int newstate )
{
#ifdef LEXDEBUG
   fprintf( stdout,
            "LEX pushstate: save state %d to %s, set new state %s, line %d\n",
            prevstateidx, states[oldstate],
            states[newstate], srcextractlineno );
#endif
   prevstate[prevstateidx] = oldstate;
   prevstateidx++;
   BEGIN newstate;
   return;
}

/* *************************************************************************
 *
 *  Private function popstate
 *
 *  DESCRIPTION
 *     Get the previous lex state off the stack created with pushstate(),
 *     and set the state stack pointer to the previous state value.
 *  RETURN VALUES
 *     The integer value of the popped lex state variable.
 *
 * ************************************************************************* */

static int popstate( void )
{
   --prevstateidx;
#ifdef LEXDEBUG
   fprintf( stdout,
            "LEX popstate: set state to %s; idx is %d\n",
            states[prevstate[prevstateidx]], prevstateidx );
#endif
   return( prevstate[prevstateidx] );
}

/* *************************************************************************
 *
 *  Private function resetstate
 *
 *  DESCRIPTION
 *     Initialize the lex state stack variables to starting conditions.
 *     The calling function specifies what the initial lex state should be.
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void resetstate( int initstate )
{
#ifdef LEXDEBUG
   fprintf( stdout,
            "LEX resetstate: prevstateidx %d, cmtlevel %d\n",
            prevstateidx, cmtlevel );
#endif

   prevstateidx = 0;
   cmtlevel     = 0;
   prevstate[prevstateidx] = initstate;
   return;
}

int column = 0;

/* *************************************************************************
 *
 *  Private function count
 *
 *  DESCRIPTION
 *     The count() function is invoked on each lexical element to keep
 *     track of the current column in the source text so that error messages
 *     can identify exactly where in the source code an error occurred.
 *     It also makes a copy of lexical elements if the parser needs the
 *     string value of the token, as specified by tokcmd.  If tokcmd is
 *     TOKECHO, count() echoes the name if LEXDEBUG is defined at compile
 *     time.  If TOKDUP is set, count() copies the token into the lex-defined
 *     srcextractlval (generically, yylval) union.
 *
 *  RETURN VALUES
 *     None (void).
 *
 * ************************************************************************* */

static void count( int tokcmd )
{
   int i;

   switch( tokcmd )
   {
      case TOKECHO :
#ifdef LEXDEBUG
         fprintf( stdout, "LEX: %s\n", yytext );
#endif
         break;
      case TOKDUP :
#ifdef LEXDEBUG
         fprintf( stdout, "LEX: %s, line %d\n", yytext, srcextractlineno );
#endif
         srcextractlval.sVal = strdup( yytext );
         break;
      default :
         break;
   }
   for (i = 0; yytext[i] != '\0'; i++)
      if (yytext[i] == '\n')
         column = 0;
      else if (yytext[i] == '\t')
         column += 8 - (column % 8);
      else
         column++;
   return;
}

/* *************************************************************************
 *
 *  Private function check_type
 *
 *  DESCRIPTION
 *     The check_type() function checks source code variable names to
 *     determine if the name is an identifier or a type name.  To do this,
 *     it compares each name in a list of previously identified type names,
 *     and if the name is found, returns TYPE_NAME.  If the name is not found
 *     in the list, then it's an IDENTIFIER.
 *
 *     The parser adds identifiers to the list typeList when it finds new
 *     enumerated type definitions.  It gets a little messy when a name is
 *     used as an enumerated type definition and then is used elsewhere as
 *     a variable name.  This part tries to not be context-sensitive so it
 *     doesn't know what the writer meant when using the name.
 *
 *  RETURN VALUES
 *     The value TYPE_NAME or IDENTIFIER, which are integers defined by the
 *     yacc grammar definition.
 *
 * ************************************************************************* */

static int check_type( void )
{
   int     i;

   srcextractlval.sVal = strdup( yytext );
   for( i = 0; i < numTypeDefs; i++ )
   {
      if( strcmp( typeList[i], yytext ) == 0 )
      {
#ifdef LEXDEBUG
         fprintf( stdout,
                  "LEX TYPE_NAME %d: %s = %s\n",
                  i, typeList[i], yytext );
#endif
         return( TYPE_NAME );
      }
   }

#ifdef LEXDEBUG
   fprintf( stdout, "LEX ID: %s\n", yytext );
#endif
   return(IDENTIFIER);
}

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (External Interface Functions)                           */
/*                                                                           */
/* ************************************************************************* */
char *srcdocFilterLexId( void );
char *srcdocFilterLexCopyright( void );
char *srcdocFilterLexUsage( void );

/* *************************************************************************
 *
 *  Public Functions srcdocFilterLexId, srcdocFilterLexCopyright
 *               and srcdocFilterLexUsage
 *
 *  DESCRIPTION
 *     Module Identification functions used to print file ID's with the
 *     command line '-z' switch.
 *
 *  RETURN VALUES
 *     Pointer to the specified strings
 *
 * ************************************************************************* */

char *srcdocFilterLexId( void )
{
   return( ident );
}

char *srcdocFilterLexCopyright( void )
{
   return( cpyrght );
}

char *srcdocFilterLexUsage( void )
{
   return( usage );
}

#ifdef UNDEF
------------------------ BEGIN ILLEGITIMATE STUFF HERE ------------------------
------------------------- END ILLEGITIMATE STUFF HERE -------------------------
#endif /* UNDEF */
