%{
#if( defined( SOLARIS ) )
static char   ident[] __attribute__((unused))
         = "%Z%%M%  control version %I%  %G% %U%";
#elif( defined( LINUX ) )
static char   ident[] __attribute__((unused))
         = "srcdocParse.y,v control version 1.8 2002/01/05 03:10:25";
#endif
static char   cpyrght[] __attribute__((unused))
         = "Copyright (c) 2000, Gordon B. Aaseng";
static char   usage[]   __attribute__((unused))
         = "Parser for the man page generator";

/*
 * copyright (c) 2000, 2001, Gordon B. Aaseng
 *                     Glendale, AZ
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this distribution; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*
 *
 * FILE NAME
 *
 * srcdocParse.y
 *
 *
 * PROGRAM INFORMATION
 *
 * Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
 * Under License:	General Public License
 *
 *
 * FILE DESCRIPTION
 *
 * This file contains the YACC specification for the man page generator
 * parser.  The parser looks for function definitions in C source code and
 * saves function names, parameters and return types.  It also saves comments
 * preceding a function definition to be used in the man page outline.  When
 * a function definition is found, it forces a state change in the lexical
 * analyzer so that the function body is ignored.
 *
 *
 * RCS CHANGE LOG
 *    NOTE:  Do not edit this section; it is managed by RCS.
 * srcdocParse.y,v
 * Revision 1.8  2002/01/05 03:10:25  gordon
 * Changed license to GPL.
 *
 * Revision 1.7  2001/12/18 13:01:41  gordon
 * Corrected a bug in handling GNU __extension__ constructs, and changed
 * the way header and internal function comments are extracted.
 *
 * Revision 1.6  2001/02/21 12:20:50  gordon
 * Added rules for parsing the GNU extension allowing code blocks in
 * expressions, e.g. if(( __extension__ ( { ... } ) )) { ... }
 *
 * Revision 1.5  2001/02/13 05:38:28  gordon
 * Added rules for parsing K&R type declarations.
 *
 * Revision 1.4  2000/08/01 04:55:28  gordon
 * Changed names from mkdoc to srcdoc, and all variables with mkdoc prefix.
 *
 * Revision 1.3  2000/07/23 13:48:03  gordon
 * Moved all actions to separate file; corrected parse errors for 'asm', type
 * names used as structure elements, labeled statements and constants used
 * as function parameters.
 *
 * Revision 1.2  2000/07/04 13:09:29  gordon
 * Added a switch to get only the comments preceding the function definition;
 * this is the default.  User can ask for all comments from the command line.
 *
 * Revision 1.1  2000/06/29 12:02:10  gordon
 * Initial revision
 *
 *
 *********************  End of RCS CHANGE LOG *******************
 *
 */


#define   _POSIX_SOURCE     1       /* posix.1             */
#ifndef   _POSIX_C_SOURCE
#define   _POSIX_C_SOURCE   199309  /* posix.1 and posix.4 */
#endif

/* ************************************************************************* */
/*                                                                           */
/* Included Files                                                            */
/*                                                                           */
/* ************************************************************************* */

/*
 * os specific header files.
 */
/*  (None.)  */

/*
 * standard c (ANSI) header files
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
 * posix header files
 */
/*  (None.)  */

/*
 * project specific header files.
 */

#include "srcdocLex.h"
#include "srcdocActions.h"

/* ************************************************************************* */
/*                                                                           */
/* External References                                                       */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Manifest Constants                                                        */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Data Structures                                                           */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Public Global Variables                                                   */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/* ************************************************************************* */
/*                                                                           */
/* Private (static) Global Variables                                         */
/*                                                                           */
/* ************************************************************************* */

static int             typedefflag = 0;

/* ************************************************************************* */
/*                                                                           */
/* Executable Code  (Locally Used Functions)                                 */
/*                                                                           */
/* ************************************************************************* */

/*  (None.)  */

/**************************************************************************
 * Private Function Prototypes
 **************************************************************************/

/*  (None.)  */

/* ---------------------------------------------------------------------- */
%}

%token STRING_LITERAL SIZEOF
%token PTR_OP INC_OP DEC_OP LEFT_OP RIGHT_OP LE_OP GE_OP EQ_OP NE_OP
%token AND_OP OR_OP MUL_ASSIGN DIV_ASSIGN MOD_ASSIGN ADD_ASSIGN
%token SUB_ASSIGN LEFT_ASSIGN RIGHT_ASSIGN AND_ASSIGN
%token XOR_ASSIGN OR_ASSIGN

%token CASE DEFAULT IF ELSE SWITCH WHILE DO FOR GOTO CONTINUE BREAK RETURN
%token ATTRIBUTE ASM

%token <sVal> IDENTIFIER CHAR SHORT INT LONG SIGNED UNSIGNED FLOAT DOUBLE
%token <sVal> CONST VOLATILE VOID TYPE_NAME EXTENSION
%token <sVal> TYPEDEF EXTERN STATIC AUTO REGISTER
%token <sVal> STRUCT UNION ENUM ELIPSIS
%token <sVal> CONSTANT INLINE TYPEOF

%type <sVal>     identifier storage_class_specifier type_specifier
%type <sVal>     struct_or_union struct_or_union_specifier enum_specifier

%type <typeVal>  declaration_specifiers
%type <tokList>  type_specifier_list type_name
%type <tokList>  pointer abstract_declarator

%type <pValList> parameter_type_list parameter_list
%type <pValList> parameter_declaration
%type <declVal>  declarator declarator2 init_declarator init_declarator_list
%type <funcVal>  function_definition external_definition file declaration
%type <funcVal>  declaration_list function_body

%start file
%%

primary_expr
	: identifier { $$.iVal = 0; }
	| CONSTANT   { $$.iVal = 0; }
/*  Changed from the original to allow broken strings, i.e. where a quoted
    string is closed on one line, then resumed on the next line 
    "like"
    "this."
*/
/*
	| STRING_LITERAL
*/
	| string_literal
	| '(' expr ')'
	;

string_literal
	: STRING_LITERAL
	| string_literal STRING_LITERAL
	;

postfix_expr
	: primary_expr
	| postfix_expr '[' expr ']'
	| postfix_expr '(' ')'
	| postfix_expr '(' argument_expr_list ')'
	| postfix_expr '.' identifier
	| postfix_expr '.' TYPE_NAME
	| postfix_expr PTR_OP identifier
	| postfix_expr INC_OP
	| postfix_expr DEC_OP
/* Added to the original grammar to handle __asm__ instructions */
	| assembly
	;

argument_expr_list
	: assignment_expr
	| argument_expr_list ',' assignment_expr
	;

unary_expr
	: postfix_expr
	| INC_OP unary_expr
	| DEC_OP unary_expr
	| unary_operator cast_expr
	| SIZEOF unary_expr
	| SIZEOF '(' type_name ')'
	;

unary_operator
	: '&'
	| '*'
	| '+'
	| '-'
	| '~'
	| '!'
	;

cast_expr
	: unary_expr
	| '(' type_name ')' cast_expr
	;

multiplicative_expr
	: cast_expr
	| multiplicative_expr '*' cast_expr
	| multiplicative_expr '/' cast_expr
	| multiplicative_expr '%' cast_expr
	;

additive_expr
	: multiplicative_expr
	| additive_expr '+' multiplicative_expr
	| additive_expr '-' multiplicative_expr
	;

shift_expr
	: additive_expr
	| shift_expr LEFT_OP additive_expr
	| shift_expr RIGHT_OP additive_expr
	;

relational_expr
	: shift_expr
	| relational_expr '<' shift_expr
	| relational_expr '>' shift_expr
	| relational_expr LE_OP shift_expr
	| relational_expr GE_OP shift_expr
	;

equality_expr
	: relational_expr
	| equality_expr EQ_OP relational_expr
	| equality_expr NE_OP relational_expr
	;

and_expr
	: equality_expr
	| and_expr '&' equality_expr
	;

exclusive_or_expr
	: and_expr
	| exclusive_or_expr '^' and_expr
	;

inclusive_or_expr
	: exclusive_or_expr
	| inclusive_or_expr '|' exclusive_or_expr
	;

logical_and_expr
	: inclusive_or_expr
	| logical_and_expr AND_OP inclusive_or_expr
	;

logical_or_expr
	: logical_and_expr
	| logical_or_expr OR_OP logical_and_expr
	;

conditional_expr
	: logical_or_expr
	| logical_or_expr '?' logical_or_expr ':' conditional_expr
	;

assignment_expr
	: conditional_expr
	| unary_expr assignment_operator assignment_expr
	;

assignment_operator
	: '='
	| MUL_ASSIGN
	| DIV_ASSIGN
	| MOD_ASSIGN
	| ADD_ASSIGN
	| SUB_ASSIGN
	| LEFT_ASSIGN
	| RIGHT_ASSIGN
	| AND_ASSIGN
	| XOR_ASSIGN
	| OR_ASSIGN
	;
expr
	: assignment_expr
	| expr ',' assignment_expr
/*
 * Added this rule to parse the GNU extension that allows an expression
 * containing a compound statement.  This is encountered in the macro
 * FD_ISSET found in 'select.h' and other macros in the linux includes. 
 * The code construct is
 *    if( ( __extension__ ( { ... } ) ) ) { ... }
*/ 
        | EXTENSION '(' compound_statement ')' { $$.iVal = 0; }
	;

constant_expr
	: conditional_expr
	;

/*  For K&R declarations ... need to combine the init_declarator_list with
    the declaration_specifiers into something ... */

declaration
	: declaration_specifiers ';' { $$ = NULL; }
	| declaration_specifiers init_declarator_list ';'
                                     { $$ = createKRFunc( $1, $2 );
                                       if( typedefflag == 1 ) typedefflag = 0;
                                     }
	;

declaration_specifiers
	: storage_class_specifier { $$ = createTypeSpec( $1, STG );  }
	| storage_class_specifier declaration_specifiers
                                  { $$ = addTypeSpec( $1, $2, STG ); }
	| EXTENSION storage_class_specifier
                                  { $$ = createTypeSpec( $2, STG );  }
	| EXTENSION storage_class_specifier declaration_specifiers
                                  { $$ = addTypeSpec( $2, $3, TYP ); }
	| type_specifier          { $$ = createTypeSpec( $1, TYP );  }
	| type_specifier declaration_specifiers
                                  { $$ = addTypeSpec( $1, $2, TYP ); }
	| EXTENSION type_specifier
                                  { $$ = createTypeSpec( $2, TYP );  }
	| EXTENSION type_specifier declaration_specifiers
                                  { $$ = addTypeSpec( $2, $3, TYP ); }
	;

/* For K&R declarations ...  these will get converted to parameter specs
   after the function is completed.
*/

init_declarator_list
	: init_declarator { $$ = $1; }
	| init_declarator_list ',' init_declarator
                          { $$ = concatDecl( $1, $3 ); }
	;

init_declarator
	: declarator
                    { if( typedefflag == 1 )
                      {
                         addTypeDef( $1 );
                      }
                      $$ = $1;
                    }
	| declarator '=' initializer
                    { if( typedefflag == 1 )
                      {
                         addTypeDef( $1 );
                      }
                      $$ = $1;
                    }
	;

storage_class_specifier
	: TYPEDEF  { $$ = $1; typedefflag = 1; }
	| EXTERN   { $$ = $1; }
	| STATIC   { $$ = $1; }
	| AUTO     { $$ = $1; }
	| REGISTER { $$ = $1; }
	| INLINE   { $$ = $1; }
	;

type_specifier
	: CHAR                          { $$ = $1; }
	| SHORT                         { $$ = $1; }
	| INT                           { $$ = $1; }
	| LONG                          { $$ = $1; }
	| SIGNED                        { $$ = $1; }
	| UNSIGNED                      { $$ = $1; }
	| FLOAT                         { $$ = $1; }
	| DOUBLE                        { $$ = $1; }
	| CONST                         { $$ = $1; }
	| VOLATILE                      { $$ = $1; }
	| VOID                          { $$ = $1; }
	| struct_or_union_specifier     { $$ = $1; }
	| enum_specifier                { $$ = $1; }
	| TYPE_NAME                     { $$ = $1; }
        | TYPEOF '(' declarator ')'     { $$ = $1; }
	;

struct_or_union_specifier
	: struct_or_union identifier '{' struct_declaration_list '}'
                                       { $$ = concatVal( $1, $2 ); }
	| struct_or_union '{' struct_declaration_list '}' { $$ = $1; }
	| struct_or_union identifier   { $$ = concatVal( $1, $2 ); }
	| struct_or_union TYPE_NAME    { $$ = concatVal( $1, $2 ); }
	;

struct_or_union
	: STRUCT     { $$ = $1; }
	| UNION      { $$ = $1; }
	;

struct_declaration_list
	: struct_declaration
	| struct_declaration_list struct_declaration
	;

struct_declaration
	: type_specifier_list struct_declarator_list ';' { $$.iVal = 0; }
	;

struct_declarator_list
	: struct_declarator
	| struct_declarator_list ',' struct_declarator
	;

struct_declarator
	: declarator { $$.iVal = 0; freeDecl( $1 ); }
	| ':' constant_expr
	| declarator ':' constant_expr { $$.iVal = 0; freeDecl( $1 ); }
/*  Added because you can name the structure element as a previously defined
 *  type name, as in 'union a { int integer; float real; }' where 'real' has
 *  already been defined as an enumerated type.  */
	| TYPE_NAME { $$.iVal = 0; }
	;

enum_specifier
	: ENUM '{' enumerator_list '}' { $$ = $1; }
	| ENUM identifier '{' enumerator_list '}'
                                       { $$ = concatVal( $1, $2 ); }
	| ENUM identifier              { $$ = concatVal( $1, $2 ); }
	;

enumerator_list
	: enumerator
	| enumerator_list ',' enumerator
	;

enumerator
	: identifier { $$.iVal = 0; }
	| identifier '=' constant_expr { $$.iVal = 0; }
	;

/*
 *  Comment collection is controlled by 'cmtMode' which separates
 *  comments prior to the function specifier from comments inside the
 *  function.  'declarator' either demarks the start of a function, or it
 *  resets the comment collection buffer and discards comments that are not
 *  directly associated with a function.
*/

declarator
	: declarator2 { $$ = $1;
                        if( cmtMode != CMTINT ) {
                            cmtMode = CMTINT;
                            hdrCommentText[hdrCmtIdx] = '\0';
                            hdrCmtIdx = 0;
                        }
                      }
	| pointer declarator2 { $2->decl->ptrLvl = $1->ptrLvl;
                                $$ = $2;
                                if( cmtMode != CMTINT ) {
                                    cmtMode = CMTINT;
                                    hdrCommentText[hdrCmtIdx] = '\0';
                                    hdrCmtIdx = 0;
                                }
                              }
	;

declarator2
	: identifier { $$ = createDecl( createTokList( $1 ) ); }
	| declarator2 attribute { $$ = $1; }
	| '(' declarator ')'  { $2->declClass = FUNC;
                                $$ = $2;
                              }
	| declarator2 '[' ']'               { $$ = $1; }
	| declarator2 '[' constant_expr ']' { $$ = $1; }
	| declarator2 '(' ')'               { $$ = $1; }
	| declarator2 '(' parameter_type_list ')'
                              { $$ = addToDecl( $1, $3 ); }
	| declarator2 '(' parameter_identifier_list ')' { $$ = $1; }
	;

/*  This is for handling the __asm__ instructions, and attempts to work for
    GNU extensions as well as ANSI C definitions.  I didn't really find a
    complete grammar for this construct, just various bits and pieces in the
    gcc info pages, various C language reference web sites, some news group
    postings, etc.  I didn't look real hard, but I did browse around.  I
    ended up using examples that were in header files, and did enough to
    parse those.  The FD_ZERO and related macros use this construct and
    this code works on that macro.

    This was made a bit more complicated because the __asm__ construct can be
    appended after a function definition, like a GNU __attribute__ construct,
    and I had already kludged that construct by declaring '__asm__' and 'asm'
    as lexical keywords and defining their use.  Then I added the __asm__
    function-like construct, and had to be careful to not clash with the
    post-function form.  This is all somewhat below the depths to which I
    normally descend into language/grammar issues.

    Please pardon my incomplete understanding of this grammar issue, and if
    ever anyone were to come across this code and see a better way to do
    this, I'd be interested in hearing from you.  (This is probably about the
    probability of someone coming across Voyageur II and getting in touch ... )
*/

assembly
	: ASM VOLATILE '(' asm_param_list ')'
	| ASM '(' asm_param_list ')'
	;

asm_param_list
	: asm_param_list ':' asm_param
	| asm_param
	;

asm_param
	: asm_param ',' asm_register
	| asm_register
	;

asm_register
	: postfix_expr
	; 

/*  End of assembly statement experiment */

attribute
	: ATTRIBUTE '(' '(' declarator2 ')' ')'
	| ATTRIBUTE '(' '(' type_specifier ')' ')'
	| ASM '(' argument_expr_list ')'
	;

pointer
	: '*'  { $$ = createTokList( NULL ); $$->ptrLvl++; }
	| '*' type_specifier_list { $$ = $2; $$->ptrLvl++; }
	| '*' pointer { $$ = $2; $$->ptrLvl++; }
	| '*' type_specifier_list pointer
               { $$ = $2; $$->ptrLvl++; }
	;

type_specifier_list
	: type_specifier { $$ = createTokList( $1 ); }
	| type_specifier_list type_specifier
              { $$ = addTokList( $2, $1, 1 ); }
	;

parameter_identifier_list
	: identifier_list
	| identifier_list ',' ELIPSIS
	;

identifier_list
	: identifier { $$.iVal = 0; }
	| identifier_list ',' identifier
                     { $$.iVal = 0; }

/*  Needed to add these for parameter lists defined with integers; saw this in
    stdio.h '__attribute__ ((__format__ (__printf__, 3, 4)));'          */
	| CONSTANT { $$.iVal = 0; }
	| identifier_list ',' CONSTANT
                     { $$.iVal = 0; }
	;

parameter_type_list
	: parameter_list { $$ = $1; }
	| parameter_list ',' ELIPSIS { $$ = $1; }
	;

parameter_list
	: parameter_declaration { $$ = $1; }
	| parameter_list ',' parameter_declaration
                                { $$ = addParmList( $1, $3 ); }
	;

parameter_declaration
	: type_specifier_list declarator
                     { $$ = addParmList(
                               createParmList( createDecl( $1 ), SCALAR ),
                               createParmList( $2, $2->declClass ) );
                     }
	| type_name  { $$ = createParmList( createDecl( $1 ), SCALAR ); }
	;

type_name
	: type_specifier_list  { $$ = $1; } 
	| type_specifier_list abstract_declarator
                               { $$ = concatTokList( $1, $2, 1 ); }
	;

abstract_declarator
	: pointer { $$ = $1; }
	| abstract_declarator2 { $$ = NULL; }
	| pointer abstract_declarator2 { $$ = $1; }
	;

abstract_declarator2
	: '(' abstract_declarator ')'
	| '[' ']'
	| '[' constant_expr ']'
	| abstract_declarator2 '[' ']'
	| abstract_declarator2 '[' constant_expr ']'
	| '(' ')'
	| '(' parameter_type_list ')'
	| abstract_declarator2 '(' ')'
	| abstract_declarator2 '(' parameter_type_list ')'
	;

initializer
	: assignment_expr
	| '{' initializer_list '}'
	| '{' initializer_list ',' '}'
	;

initializer_list
	: initializer
	| initializer_list ',' initializer
	;

statement
	: labeled_statement
	| compound_statement
	| expression_statement
	| selection_statement
	| iteration_statement
	| jump_statement
	;

/*
 *  Modified from the original grammar.  It was 'identifier ':' statement',
 *  which works when there are statements following the label, but doesn't
 *  work when there aren't, like
 *       ...
 *    DONE:
 *       }
 *  However, making the labeled_statement be just the label, not the code
 *  following it, works because it gets to 'statement' from 'statement_list'
 *  in all cases.
*/

labeled_statement
	: identifier ':' { $$.iVal = 0; }
	| CASE constant_expr ':' statement
	| DEFAULT ':' statement
	;

compound_statement
	: '{' '}'
	| '{' statement_list '}'
	| '{' declaration_list '}' { freeFuncDef( $2 ); }
	| '{' declaration_list statement_list '}' { freeFuncDef( $2 ); }
	;

declaration_list
	: declaration { $$ = $1; }
	| declaration_list declaration { $$ = addKRFunc( $1, $2 ); }
	;

statement_list
	: statement
	| statement_list statement
	;

expression_statement
	: ';'
	| expr ';'
	;

selection_statement
	: IF '(' expr ')' statement
	| IF '(' expr ')' statement ELSE statement
	| SWITCH '(' expr ')' statement
	;

iteration_statement
	: WHILE '(' expr ')' statement
	| DO statement WHILE '(' expr ')' ';'
	| FOR '(' ';' ';' ')' statement
	| FOR '(' ';' ';' expr ')' statement
	| FOR '(' ';' expr ';' ')' statement
	| FOR '(' ';' expr ';' expr ')' statement
	| FOR '(' expr ';' ';' ')' statement
	| FOR '(' expr ';' ';' expr ')' statement
	| FOR '(' expr ';' expr ';' ')' statement
	| FOR '(' expr ';' expr ';' expr ')' statement
	;

jump_statement
	: GOTO identifier ';'
	| CONTINUE ';'
	| BREAK ';'
	| RETURN ';'
	| RETURN expr ';'
	;

file
	: external_definition      { addFunc( $1 ); }
	| file external_definition { addFunc( $2 ); }
	;

external_definition
	: function_definition { $$ = $1;
                                if( cmtMode != CMTHDR ) {
                                    intCommentText[intCmtIdx] = '\0'; 
                                    intCmtIdx = 0;
                                    cmtMode = CMTHDR;
                                }
                              }
	| declaration         { $$ = NULL;
                                if( cmtMode != CMTHDR ) {
                                    intCommentText[intCmtIdx] = '\0'; 
                                    intCmtIdx = 0;
                                    cmtMode = CMTHDR;
                                }
                              }
/*
 *  Added the following rule to allow for the stray semi-colon outside of
 *  any function
*/

	| ';' { $$ = NULL; }
	;

function_definition
	: declarator function_body { $$ = createFunc( NULL, $1, NULL ); }
	| declaration_specifiers declarator function_body
                    { $$ = createFunc( $1, $2, $3 ); }
	;

function_body
	: compound_statement                  { $$ = NULL; }
	| declaration_list compound_statement { $$ = $1;   }
	;

identifier
	: IDENTIFIER { $$ = $1; }
	;
%%

#ifdef UNDEF
------------------------ BEGIN ILLEGITIMATE STUFF HERE ------------------------
------------------------- END ILLEGITIMATE STUFF HERE -------------------------
#endif /* UNDEF */
