#! /usr/bin/perl

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'srcdocui.pl,v control version 1.6 2002/01/05 03:12:03'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "Top-level user interface for the srcdoc program";


#
#  copyright (c) 2000, Gordon B. Aaseng
#                      Glendale, AZ
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  
#  This software is protected as an unpublished work under the U.S. Copyright
#  Act of 1976.  All rights reserved by Gordon B. Aaseng
#

#
#  FILE NAME
#
#  srcdocui.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  Creates the top level user interface for srcdoc.  The user interface
#  requires perl/Tk.  See the INSTALL file for details on obtaining the
#  perl/Tk distribution.  The user interface consist of menu bars to control
#  the major functions, a list of functions visible to the program, displays
#  for directories currently in use, and a status display.  For several
#  operations, additional displays are opened.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  srcdocui.pl,v
#  Revision 1.6  2002/01/05 03:12:03  gordon
#  Changed license to GPL.
#
#  Revision 1.5  2001/12/31 04:37:17  gordon
#  Changed directory and file handling to correctly process tilde
#  substitution and improved validation.
#
#  Revision 1.4  2001/12/28 06:00:35  gordon
#  Changed storage class handling; changes in option handling necessitated
#  changes in data structures that were updated in the GUI.
#
#  Revision 1.3  2001/12/01 06:00:06  gordon
#  Fixed concatenation of current directory with relative pathname of the
#  HTML index file.
#
#  Revision 1.2  2001/11/30 12:56:19  gordon
#  Added controls for HTML formatting, including directory, index and
#  format selections.
#
#  Revision 1.1  2000/12/14 05:36:26  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

use Tk;

require Tk::Menubar;
require Tk::Menu::Item;
require Tk::FileSelect;

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use subs qw( DisplayUsage DisplayCmInfo );
use subs qw( SetConfig EditSource SetDirectives SetTagFile ExitGui ); 
use subs qw( UpdateFuncList GetManDir SrcSelect GetSelection SetExtractMode);
use subs qw( ValidateDir );

# ***************************************************************
#
#  Subroutine USAGE
#
#  Print script help message
#
# ***************************************************************

sub DisplayUsage {
   printf( STDOUT "Usage:   require 'srcdocui.pl'\n" );
   return;
}

# ***************************************************************
#
#  Subroutine CMINFO
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub DisplayCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}

# ***************************************************************
#
#  Main script body
#
# ***************************************************************



# ***************************************************************
#
#  Subroutine DisplayUserIf
#
#  Description:  This function is used to display a user interface using
#                the Perl/Tk library.  The user interface displays a list
#                of the functions that were found in the source code.
#                Selecting an item and then selecting the 'edit' button
#                opens up the text in a gvim editor.
#
# ***************************************************************

sub DisplayUserIf {
   my  @func_names;
   my  $fname;
   my  $fct;
   my  $MW;
   my  $flist;
   my  $bcmd;

   $mkdocgui_cnt++;

#  An array of the function names should have been collected into a global
#  array.  This selects the storage types to be displayed, based on the
#  user interface selections made.

   if( $#GblFuncnames >= 0 ) {
      foreach $fct (@GblFuncnames) {
         if( $StorageMode{public} eq $True and
             $FuncList{$fct}->{stg} ne "static" ) {
            push( @func_names, $fct );
         }
         if( $StorageMode{static} eq $True and
             $FuncList{$fct}->{stg} eq "static" ) {
            push( @func_names, $fct );
         }
      }
   }

   %PageIndexList = ();

#  The top-level container.

   $MW = MainWindow->new;

#  This frame contains pull-down menus with application-level controls.

   my $MainControlFrame = $MW->Frame();

   $MainControlFrame->pack( -side => 'top', -fill => 'x' );


#  This frame contains the file list and some current settings that are
#  displayed in label/entry pairs.

   my $WorkFrame = $MW->Frame( -bg  => 'gray62' );
   $WorkFrame->pack( -side => 'top' );

#  This is a frame that contains the listbox for function names, and a label
#  The hierarchy is
#     flistframe          (Frame)
#        captionlabel        (Label)
#        scroll_listframe    (Frame)
#           scroll              (Scrollbar)
#           flist               (Listbox)

   my $flistframe = $WorkFrame->Frame(
               -relief     => 'raised',
               -borderwidth         => 2
             );
   $flistframe->pack( -side => 'left' );

   my $captionlabel = $flistframe->Label(
                 -text        => "Function Names",
                 -anchor      => 'n'
               );
   $captionlabel->pack( -side => 'top' );

   my $scroll_listframe = $flistframe->Frame();
   $scroll_listframe->pack( -side => 'top' );

   my $scroll = $scroll_listframe->Scrollbar();
   $scroll->pack( -side => 'right', -fill => 'y' );

   $flist = $scroll_listframe->Listbox(
               -yscrollcommand => ['set', $scroll],
               -selectmode     => 'extended',
               -relief         => 'sunken',
               -width          => 30,
               -height         => 22,
               -setgrid        => 'yes',
             );

#  Create a frame for information like the man directory, configuration file
#  and maybe other stuff.
#  The widget hierarchy is:
#     infoframe          (Frame)
#        infolabel          (Label)
#        man_dirframe       (Frame)
#           man_dirlabel       (Label)
#           man_direntry       (Entry)
#        manrcframe         (Frame)
#           manrclabel         (Label)
#           manrcentry         (Entry)

   my $infoframe = $WorkFrame->Frame(
               -bg                  => 'DarkGray',
               -borderwidth         => 2,
               -highlightthickness  => 2,
               -relief              => 'raised'
             );
   $infoframe->pack( -side => 'left', -anchor => 'n', -fill => 'y' );

   my $infolabel = $infoframe->Label(
               -text               => 'Man Page Settings',
               -width              => 39,
               -height             => 2,
               -anchor             => 'n'
             );
   $infolabel->pack( -side => 'top', -expand => 'yes', -fill => 'both' );

   my $infoht = 24;
   my $label_size = 16;

#  Create a source directory panel for displaying and setting the current
#  source code directory.

   my $src_dirframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $src_dirframe->pack( -side => 'top' );

   my $src_dirlabel = $src_dirframe->Label(
                 -text => "Source Directory",
                 -width => $label_size,
                 -anchor  => 'w'
               );

   $src_dirlabel->place( -in => $src_dirframe, -relx => 0.0 );
   my $src_direntry = $src_dirframe->Entry(
                         -width => 20,
                         -textvariable => \$DispSrcDir
                       );
   $src_direntry->place( -in => $src_dirlabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );

   $src_direntry->bind( '<Return>'   =>
                        sub { ValidateDir( \$DispSrcDir, \$SrcDir ) } );
   $src_direntry->bind( '<FocusOut>' =>
                        sub { ValidateDir( \$DispSrcDir, \$SrcDir ) } );

#  Create a man directory panel for displaying and setting the current
#  man page directory.

   my $man_dirframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $man_dirframe->pack( -side => 'top' );

   my $man_dirlabel = $man_dirframe->Label(
                 -text => "Man Directory",
                 -width => $label_size,
                 -anchor  => 'w'
               );

   $man_dirlabel->place( -in => $man_dirframe, -relx => 0.0 );

   my $man_direntry = $man_dirframe->Entry( -width => 20,
                                            -textvariable => \$DispManDir );
   $man_direntry->place( -in => $man_dirlabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );
   if( $ManDir eq "" ) {
      my $mandir;
      $mandir = qx/pwd/;
      chomp( $mandir );
      $ManDir = $mandir;
      $DispManDir = $ManDir;
   }

   $man_direntry->bind( '<Return>'   =>
                        sub { ValidateDir( \$DispManDir, \$ManDir ) } );
   $man_direntry->bind( '<FocusOut>' =>
                        sub { ValidateDir( \$DispManDir, \$ManDir ) } );

   my $html_dirframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $html_dirframe->pack( -side => 'top' );

   my $html_dirlabel = $html_dirframe->Label(
                 -text => "HTML Directory",
                 -width => $label_size,
                 -anchor  => 'w'
               );

   $html_dirlabel->place( -in => $html_dirframe, -relx => 0.0 );
   my $html_direntry = $html_dirframe->Entry( -width => 20,
                                              -textvariable => \$DispHtmlDir );
   $html_direntry->place( -in => $html_dirlabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );
   if( $HtmlDir eq "" ) {
      my $htmldir;
      $htmldir = qx/pwd/;
      chomp( $htmldir );
      $HtmlDir = $htmldir;
      $DispHtmlDir = $HtmlDir;
   }

   $html_direntry->bind( '<Return>'   =>
                         sub { ValidateDir( \$DispHtmlDir, \$HtmlDir ) } );
   $html_direntry->bind( '<FocusOut>' =>
                         sub { ValidateDir( \$DispHtmlDir, \$HtmlDir ) } );

#  Create a panel for displaying the editor tag file name.

   my $tagframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
                );
   $tagframe->pack( -side => 'top' );
   my $taglabel = $tagframe->Label(
                 -text => "Editor Tag File",
                 -width => $label_size,
                 -anchor  => 'w'
               );
   $taglabel->place( -in => $tagframe, -relx => 0.0 );
   my $tagentry = $tagframe->Entry(
                     -width        => 20,
                     -textvariable => \$DispTagFile
                    );
   $tagentry->place( -in     => $taglabel,
                     -relx   => 1.0,
                     -rely   => 0.5,
                     -anchor => 'w'
                   );

   $tagentry->bind( '<Return>' => [ \&SetTagFile ] );
   $tagentry->bind( '<FocusOut>' => [ \&SetTagFile ] );

#  Create a panel for displaying the HTML index file, if HTML is selected.

   my $indexframe = $infoframe->Frame(
                 -width               => 280,
                 -height              => $infoht
               );
   $indexframe->pack( -side => 'top' );
   my $indexlabel = $indexframe->Label(
                 -text => "HTML Index File",
                 -width => $label_size,
                 -anchor  => 'w'
               );
   $indexlabel->place( -in => $indexframe, -relx => 0.0 );
   my $indexentry = $indexframe->Entry(
                     -width        => 20,
                     -textvariable => \$DispIndexFile
                    );
   $indexentry->place( -in     => $indexlabel,
                       -relx   => 1.0,
                       -rely   => 0.5,
                       -anchor => 'w'
                     );

   $bcmd = sub {
                 my $oldfile = $IndexFile;
                 my $newfile = $DispIndexFile;
                 TildeSub( \$newfile );
                 if( $newfile eq "" ) { $IndexFile = ""; }
                 elsif ( $newfile !~ /^\// ) {
                   $newfile =~ s/\.\///;
                   $IndexFile = $SrcDir . '/' . $newfile;
                 } else { $IndexFile = $newfile; }
                 if( $IndexFile ne $oldfile ) {
                   if( $IndexFile ne "" ) {
                     PrintStat( "HTML Index set to $IndexFile", 'advise' );
                   } else { PrintStat( "HTML Index File unset", 'advise' ); }
                 }
               };

   $indexentry->bind( '<Return>'   => $bcmd );
   $indexentry->bind( '<FocusOut>' => $bcmd );

#  Create a panel that contains checkbuttons for selecting the function
#  storage types to be processed (public or static).

   my $storageframe = $infoframe->Frame( )->pack( -side => 'top', -fill =>'x' );
   my $storagelabel = $storageframe->Label(
                     -text => "Function Types"
                   );
   $storagelabel->pack( -side => 'left' );

   $bcmd = [ \&UpdateFuncList, $MW, $flist,
             \%StorageMode, \@func_names, "REFRESH" ];

   my $storage_public = $storageframe->Checkbutton(
                      -text     => "Public",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$StorageMode{public},
                      -command  => $bcmd
                   );
   $storage_public->pack( -side => 'left' );
   $storage_public->bind( '<FocusOut>' => $bcmd );

   my $storage_static = $storageframe->Checkbutton(
                      -text     => "Static",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$StorageMode{static},
                      -command  => $bcmd
                   );
   $storage_static->pack( -side => 'left' );
   $storage_static->bind( '<FocusOut>' => $bcmd );

#  Create a panel that contains checkbuttons for selecting the output
#  type, man or HTML.  Both or none can be selected.

   my $pagetypeframe = $infoframe->Frame( )->pack( -side => 'top',
                                                   -fill => 'x'
                                                  );
   my $pagetypelabel = $pagetypeframe->Label( -text => "Page Format" );
   $pagetypelabel->pack( -side => 'left' );

   my $page_man = $pagetypeframe->Checkbutton(
                      -text     => "Man Page",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$ManFormat
                   );
   $page_man->pack( -side => 'left' );

   my $page_html = $pagetypeframe->Checkbutton(
                      -text     => "HTML",
                      -onvalue  => $True,
                      -offvalue => $False,
                      -variable => \$HtmlFormat
                   );
   $page_html->pack( -side => 'left' );

#  Create a status frame for displaying activity, errors and other info

   my $StatusFrame = $infoframe->Frame(
               -height              => 5,
               -borderwidth         => 2,
             );
   $StatusFrame->pack( -side =>'bottom', -fill => 'x' );

   my $StatLabelFrame = $StatusFrame->Frame(
              -height          => 20
             )->pack( -side    => 'top' );
   my $StatMsgFrame = $StatusFrame->Frame(
             )->pack( -side    => 'top' );

   my $StatScroll = $StatMsgFrame->Scrollbar();
   $StatScroll->pack( -side => 'right', -fill => 'y' );

   my $StatLabel = $StatLabelFrame->Label(
               -text           => "Status",
               -width          => 30,
               -height         => 2 
              );
   $StatLabel->pack( -side => 'left' );

   my $StatList = $StatMsgFrame->Text(
               -yscrollcommand => ['set', $StatScroll],
               -state          => 'disabled',
               -relief         => 'sunken',
               -width          => 34,
               -height         => 10,
               -setgrid        => 'yes',
               -wrap           => 'none'
             );
   $StatList->pack( -side => 'top' );
   $StatScroll->configure( -command => ['yview', $StatList] );

   $StatWin = $StatList;


   $flist->pack( -side => 'left', -fill => 'both', -expand => 'yes' );
   $scroll->configure( -command => ['yview', $flist] );
   $flist->bind( '<Double-1>' => sub {
                        GetSelection( $flist, \@func_names, $Edit ); }
               );

#  Here it comes back to the menus and fills in the pull-downs.
#  It's done after all the other widgets are realized so that they can act
#  on the information in the working windows.

   my $FileMenu = $MainControlFrame->Menubutton(
                       -text      => "File",
                       Name       => 'filemenu'
                     );
   $FileMenu->command( -label   => "Exit",
                       -command => [ \&ExitGui ] );

   $FileMenu->pack( -side => 'left', -padx => 3 );

   my $ManMenu = $MainControlFrame->Menubutton(
                       -text     => 'ManPage',
                       Name      => 'manmenu'
                     );

   $ManMenu->command( -label   => 'Edit Man Text',
                      -command =>
                          [ \&GetSelection, $flist, \@func_names, $Edit ]
                     );
   $ManMenu->command( -label   => 'Generate Man Page',
                      -command =>
                         [ \&GetSelection, $flist, \@func_names, $Generate ]
                     );
   $ManMenu->command( -label    => 'Set Man Dir',
                      -command  =>
                            [ \&GetManDir, $MW, $man_direntry ]
                    );
   $ManMenu->command( -label    => 'Set Man Config...',
                      -command  => [ \&SetConfig, $MW ]
                    );

   $ManMenu->command( -label    => 'View Man Page',
                      -command =>
                         [ \&GetSelection, $flist, \@func_names, $View ]
                    );
   $ManMenu->command( -label    => 'Edit Man File',
                      -command =>
                         [ \&GetSelection, $flist, \@func_names, $Editman ]
                    );
   $ManMenu->command( -label    => 'Write HTML Index',
                      -command =>
                         [ \&PrintHtmlIndex ]
                    );

   $ManMenu->pack( -side => 'left', -padx => 3 );

   my $SourceMenu = $MainControlFrame->Menubutton(
                       Name      => 'srcmenu',
                       -text     => 'Source'
                     );
   $SourceMenu->command( -label   => 'View Source',
                         -command => [ \&EditSource, $MW ]
                       );
   $SourceMenu->command( -label    => 'Extract Comments',
                 -command => [ \&UpdateFuncList, $MW, $flist,
                                \%StorageMode, \@func_names, "SELECT" ]
                     );
   $SourceMenu->command( -label    => 'Set Extraction Mode ...',
                         -command  => [ \&SetExtractMode, $MW ]
                       );
   $SourceMenu->command( -label    => 'Select Tag File...',
                       -command => [ \&SrcSelect, $MW, "tag" ]
                     );
   $SourceMenu->command( -label    => 'Set CPP Directives ...',
                       -command => [ \&SetDirectives, $MW ]
                     );
   $SourceMenu->pack( -side => 'left', -padx => 3 );

   foreach $fname (@func_names) {
      $flist->insert( 'end', $fname );
   }
   MainLoop;
   return;
}

# ***************************************************************
#
#  Subroutine ValidateDir
#
#  Description:  Checks for the existence of a directory before resetting
#                the directory name to the value entered in the GUI.
#                If the directory starts with a tilde (~), substitutes
#                the home directory before the check.
#
# ***************************************************************

sub ValidateDir {
   my $dispdir = shift;
   my $realdir = shift;

   my $subst_text = $$dispdir;

   TildeSub( \$subst_text );
   return if( $subst_text eq $$realdir );

   if( defined( $subst_text ) ) {
      if( -e $subst_text ) { $$realdir = $subst_text; }
      else {
         PrintStat( "Invalid Directory $$dispdir", 'warning' );
         $$dispdir = $$realdir;
      }
   }
}

# ***************************************************************
#
#  Subroutine TildeSub
#
#  Description:  Substitutes a leading tilde with the home directory.
#
# ***************************************************************

sub TildeSub {
   my  $fname = shift;

   if( $$fname =~ /^~/ ) {
      $$fname =~ s/^~/$ENV{HOME}/;
   }
}

# ***************************************************************
#
#  Subroutine ExitGui
#
#  Description:  Does some finalization and then exits.
#
# ***************************************************************

sub ExitGui {
   if( $Verbose eq $True ) {
      PrintExecStats;
   }
   PrintHtmlIndex();
   exit( 0 );
}

1;
