#! /usr/bin/perl

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'srcproc.pl,v control version 1.4 2002/01/05 03:14:25'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "Functions for processing and displaying source code";


#
#  copyright (c) 2000, 2001, Gordon B. Aaseng
#                      Glendale, AZ
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  
#  This software is protected as an unpublished work under the U.S. Copyright
#  Act of 1976.  All rights reserved by Gordon B. Aaseng.
#

#
#  FILE NAME
#
#  srcproc.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This file contains user interface functions for selecting source code,
#  setting pre-compiler directives and displaying source code modules.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  srcproc.pl,v
#  Revision 1.4  2002/01/05 03:14:25  gordon
#  Changed license to GPL.
#
#  Revision 1.3  2001/12/31 04:40:19  gordon
#  Changed tag file handling to process tilde substitutions.
#
#  Revision 1.2  2001/12/28 06:08:32  gordon
#  Added panel for selecting comment types (header or internal function) and
#  update policy.
#
#  Revision 1.1  2000/12/14 05:37:07  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use subs qw( spUsage spCmInfo );
use subs qw( SetCflags );

# ***************************************************************
#
#  Subroutine USAGE
#
#  Print script help message
#
# ***************************************************************

sub spUsage {
   printf( STDOUT "Script synopsis:\n" );
   printf( STDOUT "Usage:   require 'srcproc.pl'\n" );
   return;
}

# ***************************************************************
#
#  Subroutine CMINFO
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub spCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}

# ***************************************************************
#
#  Main script body
#
# ***************************************************************


# ***************************************************************
#
#  Subroutine EditSource
#
#  Description:  This function opens selected source code files in a
#                'vi' editor.
#
# ***************************************************************

sub EditSource {
   my $top      = shift;

   my $srcfile;
   my $curdir;

   my $rval = SrcSelect( $top, "c" );
   if( $rval < 0 ) { return; }

   if( $SrcDir ne $DispSrcDir ) {
      $DispSrcDir = $SrcDir;
      PrintStat( "Source directory set to $SrcDir", 'advise' );
   }

   my $editcmd = "gvim +\"set tag=$TagFile\"";
   foreach $srcfile (@SrcFiles) {
      $editcmd .= ' ' . $srcfile;
   }
   $editcmd .= " >/dev/null &";
   system( $editcmd );
   return;
}

# ***************************************************************
#
#  Subroutine SetDirectives
#
#  Description:  This function brings up a panel for setting precompiler
#                directives that are used for precompiling the selected
#                source code files.  The files are then processed to
#                extract function information and comments from the file.
#
# ***************************************************************

sub SetDirectives {
   my $top = $_[0];
   my $incl_directive;
   my $def_directive;
   my $flag_directive;
   my @flags;
   my $flag;

   my $label_size = 16;

   my $dirtop = $top->Toplevel( -title => 'Compiler Directives' );

   my $dirframe = $dirtop->Frame();
   $dirframe->pack( -side => 'top' );

   my $inclframe = $dirframe->Frame(
                 -width               => 340,
                 -height              => 24
                );
   $inclframe->pack( -side => 'top' );

   my $inclabel = $inclframe->Label(
                 -text => "Include Dirs",
                 -width => $label_size,
                 -anchor  => 'w'
               );
   $inclabel->place( -in => $inclframe, -relx => 0.0 );

   my $inclentry = $inclframe->Entry( -width => 30 );
   $inclentry->place( -in => $inclabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );

   my $defframe = $dirframe->Frame(
                 -width               => 340,
                 -height              => 24
                );
   $defframe->pack( -side => 'top' );

   my $deflabel = $defframe->Label(
                   -text    => "Defines",
                   -width   => $label_size,
                   -anchor  => 'w'
                 );
   $deflabel->place( -in => $defframe, -relx => 0.0 );
   my $defentry = $defframe->Entry( -width => 30 );
   $defentry->place( -in => $deflabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );

   my $flagframe = $dirframe->Frame(
                 -width               => 340,
                 -height              => 24
                );
   $flagframe->pack( -side => 'top' );

   my $flaglabel = $flagframe->Label(
                   -text    => "Compiler Flags",
                   -width   => $label_size,
                   -anchor  => 'w'
                 );
   $flaglabel->place( -in => $flagframe, -relx => 0.0 );
   my $flagentry = $flagframe->Entry( -width => 30 );
   $flagentry->place( -in => $flaglabel,
                         -relx => 1.0,
                         -rely => 0.5,
                         -anchor => 'w'
                       );

   my $qframe = $dirframe->Frame();
   $qframe->pack( -side => 'top' );

   @flags = split( ' ', $Cflags );

   foreach $flag (@flags) {
      if( $flag =~ /\-I/ ) {
         $flag =~ s/\-I//;
         $flag .= ' ';
         $inclentry->insert( 'end', $flag );
      }
      elsif( $flag =~ /\-D/ ) {
         $flag =~ s/\-D//;
         $flag .= ' ';
         $defentry->insert( 'end', $flag );
      }
      else{
         $flag .= ' ';
         $flagentry->insert( 'end', $flag );
      }
   }

   my $qbtn = $qframe->Button(
                -text => 'OK',
                -command => sub { $incl_directive = $inclentry->get();
                                  $def_directive  = $defentry->get();
                                  $flag_directive = $flagentry->get();
                                  SetCflags( $incl_directive,
                                             $def_directive,
                                             $flag_directive );
                                  $dirtop->withdraw }
             );
   $qbtn->pack( -side => 'top' );
   return;
}

# ***************************************************************
#
#  Subroutine SetCflags
#
#  Description:  This function takes compiler flag information from the
#                compiler directives panel and formats it for use by
#                the precompiler, setting the global variable 'Cflags'.
#
# ***************************************************************

sub SetCflags {
   my ($incl_str, $def_str, $flag_str) = @_;

   my $incl;
   my $def;
   my $sep    = "";
   my $Iflags = "";
   my $dflags = "";
   my @includes = split( ' ', $incl_str );
   my @defines  = split( ' ', $def_str  );

   foreach $incl (@includes) {
      $incl =~ s/\-I//;
      if( length( $incl ) > 0 ) { $sep = ' '; }
      $Iflags .= $sep . "-I" . $incl; 
   }
   foreach $def (@defines) {
      $def =~ s/\-D//;
      if( length( $def ) > 0 ) { $sep = ' '; }
      $dflags .= $sep . "-D" . $def; 
   }
   $Cflags = $Iflags . ' ' . $dflags;
   if( length( $flag_str ) > 0 ) { $Cflags .= ' ' . $flag_str; }

   if( $Cflags ne "" ) {
      PrintStat( "Compiler directives set to $Cflags", 'advise' );
   }

   return;
}

# ***************************************************************
#
#  Subroutine SetTagFile
#
#  Description:  Sets the tag file to the value entered in the tag file
#                entry panel, after checking for name validity.
#                If a relative path name is entered, the file is set
#                relative to the source directory ($SrcDir)
#
# ***************************************************************

sub SetTagFile {

   my $newtag = "";
   my $expanded_tag = $DispTagFile;

   TildeSub( \$expanded_tag );

   if( $expanded_tag eq $TagFile ) { return; }

   if( $expanded_tag !~ /^\// ) { $newtag = $SrcDir . '/' . $DispTagFile; }
   else { $newtag = $expanded_tag; }


   if( -f $newtag ) {
      $TagFile = $newtag;
      PrintStat( "Set Tag File to $TagFile", 'advise' );
   }
   else {
      $DispTagFile = $TagFile;
      PrintStat( "Could not find $newtag", 'warning' );
   }
   return;
}

# ***************************************************************
#
#  Subroutine SetExtractMode
#
#  Description:  Pops up a panel for setting comment selection options,
#                to include when to rewrite the intermediate *.mtp file
#                and whether to use the comments internal to the function
#                source code or only the function headers.
#
# ***************************************************************

sub SetExtractMode {
   my $top   = shift;


   my $em_top = $top->Toplevel( -title => 'Comment Extraction Modes' );

   my $em_frame = $em_top->Frame();
   $em_frame->pack( -side => 'top' );

   my $em_hdr = $em_frame->Label(
                      -text   => "Specify Comment Extraction Methods",
                      -width  => 40,
                      -height => 2
                                );
   $em_hdr->pack( -side => 'top' );

   my $selectframe = $em_frame->Frame();
   $selectframe->pack( -side => 'top', -anchor => 'w', -padx => '0.2i' );

   my $modelabel = $selectframe->Label(
                       -relief => 'raised',
                       -borderwidth => 2,
                       -text => "Comment Types"
                                        );
   $modelabel->pack( -side => 'top' );

   my $modeallbtn = $selectframe->Radiobutton(
                        -variable => \$CmtFlag,
                        -value    => $AllCmt,
                        -text => "All Function Comments"
                                          );
   $modeallbtn->pack( -side => 'top', -anchor => 'w' );

   my $modehdrbtn = $selectframe->Radiobutton(
                        -variable => \$CmtFlag,
                        -value    => $HdrCmt,
                        -text     => "Header Comments Only"
                                             );
   $modehdrbtn->pack( -side => 'top', -anchor => 'w' );

#  Create a frame for the label and buttons for the ReComment setting

   my $recmtframe = $em_frame->Frame();
   $recmtframe->pack( -side => 'top', -anchor => 'w', -padx => '0.2i' );

   my $recmtlabel = $recmtframe->Label(
                       -relief => 'raised',
                       -borderwidth => 2,
                       -text => "Update Policy"
                                       );
   $recmtlabel->pack( -side => 'top' );

   my $recmtsrc = $recmtframe->Radiobutton(
                       -variable => \$ReComment,
                       -value    => $SrcCmtAlways,
                       -text     => "Always Extract"
                                           );
   $recmtsrc->pack( -side => 'top', -anchor => 'w' );

   my $recmtnew = $recmtframe->Radiobutton(
                       -variable => \$ReComment,
                       -value    => $NewCmtAlways,
                       -text     => "Compare to Source"
                                           );
   $recmtnew->pack( -side => 'top', -anchor => 'w' );

   my $recmtedt = $recmtframe->Radiobutton(
                       -variable => \$ReComment,
                       -value    => $EdtCmtAlways,
                       -text     => "Extract First Time Only"
                                           );
   $recmtedt->pack( -side => 'top', -anchor => 'w' );

   my $em_okbtn = $em_frame->Button(
                -text    => 'OK',
                -command => sub {
                   $em_top->withdraw
                }
               );
   $em_okbtn->pack( -side => 'top' );

   return;
}

1;
