#! /usr/bin/perl

my $osname;
my $ident;
my $cpyrght;
my $usage;

$osname = `uname -s`;
chomp $osname;

if( $osname eq "Solaris" ) {
   $ident = '%Z%%M%  control version %I%  %G% %U%'; }
if( $osname eq "Linux" ) {
   $ident = 'statwin.pl,v control version 1.3 2002/01/05 03:16:03'; }
$cpyrght = "Copyright (c) 2000, Gordon B. Aaseng";

$usage = "Status display functions for the srcdoc program";


#
#  copyright (c) 2000, Gordon B. Aaseng
#                      Glendale, AZ
#
#  This program is free software; you can redistribute it and/or modify it
#  under the terms of the GNU General Public License as published by the
#  Free Software Foundation; either version 2 of the License, or any later
#  version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License along
#  with this distribution; if not, write to the Free Software Foundation, Inc.,
#  59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#  

#
#  FILE NAME
#
#  statwin.pl
#
#
#  PROGRAM INFORMATION
#
#  Developer:		Gordon B. Aaseng, gba, <gaaseng@users.sourceforge.net>
#  Under License:	General Public License
#
#
#  FILE DESCRIPTION
#
#  This file contains the functions for a status display in the 'srcdoc'
#  program user interface.  It is included by srcdoc.pl and depends on
#  variables instantiated in that file.
# 
#
#  RCS CHANGE LOG
#     NOTE:  Do not edit this section; it is managed by RCS.
#  statwin.pl,v
#  Revision 1.3  2002/01/05 03:16:03  gordon
#  Changed license to GPL.
#
#  Revision 1.2  2000/12/17 14:11:10  gordon
#  Added a check of the $DisplayOpen variable to determine if it can
#  write to the GUI or should use stdout.
#
#  Revision 1.1  2000/12/14 05:38:57  gordon
#  Initial revision
#
#
# ********************  End of RCS CHANGE LOG *******************
#
#

# ***************************************************************
#
#  Global Data Declarations
#
# ***************************************************************

my $stat_line = 1;
my %StatColor;
$StatColor{warning}{fg} = "red";
$StatColor{warning}{bg} = "white";
$StatColor{caution}{fg} = "orange";
$StatColor{caution}{bg} = "black";
$StatColor{advise}{fg}  = "yellow";
$StatColor{advise}{bg}  = "black";


# ***************************************************************
#
#  Executable Code (External Modules)
#
# ***************************************************************

# (None.)

# ***************************************************************
#
#  Executable Code (Locally Used Functions)
#
# ***************************************************************

use subs qw( StatwinUsage StatwinCmInfo PrintStat );

# ***************************************************************
#
#  Subroutine USAGE
#
#  Print script help message
#
# ***************************************************************

sub StatwinUsage {
   printf( STDOUT "Script synopsis:\n" );
   printf( STDOUT "Usage:   <filename> -h -z <args>\n" );
   return;
}

# ***************************************************************
#
#  Subroutine CMINFO
#
#  Print script Ident string, copyright information and brief script
#  usage messages.
#
# ***************************************************************

sub StatwinCmInfo {
   if( $osname eq "Linux" ) {
      $ident =~ s/\\$//g && s/\,v// && s/Revision\://;
      $vcheck = $ident;
      $vcheck =~ s/[0-9]//g;
      if( $vcheck eq $ident ) {
         $ident = "File not checked in to CM library";
      }
   }
   printf( STDOUT "%s\n", $ident );
   printf( STDOUT "%s\n", $cpyrght );
   printf( STDOUT "%s\n", $usage );

   return;
}

# ***************************************************************
#
#  Main script body
#
# ***************************************************************

# ***************************************************************
#
#  Subroutine PrintStat
#
#  Description:  Print a message in the status display.  The first
#                parameter is the message to display.   The second is
#                a severity level, which sets the background color on
#                the messages.  Previous messages are uncolored, so
#                only the latest message is colored with the severity
#                level.
#
# ***************************************************************

sub PrintStat {
   my  $msg = $_[0];
   my  $slevel = $_[1];

   my  $ccmd;
   my  $pstat;
   my  $tag0;
   my  $tag1;
   my  $tag2;
   my  $fgscolor;
   my  $bgscolor;

   if( $DisplayOpen eq $False ) {
      printf( STDOUT "%s: %s\n", uc( $slevel ), $msg );
      return;
   }

   if( !defined( $slevel ) ) { $slevel = 'advise'; }
   $msg = $msg . "\n";
   $fgscolor = $StatColor{$slevel}{fg};
   $bgscolor = $StatColor{$slevel}{bg};
   $tag1 = "tag" . $stat_line;
   $StatWin->configure( -state => "normal" );
   $StatWin->insert( 'end', $msg );
   $ccmd = "add $tag1 $stat_line.0 $stat_line.end";
   $StatWin->tag(  split( ' ', $ccmd ) );
   $ccmd = "configure $tag1 -background $fgscolor";
   $StatWin->tag(  split( ' ', $ccmd ) );
   $ccmd = "configure $tag1 -foreground $bgscolor";
   $StatWin->tag(  split( ' ', $ccmd ) );

   if( $stat_line > 1 ) {
      $pstat = $stat_line - 1;
      $tag0 = "tag" . $pstat;
      $ccmd = "configure $tag0 -background orange1";
      $StatWin->tag(  split( ' ', $ccmd ) );
      $ccmd = "configure $tag0 -foreground black";
      $StatWin->tag(  split( ' ', $ccmd ) );
   }
   if( $stat_line > 2 ) {
      $pstat = $stat_line - 2;
      $tag2 = "tag" . $pstat;
      $ccmd = "delete $tag2";
      $StatWin->tag(  split( ' ', $ccmd ) );
   }
   $stat_line++;
   $StatWin->configure( -state => "disabled" );

   return;
}

1;
